from unittest import TestCase


class TestCoreModel(TestCase):
    def test_check_model(self):
        self.fail()

    def test__remove_related_files(self):
        self.fail()

    def test_simulations(self):
        self.fail()

    def test_simulation_names(self):
        self.fail()

    def test_run(self):
        self.fail()

    def test_simulated_results(self):
        self.fail()
