
l_model ={
  "$child": "Models.Core.Simulations, Models",
  "ExplorerWidth": 285,
  "Version": 159,
  "Name": "Simulations",
  "ResourceName": None,
  "Children": [
    {
      "$child": "Models.Storage.DataStore, Models",
      "useFirebird": False,
      "CustomFileName": None,
      "Name": "DataStore",
      "ResourceName": None,
      "Children": [],
      "Enabled": True,
      "ReadOnly": False
    },
    {
      "$child": "Models.Core.Folder, Models",
      "ShowInDocs": False,
      "GraphsPerPage": 6,
      "Name": "Replacements",
      "ResourceName": None,
      "Children": [
        {
          "$child": "Models.PMF.Plant, Models",
          "Name": "Maize",
          "ResourceName": "Maize",
          "Children": [
            {
              "$child": "Models.Memo, Models",
              "Text": "_Brown, H.E., Teixeira, E.I., Huth, N.I. and Holzworth, D.P._",
              "Name": "Introduction",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.OrganArbitrator, Models",
              "Name": "Arbitrator",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.PMF.BiomassTypeArbitrator, Models",
                  "Name": "DMArbitration",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Core.Folder, Models",
                      "ShowInDocs": False,
                      "GraphsPerPage": 6,
                      "Name": "PotentialPartitioningMethods",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.PMF.Arbitrator.ReallocationMethod, Models",
                          "Name": "ReallocationMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.PMF.Arbitrator.AllocateFixationMethod, Models",
                          "Name": "AllocateFixationMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.PMF.Arbitrator.RetranslocationMethod, Models",
                          "Name": "RetranslocationMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.PMF.Arbitrator.SendPotentialDMAllocationsMethod, Models",
                          "Name": "SendPotentialDMAllocationsMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Core.Folder, Models",
                      "ShowInDocs": False,
                      "GraphsPerPage": 6,
                      "Name": "AllocationMethods",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.PMF.Arbitrator.NutrientConstrainedAllocationMethod, Models",
                          "Name": "NutrientConstrainedAllocationMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.PMF.Arbitrator.DryMatterAllocationsMethod, Models",
                          "Name": "DryMatterAllocationsMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.RelativeAllocation, Models",
                      "Name": "ArbitrationMethod",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.BiomassTypeArbitrator, Models",
                  "Name": "NArbitration",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Core.Folder, Models",
                      "ShowInDocs": False,
                      "GraphsPerPage": 6,
                      "Name": "PotentialPartitioningMethods",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.PMF.Arbitrator.ReallocationMethod, Models",
                          "Name": "ReallocationMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Core.Folder, Models",
                      "ShowInDocs": False,
                      "GraphsPerPage": 6,
                      "Name": "ActualPartitioningMethods",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.PMF.Arbitrator.AllocateFixationMethod, Models",
                          "Name": "AllocateFixationMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.PMF.Arbitrator.RetranslocationMethod, Models",
                          "Name": "RetranslocationMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Core.Folder, Models",
                      "ShowInDocs": False,
                      "GraphsPerPage": 6,
                      "Name": "AllocationMethods",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.PMF.Arbitrator.NitrogenAllocationsMethod, Models",
                          "Name": "NitrogenAllocationsMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.RelativeAllocation, Models",
                      "Name": "ArbitrationMethod",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Arbitrator.AllocateUptakesMethod, Models",
                      "Name": "AllocateUptakesMethod",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Arbitrator.WaterUptakeMethod, Models",
                  "Name": "WaterUptakeMethod",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Arbitrator.NitrogenUptakeMethod, Models",
                  "Name": "NitrogenUptakeMethod",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.Phen.Phenology, Models",
              "Name": "Phenology",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Functions.MultiplyFunction, Models",
                  "Name": "ThermalTime",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.HourlyInterpolation, Models",
                      "agregationMethod": 0,
                      "Name": "BaseThermalTime",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            18.0,
                            26.0,
                            34.0,
                            44.0
                          ],
                          "Y": [
                            0.0,
                            10.0,
                            18.0,
                            26.0,
                            0.0
                          ],
                          "XVariableName": None,
                          "Name": "Response",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.ThreeHourSin, Models",
                          "TempRangeFactors": None,
                          "Name": "InterpolationMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.GerminatingPhase, Models",
                  "Start": "Sowing",
                  "End": "Germination",
                  "Name": "Germinating",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.EmergingPhase, Models",
                  "Start": "Germination",
                  "End": "Emergence",
                  "TTForTimeStep": 0.0,
                  "Name": "Emerging",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "ThermalTime",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.AddFunction, Models",
                      "Name": "Target",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 55.0,
                          "Units": None,
                          "Name": "ShootLag",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.MultiplyFunction, Models",
                          "Name": "DepthxRate",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Plant].SowingData.Depth",
                              "Name": "SowingDepth",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 0.6,
                              "Units": None,
                              "Name": "ShootRate",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.GenericPhase, Models",
                  "Start": "Emergence",
                  "End": "EndJuvenile",
                  "Name": "Juvenile",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 200.0,
                      "Units": "deg day",
                      "Name": "Target",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "Progression",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.GenericPhase, Models",
                  "Start": "EndJuvenile",
                  "End": "FloralInitiation",
                  "Name": "Photosensitive",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "Progression",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "Target",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Memo, Models",
                          "Text": "\nThe photoperiod is likely to vary with individual maize cultivars.  Therefore, the response shown here is indicative only.  See the descriptions of individual cultivars for more detail on photoperiod responses used by the maize model.\n",
                          "Name": "memo",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            12.5,
                            24.0
                          ],
                          "Y": [
                            0.0,
                            0.0,
                            10.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Phenology].Photoperiod",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.LeafAppearancePhase, Models",
                  "Start": "FloralInitiation",
                  "End": "FlagLeaf",
                  "Name": "LeafAppearance",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "ThermalTime",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Structure].FinalLeafNumber",
                      "Name": "FinalLeafNumber",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.ExpressionFunction, Models",
                      "Expression": "[Leaf].ExpandedCohortNo + [Leaf].NextExpandingLeafProportion",
                      "Name": "LeafNumber",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Leaf].ExpandedCohortNo",
                      "Name": "FullyExpandedLeafNo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Leaf].InitialisedCohortNo",
                      "Name": "InitialisedLeafNumber",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.GenericPhase, Models",
                  "Start": "FlagLeaf",
                  "End": "Flowering",
                  "Name": "FlagLeafToFlowering",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 50.0,
                      "Units": "deg day",
                      "Name": "Target",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "Progression",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.GenericPhase, Models",
                  "Start": "Flowering",
                  "End": "StartGrainFill",
                  "Name": "FloweringToGrainFilling",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 120.0,
                      "Units": "deg day",
                      "Name": "Target",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "Progression",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.GenericPhase, Models",
                  "Start": "StartGrainFill",
                  "End": "EndGrainFill",
                  "Name": "GrainFilling",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 550.0,
                      "Units": "deg day",
                      "Name": "Target",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "Progression",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.GenericPhase, Models",
                  "Start": "EndGrainFill",
                  "End": "Maturity",
                  "Name": "Maturing",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 10.0,
                      "Units": "deg day",
                      "Name": "Target",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "Progression",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.GenericPhase, Models",
                  "Start": "Maturity",
                  "End": "HarvestRipe",
                  "Name": "MaturityToHarvestRipe",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 10.0,
                      "Units": "deg day",
                      "Name": "Target",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "Progression",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.EndPhase, Models",
                  "Start": "HarvestRipe",
                  "End": "Unused",
                  "Name": "ReadyForHarvesting",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "ThermalTime",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhotoperiodFunction, Models",
                  "Twilight": -6.0,
                  "DayLength": 0.0,
                  "Name": "Photoperiod",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.OnEventFunction, Models",
                  "SetEvent": "Flowering",
                  "ReSetEvent": "never",
                  "Name": "FloweringDAS",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nA function is used to provide flowering date as days after sowing(DAS).\n",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "PreEventValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Plant].DaysAfterSowing",
                      "Name": "PostEventValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.OnEventFunction, Models",
                  "SetEvent": "Maturity",
                  "ReSetEvent": "never",
                  "Name": "MaturityDAS",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nA function is used to provide maturity date as days after sowing(DAS).",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "PreEventValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Plant].DaysAfterSowing",
                      "Name": "PostEventValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.OnEventFunction, Models",
                  "SetEvent": "Emergence",
                  "ReSetEvent": "never",
                  "Name": "EmergenceDAS",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nA function is used to provide maturity date as days after sowing(DAS).",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "PreEventValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Plant].DaysAfterSowing",
                      "Name": "PostEventValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.OnEventFunction, Models",
                  "SetEvent": "HarvestRipe",
                  "ReSetEvent": "never",
                  "Name": "HarvestDAS",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nA function is used to provide maturity date as days after sowing(DAS).",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "PreEventValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Plant].DaysAfterSowing",
                      "Name": "PostEventValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.BBCH, Models",
                  "Name": "BBCH",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.Struct.Structure, Models",
              "CohortInitialisationStage": "Germination",
              "LeafInitialisationStage": "Emergence",
              "Name": "Structure",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Memo, Models",
                  "Text": "\nThough tillering is common in some maize varieties under certain agronomic conditions, no tillering is accounted for within this model.  Therefore all branching has been parameterised out of the current maize model.\n",
                  "Name": "memo",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "StemSenescenceAge",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.VariableReference, Models",
                  "VariableName": "[Phenology].ThermalTime",
                  "Name": "ThermalTime",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhaseLookup, Models",
                  "Name": "Phyllochron",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.PhaseLookupValue, Models",
                      "Start": "Emergence",
                      "End": "FlagLeaf",
                      "Name": "Phyllochron",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.LinearInterpolationFunction, Models",
                          "Name": "Phyllochron",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.XYPairs, Models",
                              "X": [
                                1.0,
                                3.4,
                                4.0
                              ],
                              "Y": [
                                30.0,
                                30.0,
                                42.0
                              ],
                              "XVariableName": None,
                              "Name": "XYPairs",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Structure].LeafTipsAppeared",
                              "Name": "XValue",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Struct.HeightFunction, Models",
                  "Name": "HeightModel",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "PotentialHeight",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            3.0,
                            4.0,
                            5.0,
                            6.0
                          ],
                          "Y": [
                            20.0,
                            400.0,
                            450.0,
                            1700.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Phenology].Stage",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "WaterStress",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            0.6
                          ],
                          "Y": [
                            0.0,
                            1.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Maize].Leaf.Fw",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "BranchingRate",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.AddFunction, Models",
                  "Name": "BranchMortality",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "ShadeInducedBranchMortality",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "DroughtInducedBranchMortality",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 15.0,
                  "Units": None,
                  "Name": "zFinalLeafNumber",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.HoldFunction, Models",
                  "WhenToHold": "FloralInitiation",
                  "Name": "FinalLeafNumber",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.AddFunction, Models",
                      "Name": "ValueToHold",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 6.0,
                          "Units": None,
                          "Name": "PrimordiaAtEmergence",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.DivideFunction, Models",
                          "Name": "PrimordiaDuringJuvenilePhase",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.AddFunction, Models",
                              "Name": "AddFunction",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Phenology].Juvenile.Target",
                                  "Name": "JuvenilePhaseDuration",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Phenology].Photosensitive.Target",
                                  "Name": "PhotoSensitivePhaseDuration",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.PhaseLookup, Models",
                              "Name": "MeanPlastochron",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.PhaseLookupValue, Models",
                                  "Start": "Germination",
                                  "End": "Emergence",
                                  "Name": "PreEmerg",
                                  "ResourceName": None,
                                  "Children": [
                                    {
                                      "$child": "Models.Functions.Constant, Models",
                                      "FixedValue": 15.0,
                                      "Units": None,
                                      "Name": "Constant",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    }
                                  ],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.PhaseLookupValue, Models",
                                  "Start": "Emergence",
                                  "End": "FlagLeaf",
                                  "Name": "PostEmerg",
                                  "ResourceName": None,
                                  "Children": [
                                    {
                                      "$child": "Models.Functions.DivideFunction, Models",
                                      "Name": "MeanPlastochron",
                                      "ResourceName": None,
                                      "Children": [
                                        {
                                          "$child": "Models.Functions.AccumulateFunction, Models",
                                          "StartStageName": "Emergence",
                                          "EndStageName": "FloralInitiation",
                                          "ResetStageName": None,
                                          "FractionRemovedOnCut": 0.0,
                                          "FractionRemovedOnHarvest": 0.0,
                                          "FractionRemovedOnGraze": 0.0,
                                          "FractionRemovedOnPrune": 0.0,
                                          "Name": "AccumulatedPlastochron",
                                          "ResourceName": None,
                                          "Children": [
                                            {
                                              "$child": "Models.Functions.MultiplyFunction, Models",
                                              "Name": "Plastochron",
                                              "ResourceName": None,
                                              "Children": [
                                                {
                                                  "$child": "Models.Functions.VariableReference, Models",
                                                  "VariableName": "[Structure].Phyllochron",
                                                  "Name": "Phyllochron",
                                                  "ResourceName": None,
                                                  "Children": [],
                                                  "Enabled": True,
                                                  "ReadOnly": False
                                                },
                                                {
                                                  "$child": "Models.Functions.Constant, Models",
                                                  "FixedValue": 0.5,
                                                  "Units": None,
                                                  "Name": "LeafTipsPerPrimordia",
                                                  "ResourceName": None,
                                                  "Children": [],
                                                  "Enabled": True,
                                                  "ReadOnly": False
                                                }
                                              ],
                                              "Enabled": True,
                                              "ReadOnly": False
                                            }
                                          ],
                                          "Enabled": True,
                                          "ReadOnly": False
                                        },
                                        {
                                          "$child": "Models.Functions.AccumulateFunction, Models",
                                          "StartStageName": "Emergence",
                                          "EndStageName": "FloralInitiation",
                                          "ResetStageName": None,
                                          "FractionRemovedOnCut": 0.0,
                                          "FractionRemovedOnHarvest": 0.0,
                                          "FractionRemovedOnGraze": 0.0,
                                          "FractionRemovedOnPrune": 0.0,
                                          "Name": "Days",
                                          "ResourceName": None,
                                          "Children": [
                                            {
                                              "$child": "Models.Functions.Constant, Models",
                                              "FixedValue": 1.0,
                                              "Units": None,
                                              "Name": "Constant",
                                              "ResourceName": None,
                                              "Children": [],
                                              "Enabled": True,
                                              "ReadOnly": False
                                            }
                                          ],
                                          "Enabled": True,
                                          "ReadOnly": False
                                        }
                                      ],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    }
                                  ],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.Organs.ReproductiveOrgan, Models",
              "GrowthRespiration": 0.0,
              "MaintenanceRespiration": 0.0,
              "DMDemand": None,
              "NDemand": None,
              "DMSupply": None,
              "NSupply": None,
              "RipeStage": None,
              "potentialDMAllocation": None,
              "Live": {
                "$child": "Models.PMF.Biomass, Models",
                "Name": "Biomass",
                "ResourceName": None,
                "Children": [],
                "Enabled": True,
                "ReadOnly": False
              },
              "Dead": {
                "$child": "Models.PMF.Biomass, Models",
                "Name": "Biomass",
                "ResourceName": None,
                "Children": [],
                "Enabled": True,
                "ReadOnly": False
              },
              "Name": "Grain",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": None,
                  "Name": "DMConversionEfficiency",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "RemobilisationCost",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.05,
                  "Units": None,
                  "Name": "InitialGrainProportion",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.3,
                  "Units": "g",
                  "Name": "MaximumPotentialGrainSize",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 700.0,
                  "Units": "number",
                  "Name": "MaximumGrainsPerCob",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhaseLookup, Models",
                  "Name": "DMDemandFunction",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.PhaseLookupValue, Models",
                      "Start": "Flowering",
                      "End": "StartGrainFill",
                      "Name": "InitialPhase",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.DemandFunctions.FillingRateFunction, Models",
                          "Name": "FillingRateFunction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Grain].NumberFunction",
                              "Name": "NumberFunction",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Phenology].FloweringToGrainFilling.Target",
                              "Name": "FillingDuration",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Phenology].ThermalTime",
                              "Name": "ThermalTime",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.MultiplyFunction, Models",
                              "Name": "PotentialSizeIncrement",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Grain].InitialGrainProportion",
                                  "Name": "InitialProportion",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Grain].MaximumPotentialGrainSize",
                                  "Name": "MaximumSize",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.PhaseLookupValue, Models",
                      "Start": "StartGrainFill",
                      "End": "EndGrainFill",
                      "Name": "LinearPhase",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.DemandFunctions.FillingRateFunction, Models",
                          "Name": "FillingRateFunction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Grain].NumberFunction",
                              "Name": "NumberFunction",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Phenology].GrainFilling.Target",
                              "Name": "FillingDuration",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Phenology].ThermalTime",
                              "Name": "ThermalTime",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.MultiplyFunction, Models",
                              "Name": "PotentialSizeIncrement",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.SubtractFunction, Models",
                                  "Name": "ProportionLinearPhase",
                                  "ResourceName": None,
                                  "Children": [
                                    {
                                      "$child": "Models.Functions.Constant, Models",
                                      "FixedValue": 1.0,
                                      "Units": None,
                                      "Name": "One",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    },
                                    {
                                      "$child": "Models.Functions.VariableReference, Models",
                                      "VariableName": "[Grain].InitialGrainProportion",
                                      "Name": "InitialProportion",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    }
                                  ],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Grain].MaximumPotentialGrainSize",
                                  "Name": "MaximumSize",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.008,
                  "Units": "g/g",
                  "Name": "MinimumNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhaseLookup, Models",
                  "Name": "MaximumNConc",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.PhaseLookupValue, Models",
                      "Start": "Flowering",
                      "End": "StartGrainFill",
                      "Name": "InitialPhase",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 0.05,
                          "Units": None,
                          "Name": "InitialNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.PhaseLookupValue, Models",
                      "Start": "StartGrainFill",
                      "End": "EndGrainFill",
                      "Name": "LinearPhase",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 0.013,
                          "Units": None,
                          "Name": "FinalNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.12,
                  "Units": "g/g",
                  "Name": "WaterContent",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhaseLookup, Models",
                  "Name": "NFillingRate",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.PhaseLookupValue, Models",
                      "Start": "Flowering",
                      "End": "EndGrainFill",
                      "Name": "GrainFilling",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.DemandFunctions.FillingRateFunction, Models",
                          "Name": "FillingRateFunction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Grain].NumberFunction",
                              "Name": "NumberFunction",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.AddFunction, Models",
                              "Name": "FillingDuration",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Phenology].FloweringToGrainFilling.Target",
                                  "Name": "FloweringToGrainFilling",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Phenology].GrainFilling.Target",
                                  "Name": "GrainFilling",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Phenology].ThermalTime",
                              "Name": "ThermalTime",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.MultiplyFunction, Models",
                              "Name": "PotentialSizeIncrement",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Grain].MaximumNConc.LinearPhase.FinalNconc",
                                  "Name": "MaximumFinalNConc",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Grain].MaximumPotentialGrainSize",
                                  "Name": "MaximumSize",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.MultiplyFunction, Models",
                  "Name": "NumberFunction",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[MaximumGrainsPerCob]",
                      "Name": "PotentialGrainNumber",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "GrowthRateFactor",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.55,
                            3.3,
                            6.6,
                            10.0
                          ],
                          "Y": [
                            0.0,
                            0.6,
                            1.0,
                            1.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[GrowthRateGrainDevelopment]",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Maize].Population",
                      "Name": "Population",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.ExpressionFunction, Models",
                  "Expression": "[Grain].Live.Wt * 10 * 56 * 0.4536 * 2.471",
                  "Name": "YieldBuPerAcre",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.AccumulateFunction, Models",
                  "StartStageName": "FlagLeaf",
                  "EndStageName": "StartGrainFill",
                  "ResetStageName": None,
                  "FractionRemovedOnCut": 0.0,
                  "FractionRemovedOnHarvest": 0.0,
                  "FractionRemovedOnGraze": 0.0,
                  "FractionRemovedOnPrune": 0.0,
                  "Name": "GrowthGrainDevelopment",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Arbitrator].DM.TotalFixationSupply",
                      "Name": "DailyBiomassProduction",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.AccumulateFunction, Models",
                  "StartStageName": "FlagLeaf",
                  "EndStageName": "StartGrainFill",
                  "ResetStageName": None,
                  "FractionRemovedOnCut": 0.0,
                  "FractionRemovedOnHarvest": 0.0,
                  "FractionRemovedOnGraze": 0.0,
                  "FractionRemovedOnPrune": 0.0,
                  "Name": "DaysGrainDevelopment",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": "days/day",
                      "Name": "LenthOfADay",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.DivideFunction, Models",
                  "Name": "GrowthRateGrainDevelopment",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[GrowthGrainDevelopment]",
                      "Name": "AccumulatedGrowth",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[DaysGrainDevelopment]",
                      "Name": "GrowthDuration",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Maize].Population",
                      "Name": "Population",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Library.BiomassRemoval, Models",
                  "Name": "BiomassRemovalDefaults",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 1.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.0,
                      "FractionDeadToResidue": 1.0,
                      "Name": "Harvest",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 1.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.0,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Cut",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.8,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Prune",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.6,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.2,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Graze",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.4,
                  "Units": None,
                  "Name": "CarbonConcentration",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientPoolFunctions, Models",
                  "Name": "DMDemandPriorityFactors",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientPoolFunctions, Models",
                  "Name": "NDemandPriorityFactors",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.Organs.Root, Models",
              "DMSupply": None,
              "NSupply": None,
              "DMDemand": None,
              "NDemand": None,
              "potentialDMAllocation": None,
              "GrowthRespiration": 0.0,
              "MaintenanceRespiration": 0.0,
              "Name": "Root",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Functions.RootShape.RootShapeCylinder, Models",
                  "Name": "RootShape",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                  "Name": "KLModifier",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nThe parameter KL is is described by Meinke et al (1993) to combine elements of soil conductivity (K) with root length (L) in determining root water uptake.  Peake et al (2013) demonstrated that in maize KL varied with plant population through impacts on root length density.  Here we adjust KL linearly with plant population.\n",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.XYPairs, Models",
                      "X": [
                        0.0,
                        10.0
                      ],
                      "Y": [
                        1.0,
                        1.0
                      ],
                      "XVariableName": None,
                      "Name": "XYPairs",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Maize].Population",
                      "Name": "XValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": None,
                  "Name": "DMConversionEfficiency",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "RemobilisationCost",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": None,
                  "Name": "SoilWaterEffect",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nNo effect of soil water content on root growth is currently captured in the model.\n",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 20.0,
                  "Units": "g/m2/d",
                  "Name": "MaxDailyNUptake",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nMaximum daily N uptake is set to an arbitrarily high value.\n",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.005,
                  "Units": "/d",
                  "Name": "SenescenceRate",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nJones and Kiniry (1986) used a constant root senescence fraction in the CERES-Maize model.\n",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1000000.0,
                  "Units": None,
                  "Name": "MaximumRootDepth",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.01,
                  "Units": "g/g",
                  "Name": "MaximumNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.01,
                  "Units": "g/g",
                  "Name": "MinimumNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.MultiplyFunction, Models",
                  "Name": "RootFrontVelocity",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.PhaseLookup, Models",
                      "Name": "PotentialRootFrontVelocity",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Memo, Models",
                          "Text": "\n[Hammer_etal_2009] used a constant extraction front velocity of 2.5cm per day in their analysis of historical maize yield trends in the US corn belt.\n",
                          "Name": "memo",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.PhaseLookupValue, Models",
                          "Start": "Germination",
                          "End": "Flowering",
                          "Name": "PreFlowering",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 25.0,
                              "Units": "mm/d",
                              "Name": "RootFrontVelocity",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.PhaseLookupValue, Models",
                          "Start": "Flowering",
                          "End": "Maturity",
                          "Name": "PostFlowering",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 0.0,
                              "Units": "mm/d",
                              "Name": "RootFrontVelocity",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.WeightedTemperatureFunction, Models",
                      "MaximumTemperatureWeighting": 0.5,
                      "Name": "TemperatureFactor",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            8.0,
                            15.0,
                            35.0,
                            50.0
                          ],
                          "Y": [
                            0.0,
                            0.0,
                            1.0,
                            1.0,
                            0.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "WaterFactor",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SoilWaterScale, Models",
                          "Name": "SoilWaterScale",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            0.25,
                            1.0
                          ],
                          "Y": [
                            0.0,
                            1.0,
                            1.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[SoilWaterScale]",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "AFPSFactor",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            0.03,
                            1.0
                          ],
                          "Y": [
                            1.0,
                            1.0,
                            1.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Root].AirFilledPoreSpace",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "RootDepthStressFactor",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhaseLookupValue, Models",
                  "Start": "Germination",
                  "End": "Maturity",
                  "Name": "NitrogenDemandSwitch",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Constant",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Library.BiomassRemoval, Models",
                  "Name": "BiomassRemovalDefaults",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 1.0,
                      "FractionDeadToResidue": 1.0,
                      "Name": "Harvest",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.3,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Cut",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 1.0,
                      "FractionDeadToResidue": 1.0,
                      "Name": "Prune",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 1.0,
                      "FractionDeadToResidue": 1.0,
                      "Name": "Graze",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 40.0,
                  "Units": "m/g",
                  "Name": "SpecificRootLength",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.03,
                  "Units": None,
                  "Name": "KNO3",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.02,
                  "Units": None,
                  "Name": "KNH4",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                  "Name": "NUptakeSWFactor",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.XYPairs, Models",
                      "X": [
                        0.0,
                        1.0
                      ],
                      "Y": [
                        0.0,
                        1.0
                      ],
                      "XVariableName": None,
                      "Name": "XYPairs",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Root].RWC",
                      "Name": "XValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "MaintenanceRespirationFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.4,
                  "Units": None,
                  "Name": "CarbonConcentration",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientDemandFunctions, Models",
                  "Name": "DMDemands",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.DemandFunctions.PartitionFractionDemandFunction, Models",
                          "Name": "DMDemandFunction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.LinearInterpolationFunction, Models",
                              "Name": "PartitionFraction",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.XYPairs, Models",
                                  "X": [
                                    0.0,
                                    2.99,
                                    3.0,
                                    6.0,
                                    7.0
                                  ],
                                  "Y": [
                                    0.0,
                                    0.0,
                                    0.2,
                                    0.2,
                                    0.0
                                  ],
                                  "XVariableName": None,
                                  "Name": "XYPairs",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Phenology].Stage",
                                  "Name": "XValue",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 1.0,
                          "Units": None,
                          "Name": "StructuralFraction",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.DemandFunctions.StorageDMDemandFunction, Models",
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "StorageFraction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 1.0,
                              "Units": None,
                              "Name": "One",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Root].DMDemands.Structural.StructuralFraction",
                              "Name": "StructuralFraction",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStructuralPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QMetabolicPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStoragePriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientDemandFunctions, Models",
                  "Name": "NDemands",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Root].minimumNconc",
                          "Name": "MinNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Root].potentialDMAllocation.Structural",
                          "Name": "PotentialDMAllocation",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "MetabolicNconc",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Root].criticalNConc",
                              "Name": "CritNconc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Root].minimumNconc",
                              "Name": "MinNconc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Root].potentialDMAllocation.Structural",
                          "Name": "PotentialDMAllocation",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.DemandFunctions.StorageNDemandFunction, Models",
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Root].nitrogenDemandSwitch",
                          "Name": "NitrogenDemandSwitch",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Root].maximumNconc",
                          "Name": "MaxNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStructuralPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QMetabolicPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStoragePriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.VariableReference, Models",
                  "VariableName": "[Root].MinimumNConc",
                  "Name": "CriticalNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientPoolFunctions, Models",
                  "Name": "InitialWt",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.2,
                      "Units": "g/plant",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.Organs.Leaf, Models",
              "Leaves": [],
              "CurrentExpandingLeaf": 0.0,
              "StartFractionExpanded": 0.0,
              "FractionNextleafExpanded": 0.0,
              "DeadNodesYesterday": 0.0,
              "MaxCover": 1.0,
              "GrowthRespiration": 0.0,
              "DMSupply": None,
              "NSupply": None,
              "DMDemand": None,
              "NDemand": None,
              "potentialDMAllocation": None,
              "Albedo": 0.18,
              "Gsmax350": 0.009,
              "R50": 160.0,
              "LAI": 0.0,
              "Depth": 0.0,
              "Width": 0.0,
              "FRGR": 0.0,
              "PotentialEP": 0.0,
              "WaterDemand": 0.0,
              "LightProfile": None,
              "KDead": 0.03,
              "MaximumMainStemLeafNumber": 30,
              "TipsAtEmergence": 0,
              "CohortsAtInitialisation": 0,
              "FractionDied": 0.0,
              "Name": "Leaf",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Functions.ExpressionFunction, Models",
                  "Expression": "(163 - [IWeather].MeanT)/(5 - 0.1*[IWeather].MeanT)",
                  "Name": "CO2internal",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.DivideFunction, Models",
                  "Name": "StomatalConductanceCO2Modifier",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Leaf].Photosynthesis.FCO2",
                      "Name": "FCO2",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.ExpressionFunction, Models",
                      "Expression": "([IWeather].CO2 - [Leaf].CO2internal)/(350 - [Leaf].CO2internal)",
                      "Name": "RelativeCO2Gradient",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Organs.LeafCohort, Models",
                  "ApexCohort": None,
                  "LiveStart": None,
                  "NReallocationFactor": 0.0,
                  "DMReallocationFactor": 0.0,
                  "NRetranslocationFactor": 0.0,
                  "DMRetranslocationFactor": 0.0,
                  "ShadeInducedSenRate": 0.0,
                  "SenescedFrac": 0.0,
                  "DeltaPotentialArea": 0.0,
                  "DeltaStressConstrainedArea": 0.0,
                  "DeltaCarbonConstrainedArea": 0.0,
                  "PotentialStructuralDMAllocation": 0.0,
                  "PotentialMetabolicDMAllocation": 0.0,
                  "MetabolicNReallocated": 0.0,
                  "MetabolicWtReallocated": 0.0,
                  "StorageNReallocated": 0.0,
                  "StorageWtReallocated": 0.0,
                  "MetabolicNRetranslocated": 0.0,
                  "StorageNRetrasnlocated": 0.0,
                  "DMRetranslocated": 0.0,
                  "MetabolicNAllocation": 0.0,
                  "StructuralDMAllocation": 0.0,
                  "MetabolicDMAllocation": 0.0,
                  "Rank": 1,
                  "Area": 380.0,
                  "Name": "InitialLeaves[1]",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Organs.LeafCohort, Models",
                  "ApexCohort": None,
                  "LiveStart": None,
                  "NReallocationFactor": 0.0,
                  "DMReallocationFactor": 0.0,
                  "NRetranslocationFactor": 0.0,
                  "DMRetranslocationFactor": 0.0,
                  "ShadeInducedSenRate": 0.0,
                  "SenescedFrac": 0.0,
                  "DeltaPotentialArea": 0.0,
                  "DeltaStressConstrainedArea": 0.0,
                  "DeltaCarbonConstrainedArea": 0.0,
                  "PotentialStructuralDMAllocation": 0.0,
                  "PotentialMetabolicDMAllocation": 0.0,
                  "MetabolicNReallocated": 0.0,
                  "MetabolicWtReallocated": 0.0,
                  "StorageNReallocated": 0.0,
                  "StorageWtReallocated": 0.0,
                  "MetabolicNRetranslocated": 0.0,
                  "StorageNRetrasnlocated": 0.0,
                  "DMRetranslocated": 0.0,
                  "MetabolicNAllocation": 0.0,
                  "StructuralDMAllocation": 0.0,
                  "MetabolicDMAllocation": 0.0,
                  "Rank": 2,
                  "Area": 0.0,
                  "Name": "InitialLeaves[2]",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Organs.Leaf+LeafCohortParameters, Models",
                  "ExpansionStressValue": 0.0,
                  "CellDivisionStressValue": 0.0,
                  "LagAccelerationValue": 0.0,
                  "SenescenceAccelerationValue": 0.0,
                  "ShadeInducedSenescenceRateValue": 0.0,
                  "SenessingLeafRelativeSizeValue": 0.0,
                  "Name": "CohortParameters",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "DMReallocationFactor",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "RemobilisationCost",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.01,
                      "Units": None,
                      "Name": "LeafSizeShapeParameter",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "MaintenanceRespirationFunction",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": "0-1",
                      "Name": "SenessingLeafRelativeSize",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.BellCurveFunction, Models",
                      "Name": "MaxArea",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Memo, Models",
                          "Text": "\nPotential leaf area is calculated from leaf number using the approach of Keating and Wafula (1992).\nIndividual leaf size is described using a bell-shaped curve, the parameters of which are derived from final leaf number.\n",
                          "Name": "Memo",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.MultiplyFunction, Models",
                          "Name": "LargestLeafPosition",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Structure].FinalLeafNumber",
                              "Name": "FinalLeafNumber",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 0.67,
                              "Units": None,
                              "Name": "RelativePositionOfLargestLeaf",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.LinearInterpolationFunction, Models",
                          "Name": "AreaMax",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.XYPairs, Models",
                              "X": [
                                0.0,
                                2.0,
                                7.0,
                                12.0,
                                17.0,
                                22.0
                              ],
                              "Y": [
                                32530.0,
                                39259.0,
                                49659.0,
                                60000.0,
                                75000.0,
                                90000.0
                              ],
                              "XVariableName": None,
                              "Name": "XYPairs",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Structure].FinalLeafNumber",
                              "Name": "XValue",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.LinearInterpolationFunction, Models",
                          "Name": "Breadth",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.XYPairs, Models",
                              "X": [
                                5.0,
                                10.0,
                                15.0,
                                20.0,
                                25.0
                              ],
                              "Y": [
                                -0.377,
                                -0.144,
                                -0.059,
                                -0.027,
                                -0.016
                              ],
                              "XVariableName": None,
                              "Name": "XYPairs",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Structure].FinalLeafNumber",
                              "Name": "XValue",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 0.00025,
                          "Units": None,
                          "Name": "Skewness",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "NReallocationFactor",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "NRetranslocationFactor",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.5,
                      "Units": None,
                      "Name": "DMRetranslocationFactor",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "CellDivisionStress",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.9,
                            1.0
                          ],
                          "Y": [
                            1.0,
                            1.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Maize].Leaf.Fw",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.MinimumFunction, Models",
                      "Name": "ExpansionStress",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.LinearInterpolationFunction, Models",
                          "Name": "WaterStressEffect",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.XYPairs, Models",
                              "X": [
                                0.0,
                                0.9,
                                1.3
                              ],
                              "Y": [
                                0.0,
                                1.0,
                                1.0
                              ],
                              "XVariableName": None,
                              "Name": "XYPairs",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Root].WaterTensionFactor",
                              "Name": "XValue",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.LinearInterpolationFunction, Models",
                          "Name": "TemperatureEffect",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.XYPairs, Models",
                              "X": [
                                0.0,
                                12.0,
                                14.0
                              ],
                              "Y": [
                                0.0,
                                1.0,
                                1.0
                              ],
                              "XVariableName": None,
                              "Name": "XYPairs",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[IWeather].MeanT",
                              "Name": "XValue",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.LinearInterpolationFunction, Models",
                          "Name": "NitrogenStressEffect",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.XYPairs, Models",
                              "X": [
                                0.0,
                                0.5,
                                1.0
                              ],
                              "Y": [
                                0.1,
                                0.1,
                                1.0
                              ],
                              "XVariableName": None,
                              "Name": "XYPairs",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Maize].Leaf.Fn",
                              "Name": "XValue",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "GrowthDuration",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            0.3,
                            0.5,
                            1.0
                          ],
                          "Y": [
                            1.0,
                            200.0,
                            250.0,
                            100.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Structure].RelativeNodeApperance",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "LagDuration",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            0.5,
                            0.7,
                            0.8,
                            1.0
                          ],
                          "Y": [
                            300.0,
                            300.0,
                            700.0,
                            900.0,
                            1300.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Structure].RelativeNodeApperance",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "SenescenceDuration",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            0.3,
                            0.5
                          ],
                          "Y": [
                            150.0,
                            150.0,
                            350.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Structure].RelativeNodeApperance",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1000000.0,
                      "Units": None,
                      "Name": "DetachmentLagDuration",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1000000.0,
                      "Units": None,
                      "Name": "DetachmentDuration",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.AddFunction, Models",
                      "Name": "MaximumNConc",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].CohortParameters.CriticalNConc",
                          "Name": "CriticalNConc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.MultiplyFunction, Models",
                          "Name": "StorageNConc",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.LinearInterpolationFunction, Models",
                              "Name": "NonStructuralNConc",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.XYPairs, Models",
                                  "X": [
                                    1.0,
                                    10.0
                                  ],
                                  "Y": [
                                    0.034,
                                    0.01
                                  ],
                                  "XVariableName": None,
                                  "Name": "XYPairs",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Structure].LeafTipsAppeared",
                                  "Name": "XValue",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.LinearInterpolationFunction, Models",
                              "Name": "WaterStressEffect",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.XYPairs, Models",
                                  "X": [
                                    0.5,
                                    0.95,
                                    1.0
                                  ],
                                  "Y": [
                                    0.0,
                                    0.0,
                                    1.0
                                  ],
                                  "XVariableName": None,
                                  "Name": "XYPairs",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Root].WaterTensionFactor",
                                  "Name": "XValue",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.01,
                      "Units": "g/g",
                      "Name": "CriticalNConc",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.005,
                      "Units": "g/g",
                      "Name": "MinimumNConc",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.AddFunction, Models",
                      "Name": "LagAcceleration",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 1.0,
                          "Units": None,
                          "Name": "One",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.MultiplyFunction, Models",
                          "Name": "Stress",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 2.0,
                              "Units": None,
                              "Name": "StressResponseCoefficient",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.LinearInterpolationFunction, Models",
                              "Name": "StressFactor",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.XYPairs, Models",
                                  "X": [
                                    0.5,
                                    1.0
                                  ],
                                  "Y": [
                                    1.0,
                                    0.0
                                  ],
                                  "XVariableName": None,
                                  "Name": "XYPairs",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Maize].Leaf.Fw",
                                  "Name": "XValue",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.AddFunction, Models",
                      "Name": "SenescenceAcceleration",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 1.0,
                          "Units": None,
                          "Name": "One",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.MultiplyFunction, Models",
                          "Name": "Stress",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 0.0,
                              "Units": None,
                              "Name": "StressResponseCoefficient",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.LinearInterpolationFunction, Models",
                              "Name": "StressFactor",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.XYPairs, Models",
                                  "X": [
                                    0.0,
                                    1.0
                                  ],
                                  "Y": [
                                    1.0,
                                    0.0
                                  ],
                                  "XVariableName": None,
                                  "Name": "XYPairs",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Root].WaterTensionFactor",
                                  "Name": "XValue",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "SpecificLeafAreaMax",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            1.0,
                            9.0
                          ],
                          "Y": [
                            30000.0,
                            20000.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Structure].LeafTipsAppeared",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "SpecificLeafAreaMin",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            1.0,
                            9.0
                          ],
                          "Y": [
                            25000.0,
                            15000.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Structure].LeafTipsAppeared",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.3,
                      "Units": None,
                      "Name": "StructuralFraction",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.1,
                      "Units": None,
                      "Name": "StorageFraction",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "ShadeInducedSenescenceRate",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            0.7,
                            0.8,
                            0.96,
                            0.99
                          ],
                          "Y": [
                            0.0,
                            0.0,
                            0.0,
                            0.0,
                            0.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].CohortCurrentRankCoverAbove",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": "g/g",
                      "Name": "InitialNConc",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.ArrayFunction, Models",
                      "Values": "1 1 1",
                      "Units": None,
                      "Name": "LagDurationAgeMultiplier",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.ArrayFunction, Models",
                      "Values": "1 1 1",
                      "Units": None,
                      "Name": "SenescenceDurationAgeMultiplier",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.ArrayFunction, Models",
                      "Values": "1 1 1 1 1 1 1 1 1 1 1 1",
                      "Units": None,
                      "Name": "LeafSizeAgeMultiplier",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.SupplyFunctions.RUEModel, Models",
                  "Name": "Photosynthesis",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 2.0,
                      "Units": None,
                      "Name": "RUE",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.WeightedTemperatureFunction, Models",
                      "MaximumTemperatureWeighting": 0.75,
                      "Name": "FT",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            8.0,
                            15.0,
                            35.0,
                            50.0
                          ],
                          "Y": [
                            0.0,
                            0.0,
                            1.0,
                            1.0,
                            0.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "FVPD",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            10.0,
                            50.0
                          ],
                          "Y": [
                            1.0,
                            1.0,
                            1.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].Photosynthesis.VPD",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "FN",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            1.0
                          ],
                          "Y": [
                            0.3,
                            1.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].Fn",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "FW",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            0.3
                          ],
                          "Y": [
                            0.0,
                            1.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].Fw",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.SupplyFunctions.RUECO2Function, Models",
                      "PhotosyntheticPathway": "C4",
                      "Name": "FCO2",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Leaf].RadiationIntercepted",
                      "Name": "RadnInt",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.MultiplyFunction, Models",
                  "Name": "ExtinctionCoeff",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nExtinction coefficient is specified in terms of a potential extinction coefficient which is then discounted during periods of water stress to account for the impact of leaf curling.\n",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "PotentialExtinctionCoeff",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.2,
                            0.5,
                            1.0
                          ],
                          "Y": [
                            0.7,
                            0.5,
                            0.4
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Maize].SowingData.RowSpacing",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "WaterStressImpact",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            0.7
                          ],
                          "Y": [
                            0.5,
                            1.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Maize].Leaf.Fw",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                  "Name": "FrostFraction",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.XYPairs, Models",
                      "X": [
                        -2.0,
                        2.0
                      ],
                      "Y": [
                        1.0,
                        0.0
                      ],
                      "XVariableName": None,
                      "Name": "XYPairs",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Weather].MinT",
                      "Name": "XValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.VariableReference, Models",
                  "VariableName": "[Phenology].ThermalTime",
                  "Name": "ThermalTime",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "RemobilisationCost",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.MinimumFunction, Models",
                  "Name": "FRGRFunction",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Leaf].Photosynthesis.FT",
                      "Name": "FT",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Leaf].Photosynthesis.FN",
                      "Name": "FN",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Leaf].Photosynthesis.FVPD",
                      "Name": "FVPD",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.7,
                  "Units": None,
                  "Name": "StructuralFraction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": "0-1",
                  "Name": "DMConversionEfficiency",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Library.BiomassRemoval, Models",
                  "Name": "BiomassRemovalDefaults",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 1.0,
                      "FractionDeadToResidue": 1.0,
                      "Name": "Harvest",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.8,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.0,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Cut",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.6,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Prune",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.6,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.1,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Graze",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.4,
                  "Units": None,
                  "Name": "CarbonConcentration",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "WidthFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.VariableReference, Models",
                  "VariableName": "[Leaf].Height",
                  "Name": "DepthFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientPoolFunctions, Models",
                  "Name": "DMDemandPriorityFactors",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientPoolFunctions, Models",
                  "Name": "NDemandPriorityFactors",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.Organs.GenericOrgan, Models",
              "StartLive": None,
              "DMSupply": None,
              "NSupply": None,
              "DMDemand": None,
              "NDemand": None,
              "potentialDMAllocation": None,
              "IsAboveGround": True,
              "Name": "Husk",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "NReallocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "RemobilisationCost",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.1,
                  "Units": None,
                  "Name": "NRetranslocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.007,
                  "Units": "g/g",
                  "Name": "MaximumNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": "/d",
                  "Name": "SenescenceRate",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.005,
                  "Units": "g/g",
                  "Name": "MinimumNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.1,
                  "Units": None,
                  "Name": "DMRetranslocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Library.BiomassRemoval, Models",
                  "Name": "BiomassRemovalDefaults",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.5,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.1,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Harvest",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.8,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.0,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Cut",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.6,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Prune",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.6,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.2,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Graze",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": "/d",
                  "Name": "DetachmentRateFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": "0-1",
                  "Name": "MaintenanceRespirationFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": "0-1",
                  "Name": "DMConversionEfficiency",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "DMReallocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.VariableReference, Models",
                  "VariableName": "[Husk].MinimumNConc",
                  "Name": "CriticalNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhaseLookupValue, Models",
                  "Start": "Emergence",
                  "End": "StartGrainFill",
                  "Name": "NitrogenDemandSwitch",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Constant",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.4,
                  "Units": None,
                  "Name": "CarbonConcentration",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientDemandFunctions, Models",
                  "Name": "DMDemands",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.DemandFunctions.PopulationBasedDemandFunction, Models",
                          "Name": "DMDemandFunction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Phenology].ThermalTime",
                              "Name": "ThermalTime",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 5.5,
                              "Units": None,
                              "Name": "StartStage",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 300.0,
                              "Units": None,
                              "Name": "GrowthDuration",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 25.0,
                              "Units": None,
                              "Name": "MaximumOrganWt",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Rachis].CobNumber",
                              "Name": "OrganPopulation",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 0.0,
                              "Units": "/d",
                              "Name": "SenescenceRate",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.LinearInterpolationFunction, Models",
                              "Name": "ExpansionStress2",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.XYPairs, Models",
                                  "X": [
                                    0.0,
                                    0.35
                                  ],
                                  "Y": [
                                    0.0,
                                    1.0
                                  ],
                                  "XVariableName": None,
                                  "Name": "XYPairs",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Maize].Leaf.Fw",
                                  "Name": "XValue",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.MinimumFunction, Models",
                              "Name": "ExpansionStress",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                                  "Name": "WaterStressEffect",
                                  "ResourceName": None,
                                  "Children": [
                                    {
                                      "$child": "Models.Functions.XYPairs, Models",
                                      "X": [
                                        0.1,
                                        0.7,
                                        1.3
                                      ],
                                      "Y": [
                                        0.0,
                                        1.0,
                                        1.0
                                      ],
                                      "XVariableName": None,
                                      "Name": "XYPairs",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    },
                                    {
                                      "$child": "Models.Functions.VariableReference, Models",
                                      "VariableName": "[Root].WaterTensionFactor",
                                      "Name": "XValue",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    }
                                  ],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                                  "Name": "TemperatureEffect",
                                  "ResourceName": None,
                                  "Children": [
                                    {
                                      "$child": "Models.Functions.XYPairs, Models",
                                      "X": [
                                        0.0,
                                        12.0,
                                        14.0
                                      ],
                                      "Y": [
                                        0.0,
                                        1.0,
                                        1.0
                                      ],
                                      "XVariableName": None,
                                      "Name": "XYPairs",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    },
                                    {
                                      "$child": "Models.Functions.VariableReference, Models",
                                      "VariableName": "[IWeather].MeanT",
                                      "Name": "XValue",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    }
                                  ],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                                  "Name": "NitrogenStressEffect",
                                  "ResourceName": None,
                                  "Children": [
                                    {
                                      "$child": "Models.Functions.XYPairs, Models",
                                      "X": [
                                        0.0,
                                        0.5,
                                        1.0
                                      ],
                                      "Y": [
                                        0.1,
                                        0.1,
                                        1.0
                                      ],
                                      "XVariableName": None,
                                      "Name": "XYPairs",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    },
                                    {
                                      "$child": "Models.Functions.VariableReference, Models",
                                      "VariableName": "[Maize].Leaf.Fn",
                                      "Name": "XValue",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    }
                                  ],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.PhaseLookup, Models",
                          "Name": "StructuralFraction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.PhaseLookupValue, Models",
                              "Start": "Emergence",
                              "End": "StartGrainFill",
                              "Name": "PreStemElongation",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.Constant, Models",
                                  "FixedValue": 0.5,
                                  "Units": None,
                                  "Name": "StemFraction",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.PhaseLookupValue, Models",
                              "Start": "StartGrainFill",
                              "End": "HarvestRipe",
                              "Name": "StemElongation",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.Constant, Models",
                                  "FixedValue": 1.0,
                                  "Units": None,
                                  "Name": "StemFraction",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.DemandFunctions.StorageDMDemandFunction, Models",
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "StorageFraction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 1.0,
                              "Units": None,
                              "Name": "One",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Husk].DMDemands.Structural.StructuralFraction",
                              "Name": "StructuralFraction",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStructuralPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QMetabolicPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStoragePriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientDemandFunctions, Models",
                  "Name": "NDemands",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Husk].minimumNconc",
                          "Name": "MinNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Husk].potentialDMAllocation.Structural",
                          "Name": "PotentialDMAllocation",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "MetabolicNconc",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Husk].criticalNConc",
                              "Name": "CritNconc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Husk].minimumNconc",
                              "Name": "MinNconc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Husk].potentialDMAllocation.Structural",
                          "Name": "PotentialDMAllocation",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.DemandFunctions.StorageNDemandFunction, Models",
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Husk].nitrogenDemandSwitch",
                          "Name": "NitrogenDemandSwitch",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Husk].maximumNconc",
                          "Name": "MaxNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStructuralPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QMetabolicPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStoragePriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.RetranslocateNonStructural, Models",
                  "Name": "RetranslocateNitrogen",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientPoolFunctions, Models",
                  "Name": "InitialWt",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": "g/m^2",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "Photosynthesis",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.VariableReference, Models",
                  "VariableName": "[Husk].MinimumNConc",
                  "Name": "initialNConcFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.Organs.GenericOrgan, Models",
              "StartLive": None,
              "DMSupply": None,
              "NSupply": None,
              "DMDemand": None,
              "NDemand": None,
              "potentialDMAllocation": None,
              "IsAboveGround": True,
              "Name": "Rachis",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Functions.MultiplyFunction, Models",
                  "Name": "CobNumber",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Structure].TotalStemPopn",
                      "Name": "StemNumber",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.1,
                      "Units": None,
                      "Name": "CobsPerStem",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.DivideFunction, Models",
                  "Name": "CobSize",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Rachis].Live.Wt",
                      "Name": "CobMass",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Rachis].CobNumber",
                      "Name": "CobNumber",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "NReallocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "RemobilisationCost",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.1,
                  "Units": None,
                  "Name": "NRetranslocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.008,
                  "Units": "g/g",
                  "Name": "MaximumNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": "/d",
                  "Name": "SenescenceRate",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.003,
                  "Units": "g/g",
                  "Name": "MinimumNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                  "Name": "ExpansionStress",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.XYPairs, Models",
                      "X": [
                        0.1,
                        0.7
                      ],
                      "Y": [
                        0.0,
                        1.0
                      ],
                      "XVariableName": None,
                      "Name": "XYPairs",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Maize].Leaf.Fw",
                      "Name": "XValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Library.BiomassRemoval, Models",
                  "Name": "BiomassRemovalDefaults",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.5,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.1,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Harvest",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.8,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.0,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Cut",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.6,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Prune",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.6,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.2,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Graze",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": "/d",
                  "Name": "DetachmentRateFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": "0-1",
                  "Name": "MaintenanceRespirationFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": "0-1",
                  "Name": "DMConversionEfficiency",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "DMReallocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "DMRetranslocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.VariableReference, Models",
                  "VariableName": "[Rachis].MinimumNConc",
                  "Name": "CriticalNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhaseLookupValue, Models",
                  "Start": "Emergence",
                  "End": "StartGrainFill",
                  "Name": "NitrogenDemandSwitch",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Constant",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.4,
                  "Units": None,
                  "Name": "CarbonConcentration",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientDemandFunctions, Models",
                  "Name": "DMDemands",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.DemandFunctions.PopulationBasedDemandFunction, Models",
                          "Name": "DMDemandFunction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Phenology].ThermalTime",
                              "Name": "ThermalTime",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 5.5,
                              "Units": None,
                              "Name": "StartStage",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 300.0,
                              "Units": None,
                              "Name": "GrowthDuration",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 23.0,
                              "Units": None,
                              "Name": "MaximumOrganWt",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Rachis].CobNumber",
                              "Name": "OrganPopulation",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.MinimumFunction, Models",
                              "Name": "ExpansionStress",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                                  "Name": "WaterStressEffect",
                                  "ResourceName": None,
                                  "Children": [
                                    {
                                      "$child": "Models.Functions.XYPairs, Models",
                                      "X": [
                                        0.1,
                                        0.7,
                                        1.3
                                      ],
                                      "Y": [
                                        0.0,
                                        1.0,
                                        1.0
                                      ],
                                      "XVariableName": None,
                                      "Name": "XYPairs",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    },
                                    {
                                      "$child": "Models.Functions.VariableReference, Models",
                                      "VariableName": "[Root].WaterTensionFactor",
                                      "Name": "XValue",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    }
                                  ],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                                  "Name": "TemperatureEffect",
                                  "ResourceName": None,
                                  "Children": [
                                    {
                                      "$child": "Models.Functions.XYPairs, Models",
                                      "X": [
                                        0.0,
                                        12.0,
                                        14.0
                                      ],
                                      "Y": [
                                        0.0,
                                        1.0,
                                        1.0
                                      ],
                                      "XVariableName": None,
                                      "Name": "XYPairs",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    },
                                    {
                                      "$child": "Models.Functions.VariableReference, Models",
                                      "VariableName": "[IWeather].MeanT",
                                      "Name": "XValue",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    }
                                  ],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                                  "Name": "NitrogenStressEffect",
                                  "ResourceName": None,
                                  "Children": [
                                    {
                                      "$child": "Models.Functions.XYPairs, Models",
                                      "X": [
                                        0.0,
                                        0.5,
                                        1.0
                                      ],
                                      "Y": [
                                        0.1,
                                        0.1,
                                        1.0
                                      ],
                                      "XVariableName": None,
                                      "Name": "XYPairs",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    },
                                    {
                                      "$child": "Models.Functions.VariableReference, Models",
                                      "VariableName": "[Maize].Leaf.Fn",
                                      "Name": "XValue",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    }
                                  ],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 1.0,
                          "Units": "g/g",
                          "Name": "StructuralFraction",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.DemandFunctions.StorageDMDemandFunction, Models",
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "StorageFraction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 1.0,
                              "Units": None,
                              "Name": "One",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Rachis].DMDemands.Structural.StructuralFraction",
                              "Name": "StructuralFraction",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStructuralPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QMetabolicPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStoragePriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientDemandFunctions, Models",
                  "Name": "NDemands",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Rachis].minimumNconc",
                          "Name": "MinNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Rachis].potentialDMAllocation.Structural",
                          "Name": "PotentialDMAllocation",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "MetabolicNconc",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Rachis].criticalNConc",
                              "Name": "CritNconc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Rachis].minimumNconc",
                              "Name": "MinNconc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Rachis].potentialDMAllocation.Structural",
                          "Name": "PotentialDMAllocation",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.DemandFunctions.StorageNDemandFunction, Models",
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Rachis].nitrogenDemandSwitch",
                          "Name": "NitrogenDemandSwitch",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Rachis].maximumNconc",
                          "Name": "MaxNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStructuralPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QMetabolicPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStoragePriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.RetranslocateNonStructural, Models",
                  "Name": "RetranslocateNitrogen",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientPoolFunctions, Models",
                  "Name": "InitialWt",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": "g/m^2",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "Photosynthesis",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.VariableReference, Models",
                  "VariableName": "[Rachis].MinimumNConc",
                  "Name": "initialNConcFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.Organs.GenericOrgan, Models",
              "StartLive": None,
              "DMSupply": None,
              "NSupply": None,
              "DMDemand": None,
              "NDemand": None,
              "potentialDMAllocation": None,
              "IsAboveGround": True,
              "Name": "Stem",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "RemobilisationCost",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                  "Name": "MaximumNConc",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.XYPairs, Models",
                      "X": [
                        0.0,
                        0.7,
                        1.0
                      ],
                      "Y": [
                        0.025,
                        0.025,
                        0.005
                      ],
                      "XVariableName": None,
                      "Name": "XYPairs",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Structure].RelativeNodeApperance",
                      "Name": "XValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.004,
                  "Units": "g/g",
                  "Name": "MinimumNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhaseLookupValue, Models",
                  "Start": "Emergence",
                  "End": "StartGrainFill",
                  "Name": "NitrogenDemandSwitch",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Constant",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "NReallocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhaseLookup, Models",
                  "Name": "NRetranslocationFactor",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.PhaseLookupValue, Models",
                      "Start": "Emergence",
                      "End": "FlagLeaf",
                      "Name": "PreStemElongation",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 0.0,
                          "Units": None,
                          "Name": "StemFraction",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.PhaseLookupValue, Models",
                      "Start": "FlagLeaf",
                      "End": "HarvestRipe",
                      "Name": "StemElongation",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 0.05,
                          "Units": None,
                          "Name": "NRetranslocationFactor",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.1,
                  "Units": None,
                  "Name": "DMRetranslocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Library.BiomassRemoval, Models",
                  "Name": "BiomassRemovalDefaults",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 1.0,
                      "FractionDeadToResidue": 1.0,
                      "Name": "Harvest",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 1.0,
                      "FractionDeadToResidue": 1.0,
                      "Name": "Cut",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 1.0,
                      "FractionDeadToResidue": 1.0,
                      "Name": "Prune",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.6,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.2,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Graze",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": "/d",
                  "Name": "DetachmentRateFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": "/d",
                  "Name": "SenescenceRate",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": "0-1",
                  "Name": "MaintenanceRespirationFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": "0-1",
                  "Name": "DMConversionEfficiency",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "DMReallocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.AddFunction, Models",
                  "Name": "CriticalNConc",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Stem].MinimumNConc",
                      "Name": "MinimumNConc",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "MetabolicNConc",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "NonStructuralNConc",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Stem].MaximumNConc",
                              "Name": "MaximumNConc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Stem].MinimumNConc",
                              "Name": "MinimumNConc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 0.2,
                          "Units": None,
                          "Name": "Proportion",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.4,
                  "Units": None,
                  "Name": "CarbonConcentration",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientDemandFunctions, Models",
                  "Name": "DMDemands",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.DemandFunctions.PartitionFractionDemandFunction, Models",
                          "Name": "DMDemandFunction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.LinearInterpolationFunction, Models",
                              "Name": "PartitionFraction",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.XYPairs, Models",
                                  "X": [
                                    3.0,
                                    5.0,
                                    6.0,
                                    7.0,
                                    7.0001
                                  ],
                                  "Y": [
                                    0.1,
                                    0.1,
                                    1.0,
                                    1.0,
                                    0.0
                                  ],
                                  "XVariableName": None,
                                  "Name": "XYPairs",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Phenology].Stage",
                                  "Name": "XValue",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.PhaseLookup, Models",
                          "Name": "StructuralFraction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.PhaseLookupValue, Models",
                              "Start": "Emergence",
                              "End": "StartGrainFill",
                              "Name": "PreStemElongation",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.Constant, Models",
                                  "FixedValue": 0.5,
                                  "Units": None,
                                  "Name": "StemFraction",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.PhaseLookupValue, Models",
                              "Start": "StartGrainFill",
                              "End": "HarvestRipe",
                              "Name": "StemElongation",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.Constant, Models",
                                  "FixedValue": 1.0,
                                  "Units": None,
                                  "Name": "StemFraction",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.DemandFunctions.StorageDMDemandFunction, Models",
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "StorageFraction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 1.0,
                              "Units": None,
                              "Name": "One",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Stem].DMDemands.Structural.StructuralFraction",
                              "Name": "StructuralFraction",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStructuralPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QMetabolicPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStoragePriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientDemandFunctions, Models",
                  "Name": "NDemands",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Stem].minimumNconc",
                          "Name": "MinNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Stem].potentialDMAllocation.Structural",
                          "Name": "PotentialDMAllocation",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "MetabolicNconc",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Stem].criticalNConc",
                              "Name": "CritNconc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Stem].minimumNconc",
                              "Name": "MinNconc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Stem].potentialDMAllocation.Structural",
                          "Name": "PotentialDMAllocation",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.DemandFunctions.StorageNDemandFunction, Models",
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Stem].nitrogenDemandSwitch",
                          "Name": "NitrogenDemandSwitch",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Stem].maximumNconc",
                          "Name": "MaxNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStructuralPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QMetabolicPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStoragePriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.RetranslocateNonStructural, Models",
                  "Name": "RetranslocateNitrogen",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientPoolFunctions, Models",
                  "Name": "InitialWt",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": "g/m^2",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "Photosynthesis",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.VariableReference, Models",
                  "VariableName": "[Stem].MinimumNConc",
                  "Name": "initialNConcFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.CompositeBiomass, Models",
              "OrganNames": [
                "Leaf",
                "Stem",
                "Husk",
                "Grain",
                "Rachis"
              ],
              "IncludeLive": True,
              "IncludeDead": True,
              "Name": "AboveGround",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Memo, Models",
                  "Text": "\nThe AboveGround composite biomass object includes all organs occurring above the soil surface to allow output for total dry matter and Nutrient contents, including total structural, non-structural and metabolic pools.\n",
                  "Name": "memo",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.CompositeBiomass, Models",
              "OrganNames": [
                "Root"
              ],
              "IncludeLive": True,
              "IncludeDead": True,
              "Name": "BelowGround",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Memo, Models",
                  "Text": "\nThe BelowGround composite biomass object includes all root biomass pools to allow output for total dry matter and Nutrient contents, including total structural, non-structural and metabolic pools.\n",
                  "Name": "memo",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.CompositeBiomass, Models",
              "OrganNames": [
                "Leaf",
                "Stem",
                "Husk",
                "Grain",
                "Rachis",
                "Root"
              ],
              "IncludeLive": True,
              "IncludeDead": True,
              "Name": "Total",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Memo, Models",
                  "Text": "\nThe Total composite biomass object includes all organs (above+below ground) to allow output for total dry matter and Nutrient contents, including total structural, non-structural and metabolic pools.\n",
                  "Name": "memo",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.CompositeBiomass, Models",
              "OrganNames": [
                "Leaf",
                "Stem",
                "Husk",
                "Grain",
                "Rachis",
                "Root"
              ],
              "IncludeLive": True,
              "IncludeDead": False,
              "Name": "TotalLive",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Memo, Models",
                  "Text": "\nThe TotalLive composite biomass object includes the Live pools of all organs (above+below ground) to allow output for total dry matter and Nutrient contents, including total structural, non-structural and metabolic pools.\n",
                  "Name": "memo",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.CompositeBiomass, Models",
              "OrganNames": [
                "Leaf",
                "Stem",
                "Husk",
                "Grain",
                "Rachis",
                "Root"
              ],
              "IncludeLive": False,
              "IncludeDead": True,
              "Name": "TotalDead",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Memo, Models",
                  "Text": "\nThe TotalDead composite biomass object includes the Dead pools of all organs (above+below ground) to allow output for total dry matter and Nutrient contents, including total structural, non-structural and metabolic pools.\n",
                  "Name": "memo",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.CompositeBiomass, Models",
              "OrganNames": [
                "Husk",
                "Grain",
                "Rachis"
              ],
              "IncludeLive": True,
              "IncludeDead": False,
              "Name": "EarLive",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Memo, Models",
                  "Text": "\nThe EarLive composite biomass object includes the Live pools of all organs within the Ear to allow output for total dry matter and Nutrient contents, including total structural, non-structural and metabolic pools.\n",
                  "Name": "memo",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.CompositeBiomass, Models",
              "OrganNames": [
                "Leaf",
                "Stem",
                "Husk",
                "Grain",
                "Rachis"
              ],
              "IncludeLive": True,
              "IncludeDead": False,
              "Name": "AboveGroundLive",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Memo, Models",
                  "Text": "\nThe AboveGroundLive composite biomass object includes the Live pools of all organs occurring above the soil surface to allow output for total dry matter and Nutrient contents, including total structural, non-structural and metabolic pools.\n",
                  "Name": "memo",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.CompositeBiomass, Models",
              "OrganNames": [
                "Leaf",
                "Stem",
                "Husk",
                "Grain",
                "Rachis"
              ],
              "IncludeLive": False,
              "IncludeDead": True,
              "Name": "AboveGroundDead",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Memo, Models",
                  "Text": "\nThe AboveGroundDead composite biomass object includes the Dead pools of all organs occurring above the soil surface to allow output for total dry matter and Nutrient contents, including total structural, non-structural and metabolic pools.\n",
                  "Name": "memo",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.Memo, Models",
              "Text": "#Future Development requirements\n\n* Phenology not responding to stress events.  Requires better data sets for quantifying the effects of water and nitrogen on leaf apperance and the timing of development stages.\n* Phosphorus response\n* Improved parameterisation of supply and demand for N and DM from organs and arbitration of these resources.\n* Heat Stress responses in grain number\n* More validation under a wider range of environments with more detailed datasets\n* Seedling mortality\n* Water demand (MicroClimate) needs validation\n\n\n\n\n",
              "Name": "Future Development Needs",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.Core.Folder, Models",
              "ShowInDocs": False,
              "GraphsPerPage": 6,
              "Name": "CultivarFolder",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Phenology].Juvenile.Target.FixedValue = 190",
                    "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                    "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 124",
                    "[Phenology].FlagLeafToFlowering.Target.FixedValue = 10",
                    "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                    "[Phenology].GrainFilling.Target.FixedValue = 520",
                    "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 25",
                    "[Grain].MaximumGrainsPerCob.FixedValue =  550"
                  ],
                  "Name": "Hycorn_53",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Phenology].GrainFilling.Target.FixedValue = 550",
                    "[Phenology].Juvenile.Target.FixedValue = 270"
                  ],
                  "Name": "Pioneer_33M54",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Phenology].Juvenile.Target.FixedValue = 210",
                    "[Phenology].GrainFilling.Target.FixedValue = 550"
                  ],
                  "Name": "Pioneer_38H20",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Phenology].Juvenile.Target.FixedValue = 280",
                    "[Phenology].GrainFilling.Target.FixedValue = 550"
                  ],
                  "Name": "Pioneer_34K77",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nVShort duration variety released in New Zealand by Pioneer\n",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Phenology].Juvenile.Target.FixedValue = 100",
                    "[Phenology].GrainFilling.Target.FixedValue = 500",
                    "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                    "[Structure].Phyllochron.Phyllochron.Phyllochron.XYPairs.X = 1, 4, 4.2, 10.5, 11",
                    "[Structure].Phyllochron.Phyllochron.Phyllochron.XYPairs.Y = 26, 26, 46, 46, 70",
                    "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 18",
                    "[Structure].HeightModel.PotentialHeight.XYPairs.X = 3, 4, 5, 6",
                    "[Structure].HeightModel.PotentialHeight.XYPairs.Y = 20, 40, 50, 1300"
                  ],
                  "Name": "Pioneer_39V43",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nVShort duration variety released in New Zealand by Pioneer\n",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Phenology].Juvenile.Target.FixedValue = 150",
                    "[Phenology].GrainFilling.Target.FixedValue = 500",
                    "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                    "[Structure].Phyllochron.Phyllochron.Phyllochron.XYPairs.X = 1, 4, 4.2, 10.5, 11",
                    "[Structure].Phyllochron.Phyllochron.Phyllochron.XYPairs.Y = 26, 26, 46, 46, 70",
                    "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 18",
                    "[Structure].HeightModel.PotentialHeight.XYPairs.X = 3, 4, 5, 6",
                    "[Structure].HeightModel.PotentialHeight.XYPairs.Y = 20, 40, 50, 2370"
                  ],
                  "Name": "Atrium",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nMaize variety used in trials at Wageningen University\n",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Phenology].Juvenile.Target.FixedValue = 211",
                    "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                    "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 0",
                    "[Phenology].FlagLeafToFlowering.Target.FixedValue = 1",
                    "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                    "[Phenology].GrainFilling.Target.FixedValue = 812",
                    "[Phenology].Maturing.Target.FixedValue = 1",
                    "[Phenology].MaturityToHarvestRipe.Target.FixedValue = 1",
                    "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 36",
                    "[Grain].MaximumGrainsPerCob.FixedValue =  720"
                  ],
                  "Name": "Laila",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Phenology].Juvenile.Target.FixedValue = 225",
                    "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24.0",
                    "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 124",
                    "[Phenology].FlagLeafToFlowering.Target.FixedValue = 101",
                    "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                    "[Phenology].GrainFilling.Target.FixedValue = 694",
                    "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 25",
                    "[Grain].MaximumGrainsPerCob.FixedValue = 550"
                  ],
                  "Name": "GH_5019WX",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Phenology].Juvenile.Target.FixedValue = 190",
                    "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                    "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 124",
                    "[Phenology].FlagLeafToFlowering.Target.FixedValue = 50",
                    "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                    "[Phenology].GrainFilling.Target.FixedValue = 713",
                    "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 25",
                    "[Grain].MaximumGrainsPerCob.FixedValue =  550"
                  ],
                  "Name": "Hycorn_40",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Phenology].Juvenile.Target.FixedValue = 190",
                    "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                    "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 300",
                    "[Phenology].FlagLeafToFlowering.Target.FixedValue = 50",
                    "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                    "[Phenology].GrainFilling.Target.FixedValue = 713",
                    "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 25",
                    "[Grain].MaximumGrainsPerCob.FixedValue =  550"
                  ],
                  "Name": "GH_5009",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Phenology].Juvenile.Target.FixedValue = 211",
                    "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                    "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 472",
                    "[Phenology].FlagLeafToFlowering.Target.FixedValue = 101",
                    "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                    "[Phenology].GrainFilling.Target.FixedValue = 813",
                    "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 25",
                    "[Grain].MaximumGrainsPerCob.FixedValue = 550",
                    "[Grain].MaxNconc 1.75"
                  ],
                  "Name": "Dekalb_XL82",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Phenology].Juvenile.Target.FixedValue = 280",
                    "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                    "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 264.5",
                    "[Phenology].FlagLeafToFlowering.Target.FixedValue = 10",
                    "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                    "[Phenology].GrainFilling.Target.FixedValue = 600",
                    "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 14",
                    "[Grain].MaximumGrainsPerCob.FixedValue = 360",
                    "[Grain].MaximumPotentialGrainSize.FixedValue = 0.217"
                  ],
                  "Name": "malawi_local",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nThe variety malawi_local is a Malawian medium-late maturing local variety.  Photoperiod response is taken from Ellis et al Crop Sci 32:1225 for SR52.\n",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Phenology].Juvenile.Target.FixedValue = 290",
                    "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                    "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 264.5",
                    "[Phenology].FlagLeafToFlowering.Target.FixedValue = 10",
                    "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                    "[Phenology].GrainFilling.Target.FixedValue = 570",
                    "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 14",
                    "[Grain].MaximumGrainsPerCob.FixedValue = 504",
                    "[Grain].MaximumPotentialGrainSize.FixedValue = 0.361"
                  ],
                  "Name": "mh19",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nThe variety mh19 is a Malawian medium-late maturing hybrid.  Photoperiod response is taken from Ellis et al Crop Sci 32:1225 for SR52.\n",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Phenology].Juvenile.Target.FixedValue = 290",
                    "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                    "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 264.5",
                    "[Phenology].FlagLeafToFlowering.Target.FixedValue = 10",
                    "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                    "[Phenology].GrainFilling.Target.FixedValue = 570",
                    "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 14",
                    "[Grain].MaximumGrainsPerCob.FixedValue = 504",
                    "[Grain].MaximumPotentialGrainSize.FixedValue = 0.361"
                  ],
                  "Name": "mh17",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nThe variety mh17 is a Malawian medium-late maturing hybrid.  Photoperiod response is taken from Ellis et al Crop Sci 32:1225 for SR52.\n",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Phenology].Juvenile.Target.FixedValue = 290",
                    "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                    "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 264.5",
                    "[Phenology].FlagLeafToFlowering.Target.FixedValue = 10",
                    "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                    "[Phenology].GrainFilling.Target.FixedValue = 570",
                    "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 14",
                    "[Grain].MaximumGrainsPerCob.FixedValue = 504",
                    "[Grain].MaximumPotentialGrainSize.FixedValue = 0.361"
                  ],
                  "Name": "mh16",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nThe variety mh16 is a Malawian medium-late maturing hybrid.  Photoperiod response is taken from Ellis et al Crop Sci 32:1225 for SR52.\n",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Phenology].Juvenile.Target.FixedValue = 290",
                    "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                    "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 264.5",
                    "[Phenology].FlagLeafToFlowering.Target.FixedValue = 10",
                    "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                    "[Phenology].GrainFilling.Target.FixedValue = 570",
                    "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 14",
                    "[Grain].MaximumGrainsPerCob.FixedValue = 504",
                    "[Grain].MaximumPotentialGrainSize.FixedValue = 0.361"
                  ],
                  "Name": "mh12",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nThe variety mh12 is a Malawian medium-late maturing hybrid.  Photoperiod response is taken from Ellis et al Crop Sci 32:1225 for SR52.\n",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Phenology].Juvenile.Target.FixedValue = 280",
                    "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                    "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 264.5",
                    "[Phenology].FlagLeafToFlowering.Target.FixedValue = 10",
                    "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                    "[Phenology].GrainFilling.Target.FixedValue = 550",
                    "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 14",
                    "[Grain].MaximumGrainsPerCob.FixedValue = 504",
                    "[Grain].MaximumPotentialGrainSize.FixedValue = 0.361"
                  ],
                  "Name": "sc623",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nThe variety sc623 is a Zimbabwean medium-late maturing hybrid.  Photoperiod response is taken from Ellis et al Crop Sci 32:1225 for SR52.\n",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Phenology].Juvenile.Target.FixedValue = 280",
                    "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                    "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 264.5",
                    "[Phenology].FlagLeafToFlowering.Target.FixedValue = 10",
                    "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                    "[Phenology].GrainFilling.Target.FixedValue = 570",
                    "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 14",
                    "[Grain].MaximumGrainsPerCob.FixedValue = 504",
                    "[Grain].MaximumPotentialGrainSize.FixedValue = 0.361"
                  ],
                  "Name": "sc625",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nThe variety sc625 is a Zimbabwean medium-late maturing hybrid.  Photoperiod response is taken from Ellis et al Crop Sci 32:1225 for SR52.\n",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Phenology].Juvenile.Target.FixedValue = 280",
                    "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                    "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 264.5",
                    "[Phenology].FlagLeafToFlowering.Target.FixedValue = 10",
                    "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                    "[Phenology].GrainFilling.Target.FixedValue = 560",
                    "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 14",
                    "[Grain].MaximumGrainsPerCob.FixedValue = 495",
                    "[Grain].MaximumPotentialGrainSize.FixedValue = 0.361"
                  ],
                  "Name": "sc601",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nThe variety sc601 is a Zimbabwean medium-late maturing hybrid.  Photoperiod response is taken from Ellis et al Crop Sci 32:1225 for SR52.\n",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Phenology].Juvenile.Target.FixedValue = 233",
                    "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                    "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 264.5",
                    "[Phenology].FlagLeafToFlowering.Target.FixedValue = 10",
                    "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                    "[Phenology].GrainFilling.Target.FixedValue = 550",
                    "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 14",
                    "[Grain].MaximumGrainsPerCob.FixedValue = 468",
                    "[Grain].MaximumPotentialGrainSize.FixedValue = 0.361"
                  ],
                  "Name": "CG4141",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nThe variety CG4141 is a Zimbabwean early-medium maturing hybrid.  Photoperiod response is taken from Ellis et al Crop Sci 32:1225 for SR52.\n",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Phenology].Juvenile.Target.FixedValue = 260",
                    "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                    "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 264.5",
                    "[Phenology].FlagLeafToFlowering.Target.FixedValue = 10",
                    "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                    "[Phenology].GrainFilling.Target.FixedValue = 550",
                    "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 14",
                    "[Grain].MaximumGrainsPerCob.FixedValue = 486",
                    "[Grain].MaximumPotentialGrainSize.FixedValue = 0.361"
                  ],
                  "Name": "mh18",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nThe variety mh18 is a Zimbabwean early-medium maturing hybrid.  Photoperiod response is taken from Ellis et al Crop Sci 32:1225 for SR52.\n",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Phenology].Juvenile.Target.FixedValue = 250",
                    "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                    "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 264.5",
                    "[Phenology].FlagLeafToFlowering.Target.FixedValue = 10",
                    "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                    "[Phenology].GrainFilling.Target.FixedValue = 550",
                    "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 14",
                    "[Grain].MaximumGrainsPerCob.FixedValue = 468",
                    "[Grain].MaximumPotentialGrainSize.FixedValue = 0.361"
                  ],
                  "Name": "r215",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nThe variety r215 is a Zimbabwean early-medium maturing hybrid.  Photoperiod response is taken from Ellis et al Crop Sci 32:1225 for SR52.\n",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Phenology].Juvenile.Target.FixedValue = 230",
                    "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                    "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 264.5",
                    "[Phenology].FlagLeafToFlowering.Target.FixedValue = 10",
                    "[Phenology].FloweringToGrainFilling.Target.FixedValue = 160",
                    "[Phenology].GrainFilling.Target.FixedValue = 550",
                    "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 14",
                    "[Grain].MaximumGrainsPerCob.FixedValue = 387",
                    "[Grain].MaximumPotentialGrainSize.FixedValue = 0.33"
                  ],
                  "Name": "Melkassa",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nThe variety Melkassa is a local early-medium maturing Ethiopian variety.\n",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Phenology].Juvenile.Target.FixedValue = 250",
                    "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                    "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 264.5",
                    "[Phenology].FlagLeafToFlowering.Target.FixedValue = 10",
                    "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                    "[Phenology].GrainFilling.Target.FixedValue = 550",
                    "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 14",
                    "[Grain].MaximumGrainsPerCob.FixedValue = 468",
                    "[Grain].MaximumPotentialGrainSize.FixedValue = 0.361"
                  ],
                  "Name": "sr52",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nThe variety sr52 is a Zimbabwean early-medium maturing hybrid.  Photoperiod response is taken from Ellis et al Crop Sci 32:1225 for SR52.\n",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Phenology].Juvenile.Target.FixedValue = 250",
                    "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                    "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 264.5",
                    "[Phenology].FlagLeafToFlowering.Target.FixedValue = 10",
                    "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                    "[Phenology].GrainFilling.Target.FixedValue = 550",
                    "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 14",
                    "[Grain].MaximumGrainsPerCob.FixedValue = 468",
                    "[Grain].MaximumPotentialGrainSize.FixedValue = 0.361"
                  ],
                  "Name": "sc501",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nThe variety sc501 is a Zimbabwean early-medium maturing hybrid.  Photoperiod response is taken from Ellis et al Crop Sci 32:1225 for SR52.\n",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Phenology].Juvenile.Target.FixedValue = 230",
                    "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                    "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 264.5",
                    "[Phenology].FlagLeafToFlowering.Target.FixedValue = 10",
                    "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                    "[Phenology].GrainFilling.Target.FixedValue = 550",
                    "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 14",
                    "[Grain].MaximumGrainsPerCob.FixedValue = 450",
                    "[Grain].MaximumPotentialGrainSize.FixedValue = 0.375"
                  ],
                  "Name": "r201",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nThe variety r201 is a Zimbabwean early maturing hybrid.  Photoperiod response is taken from Ellis et al Crop Sci 32:1225 for SR52.\n",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Phenology].Juvenile.Target.FixedValue = 230",
                    "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                    "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 264.5",
                    "[Phenology].FlagLeafToFlowering.Target.FixedValue = 10",
                    "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                    "[Phenology].GrainFilling.Target.FixedValue = 550",
                    "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 14",
                    "[Grain].MaximumGrainsPerCob.FixedValue = 450",
                    "[Grain].MaximumPotentialGrainSize.FixedValue = 0.375"
                  ],
                  "Name": "sc401",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nThe variety NSCM_41 is a Zimbabwean early maturing hybrid.  Photoperiod response is taken from Ellis et al Crop Sci 32:1225 for SR52.\n",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Phenology].Juvenile.Target.FixedValue = 230",
                    "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                    "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 264.5",
                    "[Phenology].FlagLeafToFlowering.Target.FixedValue = 10",
                    "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                    "[Phenology].GrainFilling.Target.FixedValue = 550",
                    "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 14",
                    "[Grain].MaximumGrainsPerCob.FixedValue = 450",
                    "[Grain].MaximumPotentialGrainSize.FixedValue = 0.375"
                  ],
                  "Name": "NSCM_41",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nThe variety NSCM_41 is a Zimbabwean early maturing hybrid.  Photoperiod response is taken from Ellis et al Crop Sci 32:1225 for SR52.\n",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Phenology].Juvenile.Target.FixedValue = 150",
                    "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                    "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 115",
                    "[Phenology].FlagLeafToFlowering.Target.FixedValue = 10",
                    "[Phenology].FloweringToGrainFilling.Target.FixedValue = 120",
                    "[Phenology].GrainFilling.Target.FixedValue = 530",
                    "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 14",
                    "[Grain].MaximumGrainsPerCob.FixedValue = 432",
                    "[Grain].MaximumPotentialGrainSize.FixedValue = 0.375"
                  ],
                  "Name": "Makueni",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nThe variety Makueni Composite is a kenyan variety with similar phenology but slightly higher cob size than Katumani composite.\n",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Phenology].Juvenile.Target.FixedValue = 150",
                    "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                    "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 115",
                    "[Phenology].FlagLeafToFlowering.Target.FixedValue = 10",
                    "[Phenology].FloweringToGrainFilling.Target.FixedValue = 120",
                    "[Phenology].GrainFilling.Target.FixedValue = 530",
                    "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 14",
                    "[Grain].MaximumGrainsPerCob.FixedValue = 405",
                    "[Grain].MaximumPotentialGrainSize.FixedValue = 0.375"
                  ],
                  "Name": "Katumani",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nThe variety Katumani Composite B was formed in 1966 and released in 1968 as part of the maize breeding work undertaken at the Katumani Research Station in Kenya which started in 1956.  Katumani Composite B was an advanced generation of Katumani Synthetic V X Katumani Synthetic VI.  It has slight photoperiod sensitivity and relatively small potential cob size.\n",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    ""
                  ],
                  "Name": "Pioneer_3153",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Phenology].Juvenile.Target.FixedValue = 180",
                    "[Phenology].GrainFilling.Target.FixedValue = 500",
                    "[Grain].MaximumGrainsPerCob.FixedValue = 550",
                    "[Grain].NumberFunction.GrowthRateFactor.XYPairs.X = 0.2,2,4,10",
                    "[Grain].NumberFunction.GrowthRateFactor.XYPairs.Y = 0,0.6,1,1",
                    "[Grain].MaximumPotentialGrainSize.FixedValue = 0.25"
                  ],
                  "Name": "Pioneer_39G12",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Core.Folder, Models",
                  "ShowInDocs": False,
                  "GraphsPerPage": 6,
                  "Name": "Generic",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Juvenile.Target.FixedValue = 135",
                        "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                        "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 0",
                        "[Phenology].FlagLeafToFlowering.Target.FixedValue = 1",
                        "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                        "[Phenology].GrainFilling.Target.FixedValue = 430",
                        "[Phenology].Maturing.Target.FixedValue = 1",
                        "[Phenology].MaturityToHarvestRipe.Target.FixedValue = 100",
                        "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 36"
                      ],
                      "Name": "B_80",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Juvenile.Target.FixedValue = 155",
                        "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                        "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 0",
                        "[Phenology].FlagLeafToFlowering.Target.FixedValue = 1",
                        "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                        "[Phenology].GrainFilling.Target.FixedValue = 530",
                        "[Phenology].Maturing.Target.FixedValue = 1",
                        "[Phenology].MaturityToHarvestRipe.Target.FixedValue = 100",
                        "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 36"
                      ],
                      "Name": "B_90",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Juvenile.Target.FixedValue = 170",
                        "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                        "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 0",
                        "[Phenology].FlagLeafToFlowering.Target.FixedValue = 1",
                        "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                        "[Phenology].GrainFilling.Target.FixedValue = 578",
                        "[Phenology].Maturing.Target.FixedValue = 1",
                        "[Phenology].MaturityToHarvestRipe.Target.FixedValue = 100",
                        "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 36"
                      ],
                      "Name": "B_95",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Juvenile.Target.FixedValue = 185",
                        "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                        "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 0",
                        "[Phenology].FlagLeafToFlowering.Target.FixedValue = 1",
                        "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                        "[Phenology].GrainFilling.Target.FixedValue = 630",
                        "[Phenology].Maturing.Target.FixedValue = 1",
                        "[Phenology].MaturityToHarvestRipe.Target.FixedValue = 100",
                        "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 36"
                      ],
                      "Name": "B_100",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Juvenile.Target.FixedValue = 188",
                        "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                        "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 0",
                        "[Phenology].FlagLeafToFlowering.Target.FixedValue = 1",
                        "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                        "[Phenology].GrainFilling.Target.FixedValue = 654",
                        "[Phenology].Maturing.Target.FixedValue = 1",
                        "[Phenology].MaturityToHarvestRipe.Target.FixedValue = 100",
                        "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 36"
                      ],
                      "Name": "B_103",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Juvenile.Target.FixedValue = 192",
                        "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                        "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 0",
                        "[Phenology].FlagLeafToFlowering.Target.FixedValue = 1",
                        "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                        "[Phenology].GrainFilling.Target.FixedValue = 673",
                        "[Phenology].Maturing.Target.FixedValue = 1",
                        "[Phenology].MaturityToHarvestRipe.Target.FixedValue = 100",
                        "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 36"
                      ],
                      "Name": "B_105",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Juvenile.Target.FixedValue = 199",
                        "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                        "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 0",
                        "[Phenology].FlagLeafToFlowering.Target.FixedValue = 1",
                        "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                        "[Phenology].GrainFilling.Target.FixedValue = 701",
                        "[Phenology].Maturing.Target.FixedValue = 1",
                        "[Phenology].MaturityToHarvestRipe.Target.FixedValue = 100",
                        "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 36"
                      ],
                      "Name": "B_108",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Juvenile.Target.FixedValue = 210",
                        "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                        "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 0",
                        "[Phenology].FlagLeafToFlowering.Target.FixedValue = 1",
                        "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                        "[Phenology].GrainFilling.Target.FixedValue = 730",
                        "[Phenology].Maturing.Target.FixedValue = 1",
                        "[Phenology].MaturityToHarvestRipe.Target.FixedValue = 100",
                        "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 36"
                      ],
                      "Name": "B_110",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Juvenile.Target.FixedValue = 208",
                        "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                        "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 0",
                        "[Phenology].FlagLeafToFlowering.Target.FixedValue = 1",
                        "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                        "[Phenology].GrainFilling.Target.FixedValue = 739",
                        "[Phenology].Maturing.Target.FixedValue = 1",
                        "[Phenology].MaturityToHarvestRipe.Target.FixedValue = 100",
                        "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 36"
                      ],
                      "Name": "B_112",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Juvenile.Target.FixedValue = 215",
                        "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                        "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 0",
                        "[Phenology].FlagLeafToFlowering.Target.FixedValue = 1",
                        "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                        "[Phenology].GrainFilling.Target.FixedValue = 767",
                        "[Phenology].Maturing.Target.FixedValue = 1",
                        "[Phenology].MaturityToHarvestRipe.Target.FixedValue = 100",
                        "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 36"
                      ],
                      "Name": "B_115",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Juvenile.Target.FixedValue = 220",
                        "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                        "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 0",
                        "[Phenology].FlagLeafToFlowering.Target.FixedValue = 1",
                        "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                        "[Phenology].GrainFilling.Target.FixedValue = 815",
                        "[Phenology].Maturing.Target.FixedValue = 1",
                        "[Phenology].MaturityToHarvestRipe.Target.FixedValue = 100",
                        "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 36"
                      ],
                      "Name": "B_120",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Juvenile.Target.FixedValue = 250",
                        "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                        "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 0",
                        "[Phenology].FlagLeafToFlowering.Target.FixedValue = 1",
                        "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                        "[Phenology].GrainFilling.Target.FixedValue = 900",
                        "[Phenology].Maturing.Target.FixedValue = 1",
                        "[Phenology].MaturityToHarvestRipe.Target.FixedValue = 100",
                        "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 36"
                      ],
                      "Name": "B_130",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Juvenile.Target.FixedValue = 140",
                        "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                        "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 0",
                        "[Phenology].FlagLeafToFlowering.Target.FixedValue = 1",
                        "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                        "[Phenology].GrainFilling.Target.FixedValue = 420",
                        "[Phenology].Maturing.Target.FixedValue = 1",
                        "[Phenology].MaturityToHarvestRipe.Target.FixedValue = 100",
                        "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 36"
                      ],
                      "Name": "A_80",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Juvenile.Target.FixedValue = 170",
                        "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                        "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 0",
                        "[Phenology].FlagLeafToFlowering.Target.FixedValue = 1",
                        "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                        "[Phenology].GrainFilling.Target.FixedValue = 505",
                        "[Phenology].Maturing.Target.FixedValue = 1",
                        "[Phenology].MaturityToHarvestRipe.Target.FixedValue = 100",
                        "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 36"
                      ],
                      "Name": "A_90",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Juvenile.Target.FixedValue = 181",
                        "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                        "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 0",
                        "[Phenology].FlagLeafToFlowering.Target.FixedValue = 1",
                        "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                        "[Phenology].GrainFilling.Target.FixedValue = 558",
                        "[Phenology].Maturing.Target.FixedValue = 1",
                        "[Phenology].MaturityToHarvestRipe.Target.FixedValue = 100",
                        "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 36"
                      ],
                      "Name": "A_95",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Juvenile.Target.FixedValue = 190",
                        "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                        "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 0",
                        "[Phenology].FlagLeafToFlowering.Target.FixedValue = 1",
                        "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                        "[Phenology].GrainFilling.Target.FixedValue = 615",
                        "[Phenology].Maturing.Target.FixedValue = 1",
                        "[Phenology].MaturityToHarvestRipe.Target.FixedValue = 100",
                        "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 36"
                      ],
                      "Name": "A_100",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Juvenile.Target.FixedValue = 204",
                        "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                        "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 0",
                        "[Phenology].FlagLeafToFlowering.Target.FixedValue = 1",
                        "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                        "[Phenology].GrainFilling.Target.FixedValue = 630",
                        "[Phenology].Maturing.Target.FixedValue = 1",
                        "[Phenology].MaturityToHarvestRipe.Target.FixedValue = 100",
                        "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 36"
                      ],
                      "Name": "A_103",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Juvenile.Target.FixedValue = 209",
                        "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                        "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 0",
                        "[Phenology].FlagLeafToFlowering.Target.FixedValue = 1",
                        "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                        "[Phenology].GrainFilling.Target.FixedValue = 648",
                        "[Phenology].Maturing.Target.FixedValue = 1",
                        "[Phenology].MaturityToHarvestRipe.Target.FixedValue = 100",
                        "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 36"
                      ],
                      "Name": "A_105",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Juvenile.Target.FixedValue = 218",
                        "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                        "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 0",
                        "[Phenology].FlagLeafToFlowering.Target.FixedValue = 1",
                        "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                        "[Phenology].GrainFilling.Target.FixedValue = 674",
                        "[Phenology].Maturing.Target.FixedValue = 1",
                        "[Phenology].MaturityToHarvestRipe.Target.FixedValue = 100",
                        "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 36"
                      ],
                      "Name": "A_108",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Juvenile.Target.FixedValue = 230",
                        "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                        "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 0",
                        "[Phenology].FlagLeafToFlowering.Target.FixedValue = 1",
                        "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                        "[Phenology].GrainFilling.Target.FixedValue = 700",
                        "[Phenology].Maturing.Target.FixedValue = 1",
                        "[Phenology].MaturityToHarvestRipe.Target.FixedValue = 100",
                        "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 36"
                      ],
                      "Name": "A_110",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Juvenile.Target.FixedValue = 229",
                        "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                        "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 0",
                        "[Phenology].FlagLeafToFlowering.Target.FixedValue = 1",
                        "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                        "[Phenology].GrainFilling.Target.FixedValue = 710",
                        "[Phenology].Maturing.Target.FixedValue = 1",
                        "[Phenology].MaturityToHarvestRipe.Target.FixedValue = 100",
                        "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 36"
                      ],
                      "Name": "A_112",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Juvenile.Target.FixedValue = 238",
                        "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                        "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 0",
                        "[Phenology].FlagLeafToFlowering.Target.FixedValue = 1",
                        "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                        "[Phenology].GrainFilling.Target.FixedValue = 737",
                        "[Phenology].Maturing.Target.FixedValue = 1",
                        "[Phenology].MaturityToHarvestRipe.Target.FixedValue = 100",
                        "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 36"
                      ],
                      "Name": "A_115",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Juvenile.Target.FixedValue = 250",
                        "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                        "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 0",
                        "[Phenology].FlagLeafToFlowering.Target.FixedValue = 1",
                        "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                        "[Phenology].GrainFilling.Target.FixedValue = 775",
                        "[Phenology].Maturing.Target.FixedValue = 1",
                        "[Phenology].MaturityToHarvestRipe.Target.FixedValue = 100",
                        "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 36"
                      ],
                      "Name": "A_120",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Juvenile.Target.FixedValue = 285",
                        "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                        "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 0",
                        "[Phenology].FlagLeafToFlowering.Target.FixedValue = 1",
                        "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                        "[Phenology].GrainFilling.Target.FixedValue = 865",
                        "[Phenology].Maturing.Target.FixedValue = 1",
                        "[Phenology].MaturityToHarvestRipe.Target.FixedValue = 100",
                        "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 36"
                      ],
                      "Name": "A_130",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Juvenile.Target.FixedValue = 212",
                        "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                        "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 0",
                        "[Phenology].FlagLeafToFlowering.Target.FixedValue = 1",
                        "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                        "[Phenology].GrainFilling.Target.FixedValue = 775",
                        "[Phenology].Maturing.Target.FixedValue = 1",
                        "[Phenology].MaturityToHarvestRipe.Target.FixedValue = 100",
                        "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 36",
                        "[Leaf].Photosynthesis.RUE.FixedValue = 2.2",
                        "[Grain].MaximumPotentialGrainSize.FixedValue = 0.36",
                        "[Grain].MaximumGrainsPerCob.FixedValue = 850"
                      ],
                      "Name": "HY_110",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Juvenile.Target.FixedValue = 216",
                        "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                        "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 0",
                        "[Phenology].FlagLeafToFlowering.Target.FixedValue = 1",
                        "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                        "[Phenology].GrainFilling.Target.FixedValue = 705",
                        "[Phenology].Maturing.Target.FixedValue = 1",
                        "[Phenology].MaturityToHarvestRipe.Target.FixedValue = 100",
                        "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 36",
                        "[Leaf].Photosynthesis.RUE.FixedValue = 1.7",
                        "[Grain].MaximumPotentialGrainSize.FixedValue = 0.25",
                        "[Grain].MaximumGrainsPerCob.FixedValue = 550"
                      ],
                      "Name": "LY_110",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Juvenile.Target.FixedValue = 270",
                        "[Phenology].Photosensitive.Target.XYPairs.Y = 0,0,0",
                        "[Phenology].FlagLeafToFlowering.Target.FixedValue = 1",
                        "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                        "[Phenology].GrainFilling.Target.FixedValue = 640",
                        "[Phenology].Maturing.Target.FixedValue=50",
                        "[Phenology].MaturityToHarvestRipe.Target.FixedValue=50",
                        "[Maize].Grain.MaximumGrainsPerCob.FixedValue = 640",
                        "[Maize].Grain.MaximumPotentialGrainSize.FixedValue = 0.32",
                        "[Maize].Root.SpecificRootLength.FixedValue = 135",
                        "[Maize].Root.RootFrontVelocity.PotentialRootFrontVelocity.PreFlowering.RootFrontVelocity.FixedValue=22",
                        "[Grain].MaximumNConc.InitialPhase.InitialNconc.FixedValue=0.03"
                      ],
                      "Name": "P1197",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Phenology].Juvenile.Target.FixedValue = 211",
                    "[Phenology].Photosensitive.Target.XYPairs.X = 0, 12.5, 24",
                    "[Phenology].Photosensitive.Target.XYPairs.Y = 0, 0, 0",
                    "[Phenology].FlagLeafToFlowering.Target.FixedValue = 1",
                    "[Phenology].FloweringToGrainFilling.Target.FixedValue = 170",
                    "[Phenology].GrainFilling.Target.FixedValue = 812",
                    "[Phenology].Maturing.Target.FixedValue = 1",
                    "[Phenology].MaturityToHarvestRipe.Target.FixedValue = 1",
                    "[Rachis].DMDemands.Structural.DMDemandFunction.MaximumOrganWt.FixedValue = 36",
                    "[Grain].MaximumGrainsPerCob.FixedValue =  720",
                    "[leaf].Photosynthesis.RUE.FixedValue =1.91"
                  ],
                  "Name": "Laila1",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.CompositeBiomass, Models",
              "OrganNames": [
                "Husk",
                "Rachis"
              ],
              "IncludeLive": True,
              "IncludeDead": False,
              "Name": "Spike",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Memo, Models",
                  "Text": "\nThe SpikeLive composite biomass object includes the Live pools of then non-grain organs within the reproductive spike to allow output for total dry matter and Nutrient contents, including total structural, non-structural and metabolic pools.\n",
                  "Name": "memo",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.Functions.Constant, Models",
              "FixedValue": 0.0,
              "Units": None,
              "Name": "MortalityRate",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.Functions.Constant, Models",
              "FixedValue": 0.0,
              "Units": None,
              "Name": "SeedMortalityRate",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            }
          ],
          "Enabled": True,
          "ReadOnly": False
        },
        {
          "$child": "Models.PMF.Plant, Models",
          "Name": "Wheat",
          "ResourceName": "Wheat",
          "Children": [
            {
              "$child": "Models.Memo, Models",
              "Text": "_Brown, H.E., Huth, N.I. and Holzworth, D.P._\n\nThe APSIM wheat model has been developed using the Plant Modelling Framework (PMF) of [brown_plant_2014]. This new framework provides a library of plant organ and process submodels that can be coupled, at runtime, to construct a model in much the same way that models can be coupled to construct a simulation. This means that dynamic composition of lower level process and organ classes (e.g. photosynthesis, leaf) into larger constructions (e.g. maize, wheat, sorghum) can be achieved by the model developer without additional coding.\r\n\r\nThe wheat model consists of:\r\n\r\n* a phenology model to simulate development through sequential developmental phases  \r\n* a structure model to simulate plant morphology \r\n* a collection of organs to simulate the various plant parts  \r\n* an arbitrator to allocate resources (N, biomass) to the various plant organs  \r\n\r\nThis work builds upon earlier APSIM Wheat models such as NWheat ([Asseng200225], [KeatingNWheat]), NWheatS ([Asseng1998163]), Cropmod-Wheat ([Wang2002GenericCropModel]), and the earlier versions developed in Plant (<a href=\\\"http://www.apsim.info/Documentation/Model,CropandSoil/CropModuleDocumentation/Wheat.aspx\">APSIM Wheat 7.5</a>\") and then within the Plant Modelling Framework ([brown_plant_2014]).",
              "Name": "Introduction",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.OrganArbitrator, Models",
              "Name": "Arbitrator",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.PMF.BiomassTypeArbitrator, Models",
                  "Name": "DMArbitration",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Core.Folder, Models",
                      "ShowInDocs": True,
                      "GraphsPerPage": 6,
                      "Name": "PotentialPartitioningMethods",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.PMF.Arbitrator.ReallocationMethod, Models",
                          "Name": "ReallocationMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.PMF.Arbitrator.AllocateFixationMethod, Models",
                          "Name": "AllocateFixationMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.PMF.Arbitrator.RetranslocationMethod, Models",
                          "Name": "RetranslocationMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.PMF.Arbitrator.SendPotentialDMAllocationsMethod, Models",
                          "Name": "SendPotentialDMAllocationsMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Core.Folder, Models",
                      "ShowInDocs": True,
                      "GraphsPerPage": 6,
                      "Name": "AllocationMethods",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.PMF.Arbitrator.NutrientConstrainedAllocationMethod, Models",
                          "Name": "NutrientConstrainedAllocationMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.PMF.Arbitrator.DryMatterAllocationsMethod, Models",
                          "Name": "DryMatterAllocationsMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.RelativeAllocation, Models",
                      "Name": "ArbitrationMethod",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.BiomassTypeArbitrator, Models",
                  "Name": "NArbitration",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Core.Folder, Models",
                      "ShowInDocs": True,
                      "GraphsPerPage": 6,
                      "Name": "PotentialPartitioningMethods",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.PMF.Arbitrator.ReallocationMethod, Models",
                          "Name": "ReallocationMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Core.Folder, Models",
                      "ShowInDocs": True,
                      "GraphsPerPage": 6,
                      "Name": "ActualPartitioningMethods",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.PMF.Arbitrator.AllocateFixationMethod, Models",
                          "Name": "AllocateFixationMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.PMF.Arbitrator.RetranslocationMethod, Models",
                          "Name": "RetranslocationMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Core.Folder, Models",
                      "ShowInDocs": True,
                      "GraphsPerPage": 6,
                      "Name": "AllocationMethods",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.PMF.Arbitrator.NitrogenAllocationsMethod, Models",
                          "Name": "NitrogenAllocationsMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.RelativeAllocation, Models",
                      "Name": "ArbitrationMethod",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Arbitrator.AllocateUptakesMethod, Models",
                      "Name": "AllocateUptakesMethod",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Arbitrator.WaterUptakeMethod, Models",
                  "Name": "WaterUptakeMethod",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Arbitrator.NitrogenUptakeMethod, Models",
                  "Name": "NitrogenUptakeMethod",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.Phen.Phenology, Models",
              "Name": "Phenology",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Memo, Models",
                  "Text": "Wheat exhibits a range of developmental responses to environment and these are strongly influences by genotype characteristics.\nTemperature effects development increasing development rates and decreasing phase durations as temperatures increase.  Theset affects are captured by thermal time, .  However, wheat also exhibits vernalisation and photoperiod sensitivities in its Vegetative phase and further photoperiod sensitivity in the EarlyReproducivePhase.  Photoperiod responses are seen as a reduction in the length of a phase for a photoperiod sensitive genotype in response to a longer photoperiod.  Vernalisation responses are more complicated as they are driven by temperature but interact with photoperiod.  For vernalisation sensitive varieties (Winter types) exposure to cool temperatures or short photoperiods during the Vegetative phase will reduce the thermal time duration of the vegetative phase.  \n\nWe draw on the Kirby Framework to capture these vernalisation and photoperiod responses.  This framework assumes that the timing of anthesis is a result of the timing of flag leaf and an additional thermal time passage from there to anthesis.  It also assumes the timing of flag leaf is a result of the Final Leaf Number which sets a target, and leaf appearance rate, which sets the rate of progress toward this target.  Leaf appearance rate is a function of Thermal time and a genotype specific Phyllochron which changes with Haun stage as described by [Jamieson_LeafAppearance_1995].    \n\nFinal Leaf Number (FLN) is modeled as the sum of three numbers:\n\nFLN = MinLeafNumber + VernalLeaves + PhotoLeaves\n\nWhere MinLeafNumber is the number of leaves that a wheat crop will produce when vernalisation is satisified early in the crops duration (before 2nd True leaf) and it is grown in a long photoperiod..  VernalLeaves are the number of leaves that are added due to vernalisation effects.  For insensitive varieties this will always be zero but this is potentially a larger number for sensitivie varieties and the number progresively decreases as the crop encounters more vernalisation.  PhotoLeaves are the number of leaves that are added to the minimum leaf number as a result of short day exposure.  For insensitive varieties this will be zero but is potential larger for more sensitivie varieties and decreases as day length increases.  More detailed explinations of the components of phenology are provided below.",
                  "Name": "Memo",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.HourlyInterpolation, Models",
                  "agregationMethod": 0,
                  "Name": "ThermalTime",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.XYPairs, Models",
                      "X": [
                        0.0,
                        18.0,
                        30.0
                      ],
                      "Y": [
                        0.0,
                        18.0,
                        0.0
                      ],
                      "XVariableName": None,
                      "Name": "Response",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "Thermal time determines the rate of developmental progress through many of the crops phases and is used by organs to determing potential growth rates.",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.ThreeHourSin, Models",
                      "TempRangeFactors": None,
                      "Name": "InterpolationMethod",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.MultiplyFunction, Models",
                  "Name": "DailyVernalisation",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "Vernalisation responses are based on those described by [Brown_Anthesis_2013].   Vernalisation is assumed to be related to the expression of the Vrn1 Gene.  Its experssion is accumulated daily and daily upregulation as a function of development (DeltaHaunStage) and a TempResponseProfile that declines exponentially from a maximum at 1^o^C to zero at 0^o^C and at 18^o^C.",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].HaunStage.Delta",
                      "Name": "DeltaHaunStage",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.HourlyInterpolation, Models",
                      "agregationMethod": 0,
                      "Name": "TempResponseProfile",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            1.0,
                            2.0,
                            3.0,
                            4.0,
                            5.0,
                            6.0,
                            7.0,
                            8.0,
                            9.0,
                            10.0,
                            11.0,
                            12.0,
                            13.0,
                            14.0,
                            15.0,
                            16.0,
                            17.0,
                            18.0,
                            19.0,
                            20.0,
                            30.0,
                            32.0
                          ],
                          "Y": [
                            0.0,
                            1.205,
                            1.037,
                            0.893,
                            0.768,
                            0.661,
                            0.569,
                            0.49,
                            0.422,
                            0.363,
                            0.312,
                            0.269,
                            0.231,
                            0.199,
                            0.171,
                            0.148,
                            0.127,
                            0.109,
                            0.094,
                            0.081,
                            0.0,
                            0.0,
                            -0.5
                          ],
                          "XVariableName": None,
                          "Name": "Response",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.ThreeHourSin, Models",
                          "TempRangeFactors": None,
                          "Name": "InterpolationMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhotoperiodFunction, Models",
                  "Twilight": -6.0,
                  "DayLength": 0.0,
                  "Name": "Photoperiod",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhaseLookup, Models",
                  "Name": "PerceivedPhotoPeriod",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.PhaseLookupValue, Models",
                      "Start": "Germination",
                      "End": "Emergence",
                      "Name": "ApexBelowGround",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 0.0,
                          "Units": "h",
                          "Name": "Photoperiod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.PhaseLookupValue, Models",
                      "Start": "Emergence",
                      "End": "HarvestRipe",
                      "Name": "ApexAboveGround",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Phenology].Photoperiod",
                          "Name": "Photoperiod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.MinimumFunction, Models",
                  "Name": "Vrn1",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "Vrn1 accumulation begins when Vrn4 experssion is down regulated to zero and stops, assuming vernalisation saturation, at a value of 1",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": "Relative Experssion",
                      "Name": "Saturation",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.AccumulateFunction, Models",
                      "StartStageName": "Germination",
                      "EndStageName": "TerminalSpikelet",
                      "ResetStageName": None,
                      "FractionRemovedOnCut": 0.0,
                      "FractionRemovedOnHarvest": 0.0,
                      "FractionRemovedOnGraze": 0.0,
                      "FractionRemovedOnPrune": 0.0,
                      "Name": "CurrentExpression",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.LessThanFunction, Models",
                          "Name": "Vernalisation",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Phenology].Vrn4",
                              "Name": "If_Vrn4",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Phenology].VernLag",
                              "Name": "IsLessThanLag",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 0.0,
                              "Units": None,
                              "Name": "Return_Zero",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Phenology].DailyVernalisation",
                              "Name": "elseReturn_Vernalisation",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.MinimumFunction, Models",
                  "Name": "Vrn4",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].VernLag",
                      "Name": "MaxExpression",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.AccumulateFunction, Models",
                      "StartStageName": "Germination",
                      "EndStageName": "TerminalSpikelet",
                      "ResetStageName": None,
                      "FractionRemovedOnCut": 0.0,
                      "FractionRemovedOnHarvest": 0.0,
                      "FractionRemovedOnGraze": 0.0,
                      "FractionRemovedOnPrune": 0.0,
                      "Name": "CurrentExpression",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Phenology].DailyVernalisation",
                          "Name": "VernalisationConditioning",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 7.0,
                  "Units": "Leaves",
                  "Name": "MinimumLeafNumber",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "Number of leaves the plant will produce when fully vernalised early and grown in long photoperiod\n",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.AccumulateFunction, Models",
                  "StartStageName": "Emergence",
                  "EndStageName": "FlagLeaf",
                  "ResetStageName": None,
                  "FractionRemovedOnCut": 0.0,
                  "FractionRemovedOnHarvest": 0.0,
                  "FractionRemovedOnGraze": 0.0,
                  "FractionRemovedOnPrune": 0.0,
                  "Name": "HaunStage",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.PhaseLookup, Models",
                      "Name": "Delta",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.PhaseLookupValue, Models",
                          "Start": "Germination",
                          "End": "HarvestRipe",
                          "Name": "Growing",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.DivideFunction, Models",
                              "Name": "Value",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Phenology].ThermalTime",
                                  "Name": "ThermalTime",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Phenology].Phyllochron",
                                  "Name": "Phyllochron",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.MultiplyFunction, Models",
                  "Name": "Phyllochron",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "This is the thermal time between the emergence of leaf tips.  The model used here is based on [Jamieson_SIRIUS_1998] where leaf appearace could be described by a base phyllochron determined between leaves 2 and 7 and a phyllochron that was 70% of base phyllochron for leaves < 2 and 130% of base phyllochron for leaves > 7",
                      "Name": "Rational",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "LeafStageFactor",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            2.0,
                            3.0,
                            7.0,
                            8.0,
                            11.0,
                            12.0
                          ],
                          "Y": [
                            0.75,
                            0.75,
                            1.0,
                            1.0,
                            1.4,
                            1.4,
                            1.4
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].AppearedCohortNo",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 120.0,
                      "Units": "oC.d",
                      "Name": "BasePhyllochron",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "PhotoPeriodEffect",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            8.0,
                            12.0,
                            14.0
                          ],
                          "Y": [
                            1.6,
                            1.0,
                            1.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Phenology].Photoperiod",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.ExpressionFunction, Models",
                  "Expression": "([Structure].FinalLeafNumber - 2.85)/1.1",
                  "Name": "HaunStageTerminalSpikelet",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "The Haun stage at which Terminal Spikelet occurs is determined from final leaf number using the approach described in [Brown_Anthesis_2013].",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.HoldFunction, Models",
                  "WhenToHold": "TerminalSpikelet",
                  "Name": "FinalLeafNumber",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.AddFunction, Models",
                      "Name": "FinalNodeNumber",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Phenology].MinimumLeafNumber",
                          "Name": "MinLeafNumber",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.MultiplyFunction, Models",
                          "Name": "VernalLeaves",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Phenology].VrnSensitivity",
                              "Name": "UnVernalisedLeaves",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.SubtractFunction, Models",
                              "Name": "VernalisationReductionFactor",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.Constant, Models",
                                  "FixedValue": 1.0,
                                  "Units": None,
                                  "Name": "One",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Phenology].Vrn1",
                                  "Name": "RelativeVrn1Expression",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.MultiplyFunction, Models",
                          "Name": "PhotoPLeaves",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Phenology].PpSensitivity",
                              "Name": "ShortPpLeaves",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.LinearInterpolationFunction, Models",
                              "Name": "PhotoPeriodReductionFactor",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.XYPairs, Models",
                                  "X": [
                                    10.0,
                                    16.0
                                  ],
                                  "Y": [
                                    1.0,
                                    0.0
                                  ],
                                  "XVariableName": None,
                                  "Name": "XYPairs",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Phenology].Photoperiod",
                                  "Name": "XValue",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": "Leaves",
                  "Name": "VrnSensitivity",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "The difference between Inherent earlyness leaf number and the number of leaves produced when the plant is unvernalised and grown in long photoperiod",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": "Unitless",
                  "Name": "VernLag",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "The amount of vernalising temperature required before vernalisation response will be evident.  Value of relative to the amount offl vernalisation temperature required from the end of the lag until vernalisation saturation.  A value of 0 means there is no lag, a value of 1 means the lag is the same (in vernal time) as the vernalisation requirement and a value of 2 means the lag is twice as long as the vernalisation response phase",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 3.0,
                  "Units": "Leaves",
                  "Name": "PpSensitivity",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "The reduction in leaf number going from 8 to 16 h Pp",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "HeadEmergencePpSensitivity",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "The phyllochrons duration for the plant to go from flag leaf ligual appearance at 16 h Pp compared to the phyllochron duration for the same phase at 8 h Pp.  ",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 2.5,
                  "Units": None,
                  "Name": "HeadEmergenceLongDayBase",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "The phyllochrons duration for the plant to go from flag leaf ligual appearance at 16 h Pp.",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.GerminatingPhase, Models",
                  "Start": "Sowing",
                  "End": "Germination",
                  "Name": "Germinating",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.EmergingPhase, Models",
                  "Start": "Germination",
                  "End": "Emergence",
                  "TTForTimeStep": 0.0,
                  "Name": "Emerging",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "ThermalTime",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.AddFunction, Models",
                      "Name": "Target",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 40.0,
                          "Units": None,
                          "Name": "ShootLag",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.MultiplyFunction, Models",
                          "Name": "DepthxRate",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Plant].SowingData.Depth",
                              "Name": "SowingDepth",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 1.5,
                              "Units": None,
                              "Name": "ShootRate",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.NodeNumberPhase, Models",
                  "Start": "Emergence",
                  "End": "VernalSaturation",
                  "Name": "Vernalising",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "ThermalTime",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].HaunStageTerminalSpikelet",
                      "Name": "CompletionNodeNumber",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Structure].LeafTipsAppeared",
                      "Name": "LeafTipNumber",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.GenericPhase, Models",
                  "Start": "VernalSaturation",
                  "End": "TerminalSpikelet",
                  "Name": "SpikeletDifferentiation",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "Progression",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Target",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.LeafAppearancePhase, Models",
                  "Start": "TerminalSpikelet",
                  "End": "FlagLeaf",
                  "Name": "StemElongation",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "The Final leaf number is fixed at Terminal Spikelet and leaves contune to appear at a rate set by thermal time and phyllochron until flag leaf liguale appears and this phase is completed.",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "ThermalTime",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Structure].FinalLeafNumber",
                      "Name": "FinalLeafNumber",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.ExpressionFunction, Models",
                      "Expression": "[Leaf].ExpandedCohortNo + [Leaf].NextExpandingLeafProportion",
                      "Name": "LeafNumber",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Leaf].ExpandedCohortNo",
                      "Name": "FullyExpandedLeafNo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Leaf].InitialisedCohortNo",
                      "Name": "InitialisedLeafNumber",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.GenericPhase, Models",
                  "Start": "FlagLeaf",
                  "End": "Heading",
                  "Name": "HeadEmergence",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "Progression",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Target",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Phenology].Phyllochron.BasePhyllochron",
                          "Name": "Phyllochron",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.AddFunction, Models",
                          "Name": "Phyllochrons",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Phenology].HeadEmergenceLongDayBase",
                              "Name": "LongDayBase",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.MultiplyFunction, Models",
                              "Name": "IncreaseDueToShortPhotoPeriod",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                                  "Name": "PhotoPeriodResponse",
                                  "ResourceName": None,
                                  "Children": [
                                    {
                                      "$child": "Models.Functions.XYPairs, Models",
                                      "X": [
                                        11.0,
                                        16.0
                                      ],
                                      "Y": [
                                        1.0,
                                        0.0
                                      ],
                                      "XVariableName": None,
                                      "Name": "XYPairs",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    },
                                    {
                                      "$child": "Models.Functions.VariableReference, Models",
                                      "VariableName": "[Phenology].Photoperiod",
                                      "Name": "XValue",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    }
                                  ],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Phenology].HeadEmergencePpSensitivity",
                                  "Name": "PpSensitivity",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.GenericPhase, Models",
                  "Start": "Heading",
                  "End": "Flowering",
                  "Name": "EarlyFlowering",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "Progression",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Target",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.GenericPhase, Models",
                  "Start": "Flowering",
                  "End": "StartGrainFill",
                  "Name": "GrainDevelopment",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 120.0,
                      "Units": None,
                      "Name": "Target",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "Progression",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.GenericPhase, Models",
                  "Start": "StartGrainFill",
                  "End": "EndGrainFill",
                  "Name": "GrainFilling",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 545.0,
                      "Units": None,
                      "Name": "Target",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "Progression",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.GenericPhase, Models",
                  "Start": "EndGrainFill",
                  "End": "Maturity",
                  "Name": "Maturing",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 35.0,
                      "Units": None,
                      "Name": "Target",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "Progression",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.GenericPhase, Models",
                  "Start": "Maturity",
                  "End": "HarvestRipe",
                  "Name": "Ripening",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 300.0,
                      "Units": None,
                      "Name": "Target",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "Progression",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.EndPhase, Models",
                  "Start": "HarvestRipe",
                  "End": "Unused",
                  "Name": "ReadyForHarvesting",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "ThermalTime",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.ZadokPMFWheat, Models",
                  "Name": "Zadok",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.OnEventFunction, Models",
                  "SetEvent": "TerminalSpikelet",
                  "ReSetEvent": "never",
                  "Name": "TerminalSpikeletDAS",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nA function is used to provide flowering date as days after sowing(DAS).\n",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "PreEventValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Plant].DaysAfterSowing",
                      "Name": "PostEventValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.OnEventFunction, Models",
                  "SetEvent": "FlagLeaf",
                  "ReSetEvent": "never",
                  "Name": "FlagLeafDAS",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nA function is used to provide flowering date as days after sowing(DAS).\n",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "PreEventValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Plant].DaysAfterSowing",
                      "Name": "PostEventValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.OnEventFunction, Models",
                  "SetEvent": "Heading",
                  "ReSetEvent": "never",
                  "Name": "HeadingDAS",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nA function is used to provide heading date as days after sowing(DAS).\n",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "PreEventValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Plant].DaysAfterSowing",
                      "Name": "PostEventValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.OnEventFunction, Models",
                  "SetEvent": "Flowering",
                  "ReSetEvent": "Germination",
                  "Name": "FloweringDAS",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nA function is used to provide flowering date as days after sowing(DAS).\n",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "PreEventValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Plant].DaysAfterSowing",
                      "Name": "PostEventValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.OnEventFunction, Models",
                  "SetEvent": "Maturity",
                  "ReSetEvent": "never",
                  "Name": "MaturityDAS",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nA function is used to provide maturity date as days after sowing(DAS).",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "PreEventValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Plant].DaysAfterSowing",
                      "Name": "PostEventValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.OnEventFunction, Models",
                  "SetEvent": "Emergence",
                  "ReSetEvent": "never",
                  "Name": "EmergenceDAS",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\nA function is used to provide flowering date as days after sowing(DAS).\n",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "PreEventValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Plant].DaysAfterSowing",
                      "Name": "PostEventValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.DivideFunction, Models",
                  "Name": "PTQ",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Radn_mol_per_m2",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[IWeather].Radn",
                          "Name": "SolarRadiation",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 2.285,
                          "Units": None,
                          "Name": "Constant",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].Photosynthesis.FT",
                          "Name": "RUE Temp factor",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Memo, Models",
                          "Text": "Multiply Radn (MJ/m2/d) by 4.57 to convert to mol/m2/day and divide by 2 to convert to mol PAR/m2/d",
                          "Name": "Memo",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "ThermalTime",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.Struct.Structure, Models",
              "CohortInitialisationStage": "Germination",
              "LeafInitialisationStage": "Emergence",
              "Name": "Structure",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Functions.VariableReference, Models",
                  "VariableName": "[Phenology].ThermalTime",
                  "Name": "ThermalTime",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.MultiplyFunction, Models",
                  "Name": "BranchingRate",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "Potential branching rate is determined by the commonly observed pattern of tillering in wheat, in which each tiller emerges with the third leaf on its parent axis (e.g. first tiller emerges at the same time as the third leaf on the main stem, the first secondary tiller appears with the third leaf on tiller 1).  This is described as a simple function of main stem leaf number.",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.PhaseLookup, Models",
                      "Name": "PotentialBranchingRate",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.PhaseLookupValue, Models",
                          "Start": "Emergence",
                          "End": "TerminalSpikelet",
                          "Name": "Vegetative",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.LinearInterpolationFunction, Models",
                              "Name": "PotentialBranchingRate",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.XYPairs, Models",
                                  "X": [
                                    1.0,
                                    2.0,
                                    3.0,
                                    4.0,
                                    5.0,
                                    6.0,
                                    7.0,
                                    8.0
                                  ],
                                  "Y": [
                                    0.0,
                                    0.0,
                                    1.0,
                                    2.0,
                                    4.0,
                                    7.0,
                                    12.0,
                                    20.0
                                  ],
                                  "XVariableName": None,
                                  "Name": "XYPairs",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Structure].LeafTipsAppeared",
                                  "Name": "XValue",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.PhaseLookupValue, Models",
                          "Start": "TerminalSpikelet",
                          "End": "HarvestRipe",
                          "Name": "Reproductive",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 0.0,
                              "Units": None,
                              "Name": "Zero",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.MinimumFunction, Models",
                      "Name": "StressFactors",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.LinearInterpolationFunction, Models",
                          "Name": "NitrogenEffect",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Memo, Models",
                              "Text": "Inadequate Nitrogen supply is assumed to affect tillering prior to any effect on photosynthesis or leaf size.",
                              "Name": "Memo",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.XYPairs, Models",
                              "X": [
                                0.5,
                                2.0,
                                3.0
                              ],
                              "Y": [
                                0.0,
                                1.0,
                                1.0
                              ],
                              "XVariableName": None,
                              "Name": "XYPairs",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Arbitrator].FN",
                              "Name": "XValue",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.LinearInterpolationFunction, Models",
                          "Name": "CoverEffect",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Memo, Models",
                              "Text": "Tillering is said to cease once a threshold fraction of incoming radiation has been achieved.  This captures shading effects on tillering described by [evers2006cessation].",
                              "Name": "Memo",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.XYPairs, Models",
                              "X": [
                                0.0,
                                0.1,
                                0.25
                              ],
                              "Y": [
                                1.0,
                                1.0,
                                0.0
                              ],
                              "XVariableName": None,
                              "Name": "XYPairs",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Leaf].CoverTotal",
                              "Name": "XValue",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.LinearInterpolationFunction, Models",
                          "Name": "WaterStressEffect",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.XYPairs, Models",
                              "X": [
                                0.0,
                                1.0
                              ],
                              "Y": [
                                0.0,
                                1.0
                              ],
                              "XVariableName": None,
                              "Name": "XYPairs",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Root].WaterTensionFactor",
                              "Name": "XValue",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhaseLookup, Models",
                  "Name": "BranchMortality",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.PhaseLookupValue, Models",
                      "Start": "FlagLeaf",
                      "End": "Flowering",
                      "Name": "MortalityPhase",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.MultiplyFunction, Models",
                          "Name": "Mortality",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.LinearInterpolationFunction, Models",
                              "Name": "MortalityPerDegDay",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.XYPairs, Models",
                                  "X": [
                                    0.0,
                                    0.005
                                  ],
                                  "Y": [
                                    0.002,
                                    0.0
                                  ],
                                  "XVariableName": None,
                                  "Name": "XYPairs",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Structure].MeanTillerGrowthRate",
                                  "Name": "XValue",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Phenology].ThermalTime",
                              "Name": "ThermalTime",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Struct.HeightFunction, Models",
                  "Name": "HeightModel",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "PotentialHeight",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            3.0,
                            4.0,
                            5.0,
                            6.0,
                            7.0,
                            8.0,
                            9.0
                          ],
                          "Y": [
                            10.0,
                            200.0,
                            200.0,
                            1000.0,
                            1000.0,
                            1000.0,
                            1000.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Phenology].Stage",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "WaterStress",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.3,
                            1.0
                          ],
                          "Y": [
                            0.0,
                            1.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].Fw",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.MovingAverageFunction, Models",
                  "NumberOfDays": 5,
                  "StageToStartMovingAverage": "Emergence",
                  "Name": "MeanTillerGrowthRate",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "This is calculated to represent the supply of asslimilate to individual tillers over the past 5 days to use as an index to determine tiller mortality.  When tiller numbers are high or total crop assimilate supply is small there will not be enough assimilate to maintain all of the tillers so the model will senesce some. ",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.DivideFunction, Models",
                      "Name": "TillerGrowthRate",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Arbitrator].DM.TotalFixationSupply",
                          "Name": "DailyGrowth",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Phenology].ThermalTime",
                          "Name": "ThermalTime",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Structure].TotalStemPopn",
                          "Name": "StemPopulation",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.MovingAverageFunction, Models",
                  "NumberOfDays": 300,
                  "StageToStartMovingAverage": "Emergence",
                  "Name": "MeanPhyllochron",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].Phyllochron",
                      "Name": "Phyllochron",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "StemSenescenceAge",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.021,
                  "Units": None,
                  "Name": "MaxLAR",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.VariableReference, Models",
                  "VariableName": "[Phenology].FinalLeafNumber",
                  "Name": "FinalLeafNumber",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.VariableReference, Models",
                  "VariableName": "[Phenology].Phyllochron",
                  "Name": "Phyllochron",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.Organs.ReproductiveOrgan, Models",
              "GrowthRespiration": 0.0,
              "MaintenanceRespiration": 0.0,
              "DMDemand": None,
              "NDemand": None,
              "DMSupply": None,
              "NSupply": None,
              "RipeStage": "Ripe",
              "potentialDMAllocation": None,
              "Live": {
                "$child": "Models.PMF.Biomass, Models",
                "Name": "Biomass",
                "ResourceName": None,
                "Children": [],
                "Enabled": True,
                "ReadOnly": False
              },
              "Dead": {
                "$child": "Models.PMF.Biomass, Models",
                "Name": "Biomass",
                "ResourceName": None,
                "Children": [],
                "Enabled": True,
                "ReadOnly": False
              },
              "Name": "Grain",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Functions.HoldFunction, Models",
                  "WhenToHold": "Flowering",
                  "Name": "NumberFunction",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "GrainNumber",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 26.0,
                          "Units": "grains",
                          "Name": "GrainsPerGramOfStem",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[StemPlusSpike].Wt",
                          "Name": "StemMass",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": None,
                  "Name": "DMConversionEfficiency",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "RemobilisationCost",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.05,
                  "Units": None,
                  "Name": "InitialGrainProportion",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.05,
                  "Units": "g",
                  "Name": "MaximumPotentialGrainSize",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhaseLookup, Models",
                  "Name": "DMDemandFunction",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.PhaseLookupValue, Models",
                      "Start": "Flowering",
                      "End": "StartGrainFill",
                      "Name": "InitialPhase",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.DemandFunctions.FillingRateFunction, Models",
                          "Name": "FillingRateFunction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Grain].NumberFunction",
                              "Name": "NumberFunction",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Phenology].GrainDevelopment.Target",
                              "Name": "FillingDuration",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Phenology].ThermalTime",
                              "Name": "ThermalTime",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.MultiplyFunction, Models",
                              "Name": "PotentialSizeIncrement",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Grain].InitialGrainProportion",
                                  "Name": "InitialProportion",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Grain].MaximumPotentialGrainSize",
                                  "Name": "MaximumSize",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.PhaseLookupValue, Models",
                      "Start": "StartGrainFill",
                      "End": "EndGrainFill",
                      "Name": "LinearPhase",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.DemandFunctions.FillingRateFunction, Models",
                          "Name": "FillingRateFunction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Grain].NumberFunction",
                              "Name": "NumberFunction",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Phenology].GrainFilling.Target",
                              "Name": "FillingDuration",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Phenology].ThermalTime",
                              "Name": "ThermalTime",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.MultiplyFunction, Models",
                              "Name": "PotentialSizeIncrement",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.SubtractFunction, Models",
                                  "Name": "ProportionLinearPhase",
                                  "ResourceName": None,
                                  "Children": [
                                    {
                                      "$child": "Models.Functions.Constant, Models",
                                      "FixedValue": 1.0,
                                      "Units": None,
                                      "Name": "One",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    },
                                    {
                                      "$child": "Models.Functions.VariableReference, Models",
                                      "VariableName": "[Grain].InitialGrainProportion",
                                      "Name": "InitialProportion",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    }
                                  ],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Grain].MaximumPotentialGrainSize",
                                  "Name": "MaximumSize",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0123,
                  "Units": None,
                  "Name": "MinimumNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.03,
                  "Units": None,
                  "Name": "MaxNConcDailyGrowth",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.03,
                  "Units": None,
                  "Name": "MaximumNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhaseLookup, Models",
                  "Name": "NFillingRate",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.PhaseLookupValue, Models",
                      "Start": "Flowering",
                      "End": "EndGrainFill",
                      "Name": "GrainFilling",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.DemandFunctions.FillingRateFunction, Models",
                          "Name": "FillingRateFunction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Grain].NumberFunction",
                              "Name": "NumberFunction",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.AddFunction, Models",
                              "Name": "FillingDuration",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Phenology].GrainDevelopment.Target",
                                  "Name": "FloweringToGrainFilling",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Phenology].GrainFilling.Target",
                                  "Name": "GrainFilling",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Phenology].ThermalTime",
                              "Name": "ThermalTime",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.MultiplyFunction, Models",
                              "Name": "PotentialSizeIncrement",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Grain].MaximumPotentialGrainSize",
                                  "Name": "MaximumSize",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Grain].MaximumNConc",
                                  "Name": "MaximumNConc",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.12,
                  "Units": None,
                  "Name": "WaterContent",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.AccumulateFunction, Models",
                  "StartStageName": "Flowering",
                  "EndStageName": "Maturity",
                  "ResetStageName": None,
                  "FractionRemovedOnCut": 0.0,
                  "FractionRemovedOnHarvest": 0.0,
                  "FractionRemovedOnGraze": 0.0,
                  "FractionRemovedOnPrune": 0.0,
                  "Name": "AccumThermalTime",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "DailyThermalTimeValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.ExpressionFunction, Models",
                  "Expression": "([Grain].Live.N + [Grain].Dead.N)/([Grain].Live.Wt + [Grain].Dead.Wt) * 100 * 5.71",
                  "Name": "Protein",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Library.BiomassRemoval, Models",
                  "Name": "BiomassRemovalDefaults",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 1.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.0,
                      "FractionDeadToResidue": 1.0,
                      "Name": "Harvest",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 1.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.0,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Cut",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.8,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Prune",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.6,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.2,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Graze",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.05,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Thin",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.4,
                  "Units": None,
                  "Name": "CarbonConcentration",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientPoolFunctions, Models",
                  "Name": "DMDemandPriorityFactors",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientPoolFunctions, Models",
                  "Name": "NDemandPriorityFactors",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.Organs.Root, Models",
              "DMSupply": None,
              "NSupply": None,
              "DMDemand": None,
              "NDemand": None,
              "potentialDMAllocation": None,
              "GrowthRespiration": 0.0,
              "MaintenanceRespiration": 0.0,
              "Name": "Root",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Functions.RootShape.RootShapeCylinder, Models",
                  "Name": "RootShape",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": None,
                  "Name": "DMConversionEfficiency",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "RemobilisationCost",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": None,
                  "Name": "KLModifier",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": None,
                  "Name": "SoilWaterEffect",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 20.0,
                  "Units": None,
                  "Name": "MaxDailyNUptake",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 20.0,
                      "Units": None,
                      "Name": "MaxDailyNUptake",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.005,
                  "Units": None,
                  "Name": "SenescenceRate",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1000000.0,
                  "Units": None,
                  "Name": "MaximumRootDepth",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.01,
                  "Units": None,
                  "Name": "MaximumNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.001,
                  "Units": None,
                  "Name": "MinimumNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhaseLookupValue, Models",
                  "Start": "Emergence",
                  "End": "EndGrainFill",
                  "Name": "NitrogenDemandSwitch",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Constant",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Library.BiomassRemoval, Models",
                  "Name": "BiomassRemovalDefaults",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 1.0,
                      "FractionDeadToResidue": 1.0,
                      "Name": "Harvest",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.3,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Cut",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.1,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Prune",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.15,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Graze",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.05,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Thin",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.02,
                  "Units": None,
                  "Name": "KNO3",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.01,
                  "Units": None,
                  "Name": "KNH4",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 105.0,
                  "Units": "m/g",
                  "Name": "SpecificRootLength",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.MultiplyFunction, Models",
                  "Name": "RootFrontVelocity",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.PhaseLookup, Models",
                      "Name": "PotentialRootFrontVelocity",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.PhaseLookupValue, Models",
                          "Start": "Germination",
                          "End": "Emergence",
                          "Name": "PreEmergence",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 5.0,
                              "Units": "mm/d",
                              "Name": "Value",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.PhaseLookupValue, Models",
                          "Start": "Emergence",
                          "End": "Maturity",
                          "Name": "PostEmergence",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 20.0,
                              "Units": "mm/d",
                              "Name": "Value",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.WeightedTemperatureFunction, Models",
                      "MaximumTemperatureWeighting": 0.5,
                      "Name": "TemperatureFactor",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            15.0,
                            25.0,
                            35.0
                          ],
                          "Y": [
                            0.0,
                            1.0,
                            1.0,
                            0.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "WaterFactor",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SoilWaterScale, Models",
                          "Name": "SoilWaterScale",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            0.25,
                            1.0
                          ],
                          "Y": [
                            0.0,
                            1.0,
                            1.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[SoilWaterScale]",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                  "Name": "NUptakeSWFactor",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.SoilWaterScale, Models",
                      "Name": "SoilWaterScale",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.XYPairs, Models",
                      "X": [
                        0.0,
                        1.0
                      ],
                      "Y": [
                        0.0,
                        1.5
                      ],
                      "XVariableName": None,
                      "Name": "XYPairs",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "This is modelled in the same way as the old wheat model where potential N uptake is decreased as the soil dries as described by NUptakeSWFactor",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[SoilWaterScale]",
                      "Name": "XValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.42,
                  "Units": None,
                  "Name": "CarbonConcentration",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "MaintenanceRespirationFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientDemandFunctions, Models",
                  "Name": "DMDemands",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.DemandFunctions.PartitionFractionDemandFunction, Models",
                          "Name": "DMDemandFunction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.PhaseLookup, Models",
                              "Name": "PartitionFraction",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.PhaseLookupValue, Models",
                                  "Start": "Germination",
                                  "End": "Emergence",
                                  "Name": "PreEmergence",
                                  "ResourceName": None,
                                  "Children": [
                                    {
                                      "$child": "Models.Functions.Constant, Models",
                                      "FixedValue": 0.0,
                                      "Units": None,
                                      "Name": "Value",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    }
                                  ],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.PhaseLookupValue, Models",
                                  "Start": "Emergence",
                                  "End": "Flowering",
                                  "Name": "PreFlowering",
                                  "ResourceName": None,
                                  "Children": [
                                    {
                                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                                      "Name": "AgeFactor",
                                      "ResourceName": None,
                                      "Children": [
                                        {
                                          "$child": "Models.Functions.XYPairs, Models",
                                          "X": [
                                            3.0,
                                            4.0,
                                            5.0
                                          ],
                                          "Y": [
                                            0.5,
                                            0.2,
                                            0.2
                                          ],
                                          "XVariableName": None,
                                          "Name": "XYPairs",
                                          "ResourceName": None,
                                          "Children": [],
                                          "Enabled": True,
                                          "ReadOnly": False
                                        },
                                        {
                                          "$child": "Models.Functions.VariableReference, Models",
                                          "VariableName": "[Phenology].Stage",
                                          "Name": "XValue",
                                          "ResourceName": None,
                                          "Children": [],
                                          "Enabled": True,
                                          "ReadOnly": False
                                        }
                                      ],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    }
                                  ],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.PhaseLookupValue, Models",
                                  "Start": "Flowering",
                                  "End": "EndGrainFill",
                                  "Name": "PostFlowering",
                                  "ResourceName": None,
                                  "Children": [
                                    {
                                      "$child": "Models.Functions.Constant, Models",
                                      "FixedValue": 0.2,
                                      "Units": None,
                                      "Name": "Value",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    }
                                  ],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 1.0,
                          "Units": None,
                          "Name": "StructuralFraction",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.DemandFunctions.StorageDMDemandFunction, Models",
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "StorageFraction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 1.0,
                              "Units": None,
                              "Name": "One",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Root].DMDemands.Structural.StructuralFraction",
                              "Name": "StructuralFraction",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStructuralPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QMetabolicPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStoragePriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientDemandFunctions, Models",
                  "Name": "NDemands",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Root].minimumNconc",
                          "Name": "MinNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Root].potentialDMAllocation.Structural",
                          "Name": "PotentialDMAllocation",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "MetabolicNconc",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Root].criticalNConc",
                              "Name": "CritNconc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Root].minimumNconc",
                              "Name": "MinNconc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Root].potentialDMAllocation.Structural",
                          "Name": "PotentialDMAllocation",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.DemandFunctions.StorageNDemandFunction, Models",
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Root].nitrogenDemandSwitch",
                          "Name": "NitrogenDemandSwitch",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Root].maximumNconc",
                          "Name": "MaxNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStructuralPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QMetabolicPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStoragePriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.VariableReference, Models",
                  "VariableName": "[Root].MinimumNConc",
                  "Name": "CriticalNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientPoolFunctions, Models",
                  "Name": "InitialWt",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.005,
                      "Units": "g/plant",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.Organs.Leaf, Models",
              "Leaves": [],
              "CurrentExpandingLeaf": 0.0,
              "StartFractionExpanded": 0.0,
              "FractionNextleafExpanded": 0.0,
              "DeadNodesYesterday": 0.0,
              "MaxCover": 1.0,
              "GrowthRespiration": 0.0,
              "DMSupply": None,
              "NSupply": None,
              "DMDemand": None,
              "NDemand": None,
              "potentialDMAllocation": None,
              "Albedo": 0.25,
              "Gsmax350": 0.011,
              "R50": 150.0,
              "LAI": 0.0,
              "Depth": 0.0,
              "Width": 0.0,
              "FRGR": 0.0,
              "PotentialEP": 0.0,
              "WaterDemand": 0.0,
              "LightProfile": None,
              "KDead": 0.2,
              "MaximumMainStemLeafNumber": 30,
              "TipsAtEmergence": 0,
              "CohortsAtInitialisation": 0,
              "FractionDied": 0.0,
              "Name": "Leaf",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Functions.SupplyFunctions.StomatalConductanceCO2Modifier, Models",
                  "Name": "StomatalConductanceCO2Modifier",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Leaf].Photosynthesis.FCO2",
                      "Name": "PhotosynthesisCO2Modifier",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Organs.LeafCohort, Models",
                  "ApexCohort": None,
                  "LiveStart": None,
                  "NReallocationFactor": 0.0,
                  "DMReallocationFactor": 0.0,
                  "NRetranslocationFactor": 0.0,
                  "DMRetranslocationFactor": 0.0,
                  "ShadeInducedSenRate": 0.0,
                  "SenescedFrac": 0.0,
                  "DeltaPotentialArea": 0.0,
                  "DeltaStressConstrainedArea": 0.0,
                  "DeltaCarbonConstrainedArea": 0.0,
                  "PotentialStructuralDMAllocation": 0.0,
                  "PotentialMetabolicDMAllocation": 0.0,
                  "MetabolicNReallocated": 0.0,
                  "MetabolicWtReallocated": 0.0,
                  "StorageNReallocated": 0.0,
                  "StorageWtReallocated": 0.0,
                  "MetabolicNRetranslocated": 0.0,
                  "StorageNRetrasnlocated": 0.0,
                  "DMRetranslocated": 0.0,
                  "MetabolicNAllocation": 0.0,
                  "StructuralDMAllocation": 0.0,
                  "MetabolicDMAllocation": 0.0,
                  "Rank": 1,
                  "Area": 200.0,
                  "Name": "InitialLeaves[1]",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Organs.LeafCohort, Models",
                  "ApexCohort": None,
                  "LiveStart": None,
                  "NReallocationFactor": 0.0,
                  "DMReallocationFactor": 0.0,
                  "NRetranslocationFactor": 0.0,
                  "DMRetranslocationFactor": 0.0,
                  "ShadeInducedSenRate": 0.0,
                  "SenescedFrac": 0.0,
                  "DeltaPotentialArea": 0.0,
                  "DeltaStressConstrainedArea": 0.0,
                  "DeltaCarbonConstrainedArea": 0.0,
                  "PotentialStructuralDMAllocation": 0.0,
                  "PotentialMetabolicDMAllocation": 0.0,
                  "MetabolicNReallocated": 0.0,
                  "MetabolicWtReallocated": 0.0,
                  "StorageNReallocated": 0.0,
                  "StorageWtReallocated": 0.0,
                  "MetabolicNRetranslocated": 0.0,
                  "StorageNRetrasnlocated": 0.0,
                  "DMRetranslocated": 0.0,
                  "MetabolicNAllocation": 0.0,
                  "StructuralDMAllocation": 0.0,
                  "MetabolicDMAllocation": 0.0,
                  "Rank": 2,
                  "Area": 0.0,
                  "Name": "InitialLeaves[2]",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Organs.Leaf+LeafCohortParameters, Models",
                  "ExpansionStressValue": 0.0,
                  "CellDivisionStressValue": 0.0,
                  "LagAccelerationValue": 0.0,
                  "SenescenceAccelerationValue": 0.0,
                  "ShadeInducedSenescenceRateValue": 0.0,
                  "SenessingLeafRelativeSizeValue": 0.0,
                  "Name": "CohortParameters",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "NReallocationFactor",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "RemobilisationCost",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "MaintenanceRespirationFunction",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.03,
                      "Units": None,
                      "Name": "NRetranslocationFactor",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "DMReallocationFactor",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "DMRetranslocationFactor",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.MinimumFunction, Models",
                      "Name": "CellDivisionStress",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.LinearInterpolationFunction, Models",
                          "Name": "WaterStressEffect",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.XYPairs, Models",
                              "X": [
                                0.5,
                                1.0
                              ],
                              "Y": [
                                0.1,
                                1.0
                              ],
                              "XVariableName": None,
                              "Name": "XYPairs",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Leaf].Fw",
                              "Name": "XValue",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.LinearInterpolationFunction, Models",
                          "Name": "NitrogenStressEffect",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.XYPairs, Models",
                              "X": [
                                0.0,
                                0.5,
                                1.0
                              ],
                              "Y": [
                                0.1,
                                0.1,
                                1.0
                              ],
                              "XVariableName": None,
                              "Name": "XYPairs",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Wheat].Leaf.Fn",
                              "Name": "XValue",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.MinimumFunction, Models",
                      "Name": "ExpansionStress",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.LinearInterpolationFunction, Models",
                          "Name": "WaterStressEffect",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.XYPairs, Models",
                              "X": [
                                0.1,
                                1.1,
                                1.3
                              ],
                              "Y": [
                                0.0,
                                1.0,
                                1.0
                              ],
                              "XVariableName": None,
                              "Name": "XYPairs",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Root].WaterTensionFactor",
                              "Name": "XValue",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.LinearInterpolationFunction, Models",
                          "Name": "TemperatureEffect",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.XYPairs, Models",
                              "X": [
                                0.0,
                                12.0,
                                14.0
                              ],
                              "Y": [
                                0.0,
                                1.0,
                                1.0
                              ],
                              "XVariableName": None,
                              "Name": "XYPairs",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[IWeather].MeanT",
                              "Name": "XValue",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.LinearInterpolationFunction, Models",
                          "Name": "NitrogenStressEffect",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.XYPairs, Models",
                              "X": [
                                0.0,
                                0.5,
                                1.0
                              ],
                              "Y": [
                                0.1,
                                0.1,
                                1.0
                              ],
                              "XVariableName": None,
                              "Name": "XYPairs",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Wheat].Leaf.Fn",
                              "Name": "XValue",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "MaxArea",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 2600.0,
                          "Units": "mm^2",
                          "Name": "AreaLargestLeaves",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.LinearInterpolationFunction, Models",
                          "Name": "AgeFactor",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.XYPairs, Models",
                              "X": [
                                3.0,
                                3.45,
                                4.0,
                                5.0
                              ],
                              "Y": [
                                0.1,
                                0.5,
                                1.0,
                                1.0
                              ],
                              "XVariableName": None,
                              "Name": "XYPairs",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Phenology].Stage",
                              "Name": "XValue",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "GrowthDuration",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 1.3,
                          "Units": None,
                          "Name": "Constant",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Phenology].Phyllochron",
                          "Name": "Phyllochron",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "LagDuration",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.LinearInterpolationFunction, Models",
                          "Name": "AgeFactor",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.XYPairs, Models",
                              "X": [
                                3.0,
                                4.0,
                                5.0
                              ],
                              "Y": [
                                0.4,
                                1.0,
                                1.0
                              ],
                              "XVariableName": None,
                              "Name": "XYPairs",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Phenology].Stage",
                              "Name": "XValue",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "LastLeafDuration",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.AddFunction, Models",
                              "Name": "ThermalTimeToRipe",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Phenology].HeadEmergence.Target",
                                  "Name": "HeadEmergence",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Phenology].EarlyFlowering.Target",
                                  "Name": "EarlyFlowering",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Phenology].GrainDevelopment.Target",
                                  "Name": "GrainDevelopment",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Phenology].GrainFilling.Target",
                                  "Name": "GrainFilling",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Phenology].Maturing.Target",
                                  "Name": "Maturing",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Phenology].Ripening.Target",
                                  "Name": "Ripening",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Leaf].CohortParameters.SenescenceDuration",
                              "Name": "SenescenceDuration",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "SenescenceDuration",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 3.0,
                          "Units": None,
                          "Name": "Constant",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Phenology].Phyllochron",
                          "Name": "Phyllochron",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1000000.0,
                      "Units": None,
                      "Name": "DetachmentLagDuration",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1000000.0,
                      "Units": None,
                      "Name": "DetachmentDuration",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "RelativeBranchLeafSize",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "MaximumNConc",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            3.0,
                            4.0,
                            5.0,
                            6.0,
                            7.0,
                            8.0
                          ],
                          "Y": [
                            0.05,
                            0.05,
                            0.05,
                            0.035,
                            0.02,
                            0.02
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Phenology].Stage",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "CriticalNConc",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.LinearInterpolationFunction, Models",
                          "Name": "CriticalNConcAt350ppm",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.XYPairs, Models",
                              "X": [
                                3.0,
                                4.0,
                                5.0,
                                6.0,
                                7.0,
                                8.0
                              ],
                              "Y": [
                                0.04,
                                0.04,
                                0.04,
                                0.0275,
                                0.015,
                                0.015
                              ],
                              "XVariableName": None,
                              "Name": "XYPairs",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Phenology].Stage",
                              "Name": "XValue",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.LinearInterpolationFunction, Models",
                          "Name": "CO2Factor",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.XYPairs, Models",
                              "X": [
                                350.0,
                                700.0
                              ],
                              "Y": [
                                1.0,
                                0.93
                              ],
                              "XVariableName": None,
                              "Name": "XYPairs",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[IWeather].CO2",
                              "Name": "XValue",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "MinimumNConc",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            3.0,
                            4.0,
                            5.0,
                            6.0,
                            7.0,
                            8.0
                          ],
                          "Y": [
                            0.01,
                            0.01,
                            0.01,
                            0.01,
                            0.01,
                            0.01
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Phenology].Stage",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.AddFunction, Models",
                      "Name": "LagAcceleration",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 1.0,
                          "Units": None,
                          "Name": "One",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.MultiplyFunction, Models",
                          "Name": "Stress",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 1.0,
                              "Units": None,
                              "Name": "StressResponseCoefficient",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.LinearInterpolationFunction, Models",
                              "Name": "StressFactor",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.XYPairs, Models",
                                  "X": [
                                    0.0,
                                    1.0
                                  ],
                                  "Y": [
                                    2.0,
                                    0.0
                                  ],
                                  "XVariableName": None,
                                  "Name": "XYPairs",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Leaf].Fw",
                                  "Name": "XValue",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.AddFunction, Models",
                      "Name": "SenescenceAcceleration",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 1.0,
                          "Units": None,
                          "Name": "One",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.MultiplyFunction, Models",
                          "Name": "Stress",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 1.0,
                              "Units": None,
                              "Name": "StressResponseCoefficient",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.LinearInterpolationFunction, Models",
                              "Name": "StressFactor",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.XYPairs, Models",
                                  "X": [
                                    0.0,
                                    1.0
                                  ],
                                  "Y": [
                                    2.0,
                                    0.0
                                  ],
                                  "XVariableName": None,
                                  "Name": "XYPairs",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Leaf].Fw",
                                  "Name": "XValue",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "SpecificLeafAreaMax",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            3.0,
                            4.0,
                            5.0,
                            6.0,
                            7.0,
                            8.0
                          ],
                          "Y": [
                            17000.0,
                            20000.0,
                            20000.0,
                            18500.0,
                            17000.0,
                            17000.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Phenology].Stage",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "SpecificLeafAreaMin",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.4,
                            1.0
                          ],
                          "Y": [
                            14000.0,
                            14000.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].Fn",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "InitialNConc",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.5,
                      "Units": None,
                      "Name": "StructuralFraction",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.3,
                      "Units": None,
                      "Name": "StorageFraction",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.3,
                      "Units": None,
                      "Name": "LeafSizeShapeParameter",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": "0-1",
                      "Name": "SenessingLeafRelativeSize",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "ShadeInducedSenescenceRate",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.ArrayFunction, Models",
                      "Values": "1 1 1",
                      "Units": None,
                      "Name": "LagDurationAgeMultiplier",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.ArrayFunction, Models",
                      "Values": "1 1 1",
                      "Units": None,
                      "Name": "SenescenceDurationAgeMultiplier",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.ArrayFunction, Models",
                      "Values": "1 1 1 1 1 1 1 1 1 1 1 1",
                      "Units": None,
                      "Name": "LeafSizeAgeMultiplier",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.SupplyFunctions.RUEModel, Models",
                  "Name": "Photosynthesis",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.5,
                      "Units": None,
                      "Name": "RUE",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.WeightedTemperatureFunction, Models",
                      "MaximumTemperatureWeighting": 0.75,
                      "Name": "FT",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            15.0,
                            25.0,
                            35.0
                          ],
                          "Y": [
                            0.0,
                            1.0,
                            1.0,
                            0.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "FN",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            1.0,
                            1.5
                          ],
                          "Y": [
                            0.0,
                            1.0,
                            1.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].Fn",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "FW",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            1.0,
                            1.0
                          ],
                          "Y": [
                            0.0,
                            1.0,
                            1.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].Fw",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "FVPD",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.DailyMeanVPD, Models",
                          "MaximumVPDWeight": 0.66,
                          "Name": "DailyVPD",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            10.0,
                            40.0
                          ],
                          "Y": [
                            1.0,
                            1.0,
                            0.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[DailyVPD]",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.SupplyFunctions.RUECO2Function, Models",
                      "PhotosyntheticPathway": "C3",
                      "Name": "FCO2",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Leaf].RadiationIntercepted",
                      "Name": "RadnInt",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "FrostFraction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "RemobilisationCost",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.VariableReference, Models",
                  "VariableName": "[Phenology].ThermalTime",
                  "Name": "ThermalTime",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.MinimumFunction, Models",
                  "Name": "FRGRFunction",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.WeightedTemperatureFunction, Models",
                      "MaximumTemperatureWeighting": 0.75,
                      "Name": "FT",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            15.0,
                            25.0,
                            35.0
                          ],
                          "Y": [
                            0.0,
                            1.0,
                            1.0,
                            0.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "FN",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            1.0,
                            1.5
                          ],
                          "Y": [
                            0.0,
                            1.0,
                            1.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].Fn",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "FVPD",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.DailyMeanVPD, Models",
                          "MaximumVPDWeight": 0.66,
                          "Name": "DailyVPD",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            10.0,
                            40.0
                          ],
                          "Y": [
                            1.0,
                            1.0,
                            0.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[DailyVPD]",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Library.BiomassRemoval, Models",
                  "Name": "BiomassRemovalDefaults",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 1.0,
                      "FractionDeadToResidue": 1.0,
                      "Name": "Harvest",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.8,
                      "FractionDeadToRemove": 0.8,
                      "FractionLiveToResidue": 0.0,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Cut",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.6,
                      "FractionDeadToResidue": 0.6,
                      "Name": "Prune",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.6,
                      "FractionDeadToRemove": 0.6,
                      "FractionLiveToResidue": 0.1,
                      "FractionDeadToResidue": 0.1,
                      "Name": "Graze",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.05,
                      "FractionDeadToResidue": 0.05,
                      "Name": "Thin",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.MultiplyFunction, Models",
                  "Name": "ExtinctionCoeff",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.5,
                      "Units": None,
                      "Name": "VegetativePhase",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "DevelopmentFactor",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            3.0,
                            4.0,
                            5.0,
                            5.7,
                            6.0
                          ],
                          "Y": [
                            1.0,
                            1.0,
                            1.0,
                            1.0,
                            1.2
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Memo, Models",
                          "Text": "Exntiction coefficient increases at stem extension as leaves become more prostrate",
                          "Name": "Memo",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Phenology].Stage",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.5,
                  "Units": None,
                  "Name": "StructuralFraction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": None,
                  "Name": "DMConversionEfficiency",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.4,
                  "Units": None,
                  "Name": "CarbonConcentration",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "WidthFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.VariableReference, Models",
                  "VariableName": "[Leaf].Height",
                  "Name": "DepthFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientPoolFunctions, Models",
                  "Name": "DMDemandPriorityFactors",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientPoolFunctions, Models",
                  "Name": "NDemandPriorityFactors",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.Organs.GenericOrgan, Models",
              "StartLive": None,
              "DMSupply": None,
              "NSupply": None,
              "DMDemand": None,
              "NDemand": None,
              "potentialDMAllocation": None,
              "IsAboveGround": True,
              "Name": "Spike",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "NReallocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "RemobilisationCost",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhaseLookup, Models",
                  "Name": "NRetranslocationFactor",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.PhaseLookupValue, Models",
                      "Start": "Emergence",
                      "End": "StartGrainFill",
                      "Name": "EarlyGrowth",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 0.0,
                          "Units": None,
                          "Name": "ValueDuringEarlyGrowth",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.PhaseLookupValue, Models",
                      "Start": "StartGrainFill",
                      "End": "EndGrainFill",
                      "Name": "ReproductiveGrowth",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 0.5,
                          "Units": None,
                          "Name": "ValueDuringGrainFill",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                  "Name": "MaximumNConc",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.XYPairs, Models",
                      "X": [
                        3.0,
                        4.0,
                        5.0,
                        5.5,
                        6.5,
                        7.0,
                        8.0
                      ],
                      "Y": [
                        0.01,
                        0.01,
                        0.01,
                        0.01,
                        0.025,
                        0.025,
                        0.025
                      ],
                      "XVariableName": None,
                      "Name": "XYPairs",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].Stage",
                      "Name": "XValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.004,
                  "Units": None,
                  "Name": "MinimumNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Library.BiomassRemoval, Models",
                  "Name": "BiomassRemovalDefaults",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.5,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.1,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Harvest",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.8,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.0,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Cut",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.6,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Prune",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.6,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.2,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Graze",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.05,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Thin",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhaseLookupValue, Models",
                  "Start": "Emergence",
                  "End": "StartGrainFill",
                  "Name": "NitrogenDemandSwitch",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Constant",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhaseLookup, Models",
                  "Name": "DMRetranslocationFactor",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.PhaseLookupValue, Models",
                      "Start": "Emergence",
                      "End": "StartGrainFill",
                      "Name": "VegetativeGrowth",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 0.0,
                          "Units": None,
                          "Name": "DMRetranslocationFactor",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.PhaseLookupValue, Models",
                      "Start": "StartGrainFill",
                      "End": "EndGrainFill",
                      "Name": "ReproductiveGrowth",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 0.5,
                          "Units": None,
                          "Name": "DMRetranslocationFactor",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": "/d",
                  "Name": "DetachmentRateFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": "0-1",
                  "Name": "MaintenanceRespirationFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": "0-1",
                  "Name": "DMConversionEfficiency",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.VariableReference, Models",
                  "VariableName": "[Structure].TotalStemPopn",
                  "Name": "HeadNumber",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "DMReallocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.AddFunction, Models",
                  "Name": "CriticalNConc",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Spike].MinimumNConc",
                      "Name": "MinimumNConc",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "MetabolicNconc",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "NonStructuralN",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Spike].MaximumNConc",
                              "Name": "MaximumNConc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Spike].MinimumNConc",
                              "Name": "MinimumNConc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 0.8,
                          "Units": None,
                          "Name": "Proportion",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                  "Name": "SenescenceRate",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.XYPairs, Models",
                      "X": [
                        3.0,
                        4.0,
                        5.0,
                        6.0,
                        7.0,
                        8.0
                      ],
                      "Y": [
                        0.07,
                        0.04,
                        0.04,
                        0.015,
                        0.012,
                        0.012
                      ],
                      "XVariableName": None,
                      "Name": "XYPairs",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].Stage",
                      "Name": "XValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.4,
                  "Units": None,
                  "Name": "CarbonConcentration",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientDemandFunctions, Models",
                  "Name": "DMDemands",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.DemandFunctions.PartitionFractionDemandFunction, Models",
                          "Name": "DMDemandFunction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.LinearInterpolationFunction, Models",
                              "Name": "PartitionFraction",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.XYPairs, Models",
                                  "X": [
                                    3.0,
                                    4.0,
                                    5.0,
                                    5.7,
                                    6.0,
                                    7.0,
                                    8.0,
                                    9.0,
                                    9.000001
                                  ],
                                  "Y": [
                                    0.0,
                                    0.0,
                                    0.0,
                                    0.01,
                                    0.7,
                                    0.7,
                                    0.7,
                                    0.7,
                                    0.0
                                  ],
                                  "XVariableName": None,
                                  "Name": "XYPairs",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Phenology].Stage",
                                  "Name": "XValue",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.PhaseLookup, Models",
                          "Name": "StructuralFraction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.PhaseLookupValue, Models",
                              "Start": "Emergence",
                              "End": "StartGrainFill",
                              "Name": "VegetativeGrowth",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.Constant, Models",
                                  "FixedValue": 0.3,
                                  "Units": None,
                                  "Name": "StructuralFractionEG",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.PhaseLookupValue, Models",
                              "Start": "StartGrainFill",
                              "End": "EndGrainFill",
                              "Name": "ReproductiveGrowth",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.Constant, Models",
                                  "FixedValue": 0.0,
                                  "Units": None,
                                  "Name": "StructuralFractionGF",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.DemandFunctions.StorageDMDemandFunction, Models",
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "StorageFraction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 1.0,
                              "Units": None,
                              "Name": "One",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Spike].DMDemands.Structural.StructuralFraction",
                              "Name": "StructuralFraction",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStructuralPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QMetabolicPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStoragePriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientDemandFunctions, Models",
                  "Name": "NDemands",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Spike].minimumNconc",
                          "Name": "MinNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Spike].potentialDMAllocation.Structural",
                          "Name": "PotentialDMAllocation",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "MetabolicNconc",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Spike].criticalNConc",
                              "Name": "CritNconc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Spike].minimumNconc",
                              "Name": "MinNconc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Spike].potentialDMAllocation.Structural",
                          "Name": "PotentialDMAllocation",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.DemandFunctions.StorageNDemandFunction, Models",
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Spike].nitrogenDemandSwitch",
                          "Name": "NitrogenDemandSwitch",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Spike].maximumNconc",
                          "Name": "MaxNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStructuralPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QMetabolicPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStoragePriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.RetranslocateNonStructural, Models",
                  "Name": "RetranslocateNitrogen",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientPoolFunctions, Models",
                  "Name": "InitialWt",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": "g/m^2",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "Photosynthesis",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.VariableReference, Models",
                  "VariableName": "[Spike].MinimumNConc",
                  "Name": "initialNConcFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.Organs.GenericOrgan, Models",
              "StartLive": None,
              "DMSupply": None,
              "NSupply": None,
              "DMDemand": None,
              "NDemand": None,
              "potentialDMAllocation": None,
              "IsAboveGround": True,
              "Name": "Stem",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                  "Name": "SenescenceRate",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.XYPairs, Models",
                      "X": [
                        3.0,
                        4.0,
                        5.0,
                        6.0,
                        7.0,
                        8.0,
                        9.0,
                        10.0,
                        11.0
                      ],
                      "Y": [
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.2,
                        0.5
                      ],
                      "XVariableName": None,
                      "Name": "XYPairs",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].Stage",
                      "Name": "XValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "RemobilisationCost",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                  "Name": "MaximumNConc",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.XYPairs, Models",
                      "X": [
                        3.0,
                        4.0,
                        5.0,
                        6.0,
                        7.0,
                        8.0
                      ],
                      "Y": [
                        0.07,
                        0.04,
                        0.04,
                        0.015,
                        0.012,
                        0.012
                      ],
                      "XVariableName": None,
                      "Name": "XYPairs",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].Stage",
                      "Name": "XValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0025,
                  "Units": None,
                  "Name": "MinimumNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhaseLookupValue, Models",
                  "Start": "Emergence",
                  "End": "StartGrainFill",
                  "Name": "NitrogenDemandSwitch",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Constant",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "NReallocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhaseLookup, Models",
                  "Name": "NRetranslocationFactor",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.PhaseLookupValue, Models",
                      "Start": "Sowing",
                      "End": "Flowering",
                      "Name": "VegetativeGrowth",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 0.0,
                          "Units": None,
                          "Name": "Fraction",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.PhaseLookupValue, Models",
                      "Start": "Flowering",
                      "End": "HarvestRipe",
                      "Name": "ReproductiveGrowth",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 0.5,
                          "Units": None,
                          "Name": "Fraction",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhaseLookup, Models",
                  "Name": "DMRetranslocationFactor",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.PhaseLookupValue, Models",
                      "Start": "Emergence",
                      "End": "StartGrainFill",
                      "Name": "VegetativeGrowth",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 0.0,
                          "Units": None,
                          "Name": "DMRetranslocationFactor",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.PhaseLookupValue, Models",
                      "Start": "StartGrainFill",
                      "End": "EndGrainFill",
                      "Name": "ReproductiveGrowth",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 0.5,
                          "Units": None,
                          "Name": "DMRetranslocationFactor",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Library.BiomassRemoval, Models",
                  "Name": "BiomassRemovalDefaults",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.5,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.5,
                      "FractionDeadToResidue": 1.0,
                      "Name": "Harvest",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.8,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.0,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Cut",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.6,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Prune",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.6,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.2,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Graze",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.05,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Thin",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.DivideFunction, Models",
                  "Name": "PercentWSC",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Stem].Live.StorageWt",
                      "Name": "WSC",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Stem].Live.Wt",
                      "Name": "StemWt",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": "/d",
                  "Name": "DetachmentRateFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": "0-1",
                  "Name": "MaintenanceRespirationFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": "0-1",
                  "Name": "DMConversionEfficiency",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "DMReallocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.AddFunction, Models",
                  "Name": "CriticalNConc",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Stem].MinimumNConc",
                      "Name": "MinimumNConc",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "MetabolicNconc",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "NonStructuralN",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Stem].MaximumNConc",
                              "Name": "MaximumNConc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Stem].MinimumNConc",
                              "Name": "MinimumNConc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 0.5,
                          "Units": None,
                          "Name": "Proportion",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.4,
                  "Units": None,
                  "Name": "CarbonConcentration",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientDemandFunctions, Models",
                  "Name": "DMDemands",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.DemandFunctions.PartitionFractionDemandFunction, Models",
                          "Name": "DMDemandFunction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.PhaseLookup, Models",
                              "Name": "PartitionFraction",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Memo, Models",
                                  "Text": "",
                                  "Name": "memo",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.PhaseLookupValue, Models",
                                  "Start": "Emergence",
                                  "End": "TerminalSpikelet",
                                  "Name": "PreStemElongation",
                                  "ResourceName": None,
                                  "Children": [
                                    {
                                      "$child": "Models.Functions.Constant, Models",
                                      "FixedValue": 0.2,
                                      "Units": None,
                                      "Name": "StemFraction",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    }
                                  ],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.PhaseLookupValue, Models",
                                  "Start": "TerminalSpikelet",
                                  "End": "StartGrainFill",
                                  "Name": "StemElongation",
                                  "ResourceName": None,
                                  "Children": [
                                    {
                                      "$child": "Models.Functions.Constant, Models",
                                      "FixedValue": 0.7,
                                      "Units": "0",
                                      "Name": "StemFraction",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    }
                                  ],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.PhaseLookupValue, Models",
                                  "Start": "StartGrainFill",
                                  "End": "HarvestRipe",
                                  "Name": "EarEmergence",
                                  "ResourceName": None,
                                  "Children": [
                                    {
                                      "$child": "Models.Functions.Constant, Models",
                                      "FixedValue": 0.0,
                                      "Units": None,
                                      "Name": "StemFraction",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    }
                                  ],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.PhaseLookup, Models",
                          "Name": "StructuralFraction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.PhaseLookupValue, Models",
                              "Start": "Sowing",
                              "End": "StartGrainFill",
                              "Name": "VegetativeGrowth",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.Constant, Models",
                                  "FixedValue": 0.3,
                                  "Units": None,
                                  "Name": "Fraction",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.PhaseLookupValue, Models",
                              "Start": "StartGrainFill",
                              "End": "HarvestRipe",
                              "Name": "ReproductiveGrowth",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.Constant, Models",
                                  "FixedValue": 0.01,
                                  "Units": None,
                                  "Name": "Fraction",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.DemandFunctions.StorageDMDemandFunction, Models",
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "StorageFraction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 1.0,
                              "Units": None,
                              "Name": "One",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Stem].DMDemands.Structural.StructuralFraction",
                              "Name": "StructuralFraction",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStructuralPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QMetabolicPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStoragePriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientDemandFunctions, Models",
                  "Name": "NDemands",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Stem].minimumNconc",
                          "Name": "MinNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Stem].potentialDMAllocation.Structural",
                          "Name": "PotentialDMAllocation",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "MetabolicNconc",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Stem].criticalNConc",
                              "Name": "CritNconc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Stem].minimumNconc",
                              "Name": "MinNconc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Stem].potentialDMAllocation.Structural",
                          "Name": "PotentialDMAllocation",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.DemandFunctions.StorageNDemandFunction, Models",
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Stem].nitrogenDemandSwitch",
                          "Name": "NitrogenDemandSwitch",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Stem].maximumNconc",
                          "Name": "MaxNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStructuralPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QMetabolicPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStoragePriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.RetranslocateNonStructural, Models",
                  "Name": "RetranslocateNitrogen",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientPoolFunctions, Models",
                  "Name": "InitialWt",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": "g/m^2",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "Photosynthesis",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.VariableReference, Models",
                  "VariableName": "[Stem].MinimumNConc",
                  "Name": "initialNConcFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.CompositeBiomass, Models",
              "OrganNames": [
                "Leaf",
                "Stem",
                "Spike",
                "Grain"
              ],
              "IncludeLive": True,
              "IncludeDead": True,
              "Name": "AboveGround",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.CompositeBiomass, Models",
              "OrganNames": [
                "Leaf",
                "Stem",
                "Spike",
                "Grain"
              ],
              "IncludeLive": True,
              "IncludeDead": False,
              "Name": "AboveGroundLive",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.CompositeBiomass, Models",
              "OrganNames": [
                "Leaf",
                "Stem",
                "Spike",
                "Grain"
              ],
              "IncludeLive": False,
              "IncludeDead": True,
              "Name": "AboveGroundDead",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.CompositeBiomass, Models",
              "OrganNames": [
                "Root"
              ],
              "IncludeLive": True,
              "IncludeDead": True,
              "Name": "BelowGround",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.CompositeBiomass, Models",
              "OrganNames": [
                "Leaf",
                "Stem",
                "Root",
                "Spike",
                "Grain"
              ],
              "IncludeLive": True,
              "IncludeDead": True,
              "Name": "Total",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.CompositeBiomass, Models",
              "OrganNames": [
                "Leaf",
                "Stem",
                "Root",
                "Spike",
                "Grain"
              ],
              "IncludeLive": True,
              "IncludeDead": False,
              "Name": "TotalLive",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.CompositeBiomass, Models",
              "OrganNames": [
                "Leaf",
                "Stem",
                "Root",
                "Spike",
                "Grain"
              ],
              "IncludeLive": False,
              "IncludeDead": True,
              "Name": "TotalDead",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.CompositeBiomass, Models",
              "OrganNames": [
                "Spike",
                "Grain"
              ],
              "IncludeLive": True,
              "IncludeDead": True,
              "Name": "Ear",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Memo, Models",
                  "Text": "\nThe Ear composite biomass object includes the Live and dead pools of all organs within the Ear to allow output for total dry matter and Nutrient contents, including total structural, non-structural and metabolic pools.\n",
                  "Name": "memo",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.CompositeBiomass, Models",
              "OrganNames": [
                "Stem",
                "Spike"
              ],
              "IncludeLive": True,
              "IncludeDead": True,
              "Name": "StemPlusSpike",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.Core.Folder, Models",
              "ShowInDocs": False,
              "GraphsPerPage": 6,
              "Name": "Cultivars",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Core.Folder, Models",
                  "ShowInDocs": False,
                  "GraphsPerPage": 6,
                  "Name": "New Zealand",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 9",
                        "[Phenology].VrnSensitivity.FixedValue = 7",
                        "[Phenology].PpSensitivity.FixedValue = 3"
                      ],
                      "Name": "Alberic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 9",
                        "[Phenology].VrnSensitivity.FixedValue = 10",
                        "[Phenology].PpSensitivity.FixedValue = 6",
                        "[Phenology].HeadEmergencePpSensitivity.FixedValue = 5",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 80"
                      ],
                      "Name": "Amarok",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 9",
                        "[Phenology].VrnSensitivity.FixedValue = 7",
                        "[Phenology].PpSensitivity.FixedValue = 3"
                      ],
                      "Name": "Aspiring",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 8",
                        "[Phenology].VrnSensitivity.FixedValue = 8",
                        "[Phenology].PpSensitivity.FixedValue = 5",
                        "[Phenology].HeadEmergencePpSensitivity.FixedValue = 3",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100",
                        "[Leaf].ExtinctionCoeff.VegetativePhase.FixedValue = 0.7",
                        "[Grain].NumberFunction.GrainNumber.GrainsPerGramOfStem.FixedValue = 22"
                      ],
                      "Name": "BattenWinter",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 7",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 5",
                        "[Phenology].HeadEmergencePpSensitivity.FixedValue = 3",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100",
                        "[Leaf].ExtinctionCoeff.VegetativePhase.FixedValue = 0.7",
                        "[Grain].NumberFunction.GrainNumber.GrainsPerGramOfStem.FixedValue = 22"
                      ],
                      "Name": "BattenSpring",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 9",
                        "[Phenology].VrnSensitivity.FixedValue = 7",
                        "[Phenology].PpSensitivity.FixedValue = 3"
                      ],
                      "Name": "Centaur",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 8",
                        "[Phenology].VrnSensitivity.FixedValue = 8",
                        "[Phenology].PpSensitivity.FixedValue = 8",
                        "[Phenology].HeadEmergencePpSensitivity.FixedValue = 3",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100"
                      ],
                      "Name": "Claire",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 7",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 3",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100"
                      ],
                      "Name": "Conquest",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 14",
                        "[Phenology].VrnSensitivity.FixedValue = 1",
                        "[Phenology].PpSensitivity.FixedValue = 2"
                      ],
                      "Name": "CRWT153",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 8.5",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 0",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 80",
                        "[Leaf].CohortParameters.MaxArea.AreaLargestLeaves.FixedValue = 1600",
                        "[Grain].MaximumPotentialGrainSize.FixedValue = 0.051",
                        "[Grain].NumberFunction.GrainNumber.GrainsPerGramOfStem.FixedValue = 35",
                        "[Leaf].ExtinctionCoeff.VegetativePhase.FixedValue = 0.5",
                        "[Structure].BranchingRate.PotentialBranchingRate.Vegetative.PotentialBranchingRate.XYPairs.Y = 0,0,0,0,4,7,12,20",
                        "[Grain].MaximumNConc.FixedValue = 0.025",
                        "[Grain].MaxNConcDailyGrowth.FixedValue = 0.025",
                        "[Leaf].ExtinctionCoeff.DevelopmentFactor.XYPairs.Y = 1,1,1,1,1.4",
                        "[Leaf].CohortParameters.MinimumNConc.XYPairs.Y = 0.01,0.008, 0.008,0.008, 0.008, 0.008",
                        "[Stem].MinimumNConc.FixedValue = 0.006",
                        "[Spike].DMRetranslocationFactor.ReproductiveGrowth.DMRetranslocationFactor.FixedValue = 0.05"
                      ],
                      "Name": "Discovery",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Memo, Models",
                          "Text": "This is a spring wheat with a high grain number and later tillering",
                          "Name": "Memo",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 9",
                        "[Phenology].VrnSensitivity.FixedValue = 7",
                        "[Phenology].PpSensitivity.FixedValue = 3"
                      ],
                      "Name": "Einstein",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 9",
                        "[Phenology].VrnSensitivity.FixedValue = 7",
                        "[Phenology].PpSensitivity.FixedValue = 3"
                      ],
                      "Name": "Exceed",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 9",
                        "[Phenology].VrnSensitivity.FixedValue = 7",
                        "[Phenology].PpSensitivity.FixedValue = 3"
                      ],
                      "Name": "Majestic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 9",
                        "[Phenology].VrnSensitivity.FixedValue = 7",
                        "[Phenology].PpSensitivity.FixedValue = 3"
                      ],
                      "Name": "Option",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 6",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Phenology].HeadEmergencePpSensitivity.FixedValue = 2",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100"
                      ],
                      "Name": "Otane",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 9",
                        "[Phenology].VrnSensitivity.FixedValue = 7",
                        "[Phenology].PpSensitivity.FixedValue = 3"
                      ],
                      "Name": "Pennant",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 9",
                        "[Phenology].VrnSensitivity.FixedValue = 7",
                        "[Phenology].PpSensitivity.FixedValue = 3"
                      ],
                      "Name": "Regency",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 9",
                        "[Phenology].VrnSensitivity.FixedValue = 7",
                        "[Phenology].PpSensitivity.FixedValue = 3"
                      ],
                      "Name": "Richmond",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 9",
                        "[Phenology].VrnSensitivity.FixedValue = 7",
                        "[Phenology].PpSensitivity.FixedValue = 3"
                      ],
                      "Name": "Robigus",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 7",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 2",
                        "[Phenology].HeadEmergencePpSensitivity.FixedValue = 3",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100"
                      ],
                      "Name": "Rongotea",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 9",
                        "[Phenology].VrnSensitivity.FixedValue = 7",
                        "[Phenology].PpSensitivity.FixedValue = 3"
                      ],
                      "Name": "Rubric",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 9",
                        "[Phenology].VrnSensitivity.FixedValue = 7",
                        "[Phenology].PpSensitivity.FixedValue = 3"
                      ],
                      "Name": "Sage",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 9",
                        "[Phenology].VrnSensitivity.FixedValue = 7",
                        "[Phenology].PpSensitivity.FixedValue = 3"
                      ],
                      "Name": "Saracen",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 9",
                        "[Phenology].VrnSensitivity.FixedValue = 7",
                        "[Phenology].PpSensitivity.FixedValue = 3"
                      ],
                      "Name": "Savannah",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 9",
                        "[Phenology].VrnSensitivity.FixedValue = 7",
                        "[Phenology].PpSensitivity.FixedValue = 3"
                      ],
                      "Name": "Solstice",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 9",
                        "[Phenology].VrnSensitivity.FixedValue = 7",
                        "[Phenology].PpSensitivity.FixedValue = 3"
                      ],
                      "Name": "Tanker",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 9",
                        "[Phenology].VrnSensitivity.FixedValue = 7",
                        "[Phenology].PpSensitivity.FixedValue = 3"
                      ],
                      "Name": "Tribute",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 8",
                        "[Phenology].VrnSensitivity.FixedValue = 10",
                        "[Phenology].PpSensitivity.FixedValue = 0",
                        "[Phenology].HeadEmergenceLongDayBase.FixedValue = 4",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100",
                        "[Phenology].DailyVernalisation.TempResponseProfile.Response.X = 0,1,8",
                        "[Phenology].DailyVernalisation.TempResponseProfile.Response.Y = 0,1.204,0",
                        "[Leaf].ExtinctionCoeff.VegetativePhase.FixedValue = 0.7",
                        "[Grain].NumberFunction.GrainNumber.GrainsPerGramOfStem.FixedValue = 22",
                        "[Grain].MaximumPotentialGrainSize.FixedValue = 0.045",
                        "[Leaf].CohortParameters.MaxArea.AreaLargestLeaves.FixedValue = 3000"
                      ],
                      "Name": "Wakanui",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 9",
                        "[Phenology].VrnSensitivity.FixedValue = 7",
                        "[Phenology].PpSensitivity.FixedValue = 3"
                      ],
                      "Name": "Weston",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Core.Folder, Models",
                  "ShowInDocs": False,
                  "GraphsPerPage": 6,
                  "Name": "Australia",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 5",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Phenology].HeadEmergencePpSensitivity.FixedValue = 0",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100"
                      ],
                      "Name": "Axe",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 6",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100"
                      ],
                      "Name": "Bolac",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 6",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100"
                      ],
                      "Name": "Cunningham",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Anlace",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Arrivato",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Babbler",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Barham",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Baxter",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Bowie",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Buckley",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Burunga",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Camm",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Carinya",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Chief",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Dagger",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Aurora",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Elmore",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Espada",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Frame",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Gazelle",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Halberd",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Meteor",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Preston",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Saphire",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Stiletto",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Wallup",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 6",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100"
                      ],
                      "Name": "Derrimut",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 6",
                        "[Phenology].VrnSensitivity.FixedValue = 1",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 85"
                      ],
                      "Name": "Gregory",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Suneca",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 5",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 3.5",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100",
                        "[Phenology].HeadEmergencePpSensitivity.FixedValue = 0"
                      ],
                      "Name": "Gamenya",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 5",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100"
                      ],
                      "Name": "Gauntlet",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Chara",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Drysdale",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Envoy",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Estoc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Giles",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Harper",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Krichauff",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Phantom",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Pugsley",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Sentinel",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Silverstar",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Sunlin",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Sunvale",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 6",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 5",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100"
                      ],
                      "Name": "Gladius",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 6",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 5",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100",
                        "[Grain].MaximumPotentialGrainSize.FixedValue = 0.041",
                        "[Grain].NumberFunction.GrainNumber.GrainsPerGramOfStem.FixedValue = 17"
                      ],
                      "Name": "Gutha",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 6",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100",
                        "[Grain].NumberFunction.GrainNumber.GrainsPerGramOfStem.FixedValue = 20"
                      ],
                      "Name": "H45",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Saintly",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Tamarinrock",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Waagin",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Wollaroi",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Zippy",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 8",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100"
                      ],
                      "Name": "H46",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 6",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 5",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100",
                        "[Grain].MaximumPotentialGrainSize.FixedValue = 0.041"
                      ],
                      "Name": "Hartog",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Annuello",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Carnamah",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Cobra",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Corack",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Correll",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Diamondbird",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Goroke",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Guardian",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Impala",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Justica",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Kord",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Merlin",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Mitre",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Shield",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Sunguard",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Yallaroi",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Yawa",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 7",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Phenology].HeadEmergenceLongDayBase.FixedValue = 3.5",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 70"
                      ],
                      "Name": "Wills",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 7",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Phenology].HeadEmergenceLongDayBase.FixedValue = 3.5",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 80"
                      ],
                      "Name": "Mercury",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Arrino",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "BonnieRock",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Bowerbird",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Dart",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Dollarbird",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "EmuRock",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Hyperno",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Impose",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Kalka",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Kukri",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Hunter",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Arrow",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Tamaroi",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Katana",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 7",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 5",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100"
                      ],
                      "Name": "Batavia",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 7",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 5",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue =80"
                      ],
                      "Name": "Egret",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 6",
                        "[Phenology].VrnSensitivity.FixedValue = 1",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100"
                      ],
                      "Name": "Janz",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 8",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100"
                      ],
                      "Name": "Kellalac",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Endure",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Scenario",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Zen",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 8",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100"
                      ],
                      "Name": "Kennedy",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 6",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 5",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100"
                      ],
                      "Name": "Lang",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 5",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100"
                      ],
                      "Name": "Livingston",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 6",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100"
                      ],
                      "Name": "Lincoln",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 6",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100"
                      ],
                      "Name": "Mace",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 7",
                        "[Phenology].VrnSensitivity.FixedValue = 5",
                        "[Phenology].PpSensitivity.FixedValue = 5",
                        "[Phenology].HeadEmergenceLongDayBase.FixedValue = 3.5 ",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100"
                      ],
                      "Name": "MacKellar",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Adagio",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Brennan",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Declic",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 7",
                        "[Phenology].VrnSensitivity.FixedValue = 1",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100",
                        "[Grain].NumberFunction.GrainNumber.GrainsPerGramOfStem.FixedValue = 22"
                      ],
                      "Name": "Matong",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 8",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 5",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 90"
                      ],
                      "Name": "McCubbin",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 8",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100"
                      ],
                      "Name": "Ruby",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 7",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100",
                        "[Grain].MaximumPotentialGrainSize.FixedValue = 0.045"
                      ],
                      "Name": "Spear",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 7",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 5",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100"
                      ],
                      "Name": "Sunbri",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 6.5",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 3.5",
                        "[Phenology].HeadEmergencePpSensitivity.FixedValue = 4",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100"
                      ],
                      "Name": "Sunco",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Advantage",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Bellaroi",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Caparoi",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 8",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100"
                      ],
                      "Name": "Ventura",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 8",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 90"
                      ],
                      "Name": "Eaglehawk",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 8",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100"
                      ],
                      "Name": "Wedgetail",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Lorikeet",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Core.Alias, Models",
                          "Name": "Wylah",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 5",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100"
                      ],
                      "Name": "Westonia",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 6",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100",
                        "[Grain].NumberFunction.GrainNumber.GrainsPerGramOfStem.FixedValue = 20",
                        "[Grain].MaximumPotentialGrainSize.FixedValue = 0.045"
                      ],
                      "Name": "Wilgoyne",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 8",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100"
                      ],
                      "Name": "Wyalkatchem",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 8",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 5",
                        "[Phenology].HeadEmergenceLongDayBase.FixedValue = 2.5",
                        "[Phenology].HeadEmergencePpSensitivity.FixedValue = 5",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 80"
                      ],
                      "Name": "Yitpi",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 8",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100"
                      ],
                      "Name": "Young",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 6.3",
                        "[Phenology].PpSensitivity.FixedValue = 1.7",
                        "[Phenology].VrnSensitivity.FixedValue = 5.0"
                      ],
                      "Name": "Scepter",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 5.7",
                        "[Phenology].PpSensitivity.FixedValue = 5.3",
                        "[Phenology].VrnSensitivity.FixedValue = 2.3"
                      ],
                      "Name": "Cutlass",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 8.0",
                        "[Phenology].PpSensitivity.FixedValue = 1.7",
                        "[Phenology].VrnSensitivity.FixedValue = 8.0"
                      ],
                      "Name": "Longsword",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 7.3",
                        "[Phenology].PpSensitivity.FixedValue = 0.0",
                        "[Phenology].VrnSensitivity.FixedValue = 6.3"
                      ],
                      "Name": "CSIROW007",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 7.0",
                        "[Phenology].PpSensitivity.FixedValue = 1.7",
                        "[Phenology].VrnSensitivity.FixedValue = 5.3"
                      ],
                      "Name": "CSIROW023",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 7.7",
                        "[Phenology].PpSensitivity.FixedValue = 1.3",
                        "[Phenology].VrnSensitivity.FixedValue = 6.3"
                      ],
                      "Name": "CSIROW073",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Core.Folder, Models",
                  "ShowInDocs": False,
                  "GraphsPerPage": 6,
                  "Name": "Turkey",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 7",
                        "[Phenology].VrnSensitivity.FixedValue = 4",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100",
                        "[Grain].NumberFunction.GrainNumber.GrainsPerGramOfStem.FixedValue = 16",
                        "[Grain].MaximumPotentialGrainSize.FixedValue = 0.050",
                        "[Grain].MaximumNConc.FixedValue = 0.026",
                        "[Phenology].GrainFilling.Target.FixedValue = 450",
                        "[Stem].DMRetranslocationFactor.ReproductiveGrowth.DMRetranslocationFactor.FixedValue = 0.2",
                        "[Stem].DMDemands.Structural.StructuralFraction.VegetativeGrowth.Fraction.FixedValue = 0.6",
                        "[Leaf].ExtinctionCoeff.VegetativePhase.FixedValue = 0.8"
                      ],
                      "Name": "Konya",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Core.Folder, Models",
                  "ShowInDocs": False,
                  "GraphsPerPage": 6,
                  "Name": "China",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 7",
                        "[Phenology].VrnSensitivity.FixedValue = 5",
                        "[Phenology].PpSensitivity.FixedValue = 3",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100"
                      ],
                      "Name": "Keyu13",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Core.Folder, Models",
                  "ShowInDocs": False,
                  "GraphsPerPage": 6,
                  "Name": "USA",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 6",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 90",
                        "[Leaf].CohortParameters.MaxArea.AreaLargestLeaves.FixedValue = 4000"
                      ],
                      "Name": "Yecora",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 7",
                        "[Phenology].VrnSensitivity.FixedValue = 6",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Grain].NumberFunction.GrainNumber.GrainsPerGramOfStem.FixedValue = 18",
                        "[Grain].MaximumNConc.FixedValue = 0.02"
                      ],
                      "Name": "Rex",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 7",
                        "[Phenology].VrnSensitivity.FixedValue = 6",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Grain].NumberFunction.GrainNumber.GrainsPerGramOfStem.FixedValue = 20",
                        "[Grain].MaximumNConc.FixedValue = 0.02"
                      ],
                      "Name": "Nugaines",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 7",
                        "[Phenology].VrnSensitivity.FixedValue = 6",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Grain].NumberFunction.GrainNumber.GrainsPerGramOfStem.FixedValue = 22",
                        "[Grain].MaximumNConc.FixedValue = 0.02"
                      ],
                      "Name": "Hyslop",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 7",
                        "[Phenology].VrnSensitivity.FixedValue = 6",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Grain].NumberFunction.GrainNumber.GrainsPerGramOfStem.FixedValue = 25",
                        "[Grain].MaximumNConc.FixedValue = 0.02"
                      ],
                      "Name": "Stephens",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 6",
                        "[Phenology].VrnSensitivity.FixedValue = 5",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 75",
                        "[Leaf].CohortParameters.MaxArea.AreaLargestLeaves.FixedValue = 4000",
                        "//[Leaf].Photosynthesis.RUE.FixedValue =1",
                        "[Leaf].ExtinctionCoeff.VegetativePhase.FixedValue = 0.5",
                        "[Phenology].ThermalTime.Response.Y =0.0, 14.0, 0.0",
                        "[Phenology].ThermalTime.Response.X =0.0, 17.0, 30.0"
                      ],
                      "Name": "YecoraCovercrop",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Core.Folder, Models",
                  "ShowInDocs": False,
                  "GraphsPerPage": 6,
                  "Name": "Europe",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 9",
                        "[Phenology].VrnSensitivity.FixedValue = 7",
                        "[Phenology].PpSensitivity.FixedValue = 3",
                        "[Phenology].HeadEmergencePpSensitivity.FixedValue = 3",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100",
                        "[Leaf].ExtinctionCoeff.VegetativePhase.FixedValue = 0.8"
                      ],
                      "Name": "Dekan",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 9",
                        "[Phenology].VrnSensitivity.FixedValue = 7",
                        "[Phenology].PpSensitivity.FixedValue = 3",
                        "[Phenology].HeadEmergencePpSensitivity.FixedValue = 3",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100",
                        "[Leaf].ExtinctionCoeff.VegetativePhase.FixedValue = 0.8"
                      ],
                      "Name": "Rosario",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 9",
                        "[Phenology].VrnSensitivity.FixedValue = 7",
                        "[Phenology].PpSensitivity.FixedValue = 3",
                        "[Phenology].HeadEmergencePpSensitivity.FixedValue = 3",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100",
                        "[Leaf].ExtinctionCoeff.VegetativePhase.FixedValue = 0.8"
                      ],
                      "Name": "Ararat",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 9",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 0",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 90"
                      ],
                      "Name": "Tybalt",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Core.Folder, Models",
                  "ShowInDocs": False,
                  "GraphsPerPage": 6,
                  "Name": "Africa",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 7",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 3",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100",
                        "[Grain].NumberFunction.GrainNumber.GrainsPerGramOfStem.FixedValue = 17",
                        "[Phenology].GrainFilling.Target.FixedValue = 350"
                      ],
                      "Name": "HAR1685",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Core.Folder, Models",
                  "ShowInDocs": False,
                  "GraphsPerPage": 6,
                  "Name": "Iran",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].MinimumLeafNumber.FixedValue = 7",
                        "[Phenology].VrnSensitivity.FixedValue = 0",
                        "[Phenology].PpSensitivity.FixedValue = 4",
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100"
                      ],
                      "Name": "Gorgan",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Core.Folder, Models",
                  "ShowInDocs": False,
                  "GraphsPerPage": 6,
                  "Name": "Just Added",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100.4",
                        "[Phenology].MinimumLeafNumber.FixedValue = 8.0",
                        "[Phenology].PpSensitivity.FixedValue = 1.7",
                        "[Phenology].VrnSensitivity.FixedValue = 5.0"
                      ],
                      "Name": "Adv08_0008",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 95.1",
                        "[Phenology].MinimumLeafNumber.FixedValue = 8.3",
                        "[Phenology].PpSensitivity.FixedValue = 0.7",
                        "[Phenology].VrnSensitivity.FixedValue = 5.7"
                      ],
                      "Name": "Bennett",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 84.4",
                        "[Phenology].MinimumLeafNumber.FixedValue = 6.7",
                        "[Phenology].PpSensitivity.FixedValue = 2.0",
                        "[Phenology].VrnSensitivity.FixedValue = 5.7"
                      ],
                      "Name": "Scythe",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 100.0",
                        "[Phenology].MinimumLeafNumber.FixedValue = 9.0",
                        "[Phenology].PpSensitivity.FixedValue = 1.0",
                        "[Phenology].VrnSensitivity.FixedValue = 6.3"
                      ],
                      "Name": "Beaufort",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 91.5",
                        "[Phenology].MinimumLeafNumber.FixedValue = 7.3",
                        "[Phenology].PpSensitivity.FixedValue = 3.7",
                        "[Phenology].VrnSensitivity.FixedValue = 1.3"
                      ],
                      "Name": "Braewood",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 80.8",
                        "[Phenology].MinimumLeafNumber.FixedValue = 7.7",
                        "[Phenology].PpSensitivity.FixedValue = 1.0",
                        "[Phenology].VrnSensitivity.FixedValue = 3.7"
                      ],
                      "Name": "Calingiri",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 91.8",
                        "[Phenology].MinimumLeafNumber.FixedValue = 6.0",
                        "[Phenology].PpSensitivity.FixedValue = 2.3",
                        "[Phenology].VrnSensitivity.FixedValue = 1.0"
                      ],
                      "Name": "Catalina",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 73.6",
                        "[Phenology].MinimumLeafNumber.FixedValue = 7.0",
                        "[Phenology].PpSensitivity.FixedValue = 1.7",
                        "[Phenology].VrnSensitivity.FixedValue = 4.0"
                      ],
                      "Name": "Condo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 76.0",
                        "[Phenology].MinimumLeafNumber.FixedValue = 6.0",
                        "[Phenology].PpSensitivity.FixedValue = 2.0",
                        "[Phenology].VrnSensitivity.FixedValue = 6.7"
                      ],
                      "Name": "Crusader",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 88.9",
                        "[Phenology].MinimumLeafNumber.FixedValue = 6.3",
                        "[Phenology].PpSensitivity.FixedValue = 1.0",
                        "[Phenology].VrnSensitivity.FixedValue = 3.7"
                      ],
                      "Name": "CSIROW002",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 90.1",
                        "[Phenology].MinimumLeafNumber.FixedValue = 6.0",
                        "[Phenology].PpSensitivity.FixedValue = 0.7",
                        "[Phenology].VrnSensitivity.FixedValue = 4.0"
                      ],
                      "Name": "CSIROW003",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 81.3",
                        "[Phenology].MinimumLeafNumber.FixedValue = 6.7",
                        "[Phenology].PpSensitivity.FixedValue = 0.7",
                        "[Phenology].VrnSensitivity.FixedValue = 5.0"
                      ],
                      "Name": "CSIROW005",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 93.0",
                        "[Phenology].MinimumLeafNumber.FixedValue = 6.0",
                        "[Phenology].PpSensitivity.FixedValue = 1.0",
                        "[Phenology].VrnSensitivity.FixedValue = 0.0"
                      ],
                      "Name": "CSIROW011",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 89.2",
                        "[Phenology].MinimumLeafNumber.FixedValue = 6.0",
                        "[Phenology].PpSensitivity.FixedValue = 1.3",
                        "[Phenology].VrnSensitivity.FixedValue = 3.7"
                      ],
                      "Name": "CSIROW018",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 85.1",
                        "[Phenology].MinimumLeafNumber.FixedValue = 8.0",
                        "[Phenology].PpSensitivity.FixedValue = 1.3",
                        "[Phenology].VrnSensitivity.FixedValue = 5.3"
                      ],
                      "Name": "CSIROW021",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 89.2",
                        "[Phenology].MinimumLeafNumber.FixedValue = 5.7",
                        "[Phenology].PpSensitivity.FixedValue = 1.0",
                        "[Phenology].VrnSensitivity.FixedValue = 0.3"
                      ],
                      "Name": "CSIROW027",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 96.6",
                        "[Phenology].MinimumLeafNumber.FixedValue = 5.7",
                        "[Phenology].PpSensitivity.FixedValue = 0.3",
                        "[Phenology].VrnSensitivity.FixedValue = 0.3"
                      ],
                      "Name": "CSIROW029",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 93.7",
                        "[Phenology].MinimumLeafNumber.FixedValue = 5.3",
                        "[Phenology].PpSensitivity.FixedValue = 0.7",
                        "[Phenology].VrnSensitivity.FixedValue = 0.7"
                      ],
                      "Name": "CSIROW077",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 91.5",
                        "[Phenology].MinimumLeafNumber.FixedValue = 6.0",
                        "[Phenology].PpSensitivity.FixedValue = 1.3",
                        "[Phenology].VrnSensitivity.FixedValue = 0.0"
                      ],
                      "Name": "CSIROW087",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 92.9",
                        "[Phenology].MinimumLeafNumber.FixedValue = 6.0",
                        "[Phenology].PpSensitivity.FixedValue = 1.0",
                        "[Phenology].VrnSensitivity.FixedValue = 2.0"
                      ],
                      "Name": "CSIROW102",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 94.0",
                        "[Phenology].MinimumLeafNumber.FixedValue = 5.0",
                        "[Phenology].PpSensitivity.FixedValue = 1.0",
                        "[Phenology].VrnSensitivity.FixedValue = 1.0"
                      ],
                      "Name": "CSIROW105",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 97.0",
                        "[Phenology].MinimumLeafNumber.FixedValue = 6.0",
                        "[Phenology].PpSensitivity.FixedValue = 2.7",
                        "[Phenology].VrnSensitivity.FixedValue = 1.0"
                      ],
                      "Name": "Hume",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 103.7",
                        "[Phenology].MinimumLeafNumber.FixedValue = 6.0",
                        "[Phenology].PpSensitivity.FixedValue = 4.0",
                        "[Phenology].VrnSensitivity.FixedValue = 4.7"
                      ],
                      "Name": "Ellison",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 95.7",
                        "[Phenology].MinimumLeafNumber.FixedValue = 6.0",
                        "[Phenology].PpSensitivity.FixedValue = 1.3",
                        "[Phenology].VrnSensitivity.FixedValue = 0.0"
                      ],
                      "Name": "Emu_Rock",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 88.0",
                        "[Phenology].MinimumLeafNumber.FixedValue = 6.0",
                        "[Phenology].PpSensitivity.FixedValue = 5.0",
                        "[Phenology].VrnSensitivity.FixedValue = 1.3"
                      ],
                      "Name": "Forrest",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 97.5",
                        "[Phenology].MinimumLeafNumber.FixedValue = 6.0",
                        "[Phenology].PpSensitivity.FixedValue = 2.7",
                        "[Phenology].VrnSensitivity.FixedValue = 1.3"
                      ],
                      "Name": "Grenade",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 82.8",
                        "[Phenology].MinimumLeafNumber.FixedValue = 8.0",
                        "[Phenology].PpSensitivity.FixedValue = 3.0",
                        "[Phenology].VrnSensitivity.FixedValue = 9.0"
                      ],
                      "Name": "Kittyhawk",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 92.3",
                        "[Phenology].MinimumLeafNumber.FixedValue = 7.0",
                        "[Phenology].PpSensitivity.FixedValue = 3.7",
                        "[Phenology].VrnSensitivity.FixedValue = 0.7"
                      ],
                      "Name": "Lancer",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 91.7",
                        "[Phenology].MinimumLeafNumber.FixedValue = 8.0",
                        "[Phenology].PpSensitivity.FixedValue = 1.7",
                        "[Phenology].VrnSensitivity.FixedValue = 4.0"
                      ],
                      "Name": "Nighthawk",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 99.4",
                        "[Phenology].MinimumLeafNumber.FixedValue = 6.7",
                        "[Phenology].PpSensitivity.FixedValue = 1.7",
                        "[Phenology].VrnSensitivity.FixedValue = 5.7"
                      ],
                      "Name": "Magenta",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 109.8",
                        "[Phenology].MinimumLeafNumber.FixedValue = 9.0",
                        "[Phenology].PpSensitivity.FixedValue = 2.3",
                        "[Phenology].VrnSensitivity.FixedValue = 5.0"
                      ],
                      "Name": "Manning",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 94.9",
                        "[Phenology].MinimumLeafNumber.FixedValue = 7.3",
                        "[Phenology].PpSensitivity.FixedValue = 1.0",
                        "[Phenology].VrnSensitivity.FixedValue = 0.0"
                      ],
                      "Name": "Merinda",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 76.6",
                        "[Phenology].MinimumLeafNumber.FixedValue = 7.0",
                        "[Phenology].PpSensitivity.FixedValue = 2.0",
                        "[Phenology].VrnSensitivity.FixedValue = 4.0"
                      ],
                      "Name": "Mitch",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 81.7",
                        "[Phenology].MinimumLeafNumber.FixedValue = 6.7",
                        "[Phenology].PpSensitivity.FixedValue = 3.3",
                        "[Phenology].VrnSensitivity.FixedValue = 1.0"
                      ],
                      "Name": "Ouyen",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 79.3",
                        "[Phenology].MinimumLeafNumber.FixedValue = 6.0",
                        "[Phenology].PpSensitivity.FixedValue = 2.3",
                        "[Phenology].VrnSensitivity.FixedValue = 1.0"
                      ],
                      "Name": "Peake",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 109.7",
                        "[Phenology].MinimumLeafNumber.FixedValue = 9.0",
                        "[Phenology].PpSensitivity.FixedValue = 1.5",
                        "[Phenology].VrnSensitivity.FixedValue = 7.5"
                      ],
                      "Name": "Revenue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 85.4",
                        "[Phenology].MinimumLeafNumber.FixedValue = 8.7",
                        "[Phenology].PpSensitivity.FixedValue = 2.7",
                        "[Phenology].VrnSensitivity.FixedValue = 8.8"
                      ],
                      "Name": "Rosella",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 75.8",
                        "[Phenology].MinimumLeafNumber.FixedValue = 7.0",
                        "[Phenology].PpSensitivity.FixedValue = 1.3",
                        "[Phenology].VrnSensitivity.FixedValue = 5.3"
                      ],
                      "Name": "Scout",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 84.4",
                        "[Phenology].MinimumLeafNumber.FixedValue = 7.0",
                        "[Phenology].PpSensitivity.FixedValue = 1.3",
                        "[Phenology].VrnSensitivity.FixedValue = 4.0"
                      ],
                      "Name": "Spitfire",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 82.7",
                        "[Phenology].MinimumLeafNumber.FixedValue = 7.3",
                        "[Phenology].PpSensitivity.FixedValue = 1.7",
                        "[Phenology].VrnSensitivity.FixedValue = 5.3"
                      ],
                      "Name": "Strzelecki",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 91.3",
                        "[Phenology].MinimumLeafNumber.FixedValue = 6.0",
                        "[Phenology].PpSensitivity.FixedValue = 1.3",
                        "[Phenology].VrnSensitivity.FixedValue = 4.0"
                      ],
                      "Name": "Sunbee",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 93.8",
                        "[Phenology].MinimumLeafNumber.FixedValue = 7.3",
                        "[Phenology].PpSensitivity.FixedValue = 3.7",
                        "[Phenology].VrnSensitivity.FixedValue = 6.7"
                      ],
                      "Name": "Sunlamb",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 84.1",
                        "[Phenology].MinimumLeafNumber.FixedValue = 6.7",
                        "[Phenology].PpSensitivity.FixedValue = 1.0",
                        "[Phenology].VrnSensitivity.FixedValue = 3.3"
                      ],
                      "Name": "Sunstate",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 92.2",
                        "[Phenology].MinimumLeafNumber.FixedValue = 7.0",
                        "[Phenology].PpSensitivity.FixedValue = 1.3",
                        "[Phenology].VrnSensitivity.FixedValue = 1.0"
                      ],
                      "Name": "Suntop",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 85.8",
                        "[Phenology].MinimumLeafNumber.FixedValue = 6.7",
                        "[Phenology].PpSensitivity.FixedValue = 3.3",
                        "[Phenology].VrnSensitivity.FixedValue = 4.0"
                      ],
                      "Name": "Trojan",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 92.0",
                        "[Phenology].MinimumLeafNumber.FixedValue = 8.0",
                        "[Phenology].PpSensitivity.FixedValue = 1.0",
                        "[Phenology].VrnSensitivity.FixedValue = 6.7"
                      ],
                      "Name": "Illabo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Phyllochron.BasePhyllochron.FixedValue = 91.5",
                        "[Phenology].MinimumLeafNumber.FixedValue = 8.3",
                        "[Phenology].PpSensitivity.FixedValue = 2.0",
                        "[Phenology].VrnSensitivity.FixedValue = 6.2"
                      ],
                      "Name": "Whistler",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.Functions.Constant, Models",
              "FixedValue": 0.0,
              "Units": None,
              "Name": "MortalityRate",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.Functions.Constant, Models",
              "FixedValue": 0.0,
              "Units": None,
              "Name": "SeedMortalityRate",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            }
          ],
          "Enabled": True,
          "ReadOnly": False
        },
        {
          "$child": "Models.PMF.Plant, Models",
          "Name": "Soybean",
          "ResourceName": "Soybean",
          "Children": [
            {
              "$child": "Models.PMF.OrganArbitrator, Models",
              "Name": "Arbitrator",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.PMF.BiomassTypeArbitrator, Models",
                  "Name": "DMArbitration",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Core.Folder, Models",
                      "ShowInDocs": True,
                      "GraphsPerPage": 6,
                      "Name": "PotentialPartitioningMethods",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.PMF.Arbitrator.ReallocationMethod, Models",
                          "Name": "ReallocationMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.PMF.Arbitrator.AllocateFixationMethod, Models",
                          "Name": "AllocateFixationMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.PMF.Arbitrator.RetranslocationMethod, Models",
                          "Name": "RetranslocationMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.PMF.Arbitrator.SendPotentialDMAllocationsMethod, Models",
                          "Name": "SendPotentialDMAllocationsMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Core.Folder, Models",
                      "ShowInDocs": True,
                      "GraphsPerPage": 6,
                      "Name": "AllocationMethods",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.PMF.Arbitrator.NutrientConstrainedAllocationMethod, Models",
                          "Name": "NutrientConstrainedAllocationMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.PMF.Arbitrator.DryMatterAllocationsMethod, Models",
                          "Name": "DryMatterAllocationsMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.RelativeAllocation, Models",
                      "Name": "ArbitrationMethod",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.BiomassTypeArbitrator, Models",
                  "Name": "NArbitration",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Core.Folder, Models",
                      "ShowInDocs": True,
                      "GraphsPerPage": 6,
                      "Name": "PotentialPartitioningMethods",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.PMF.Arbitrator.ReallocationMethod, Models",
                          "Name": "ReallocationMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Core.Folder, Models",
                      "ShowInDocs": True,
                      "GraphsPerPage": 6,
                      "Name": "ActualPartitioningMethods",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.PMF.Arbitrator.AllocateFixationMethod, Models",
                          "Name": "AllocateFixationMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.PMF.Arbitrator.RetranslocationMethod, Models",
                          "Name": "RetranslocationMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Core.Folder, Models",
                      "ShowInDocs": True,
                      "GraphsPerPage": 6,
                      "Name": "AllocationMethods",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.PMF.Arbitrator.NitrogenAllocationsMethod, Models",
                          "Name": "NitrogenAllocationsMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.RelativeAllocation, Models",
                      "Name": "ArbitrationMethod",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Arbitrator.AllocateUptakesMethod, Models",
                      "Name": "AllocateUptakesMethod",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Arbitrator.WaterUptakeMethod, Models",
                  "Name": "WaterUptakeMethod",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Arbitrator.NitrogenUptakeMethod, Models",
                  "Name": "NitrogenUptakeMethod",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.Phen.Phenology, Models",
              "Name": "Phenology",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Memo, Models",
                  "Text": "In the new model we simplified phenology by taking out stages that are not measurable (e.g. end of juvenile stage) and by adding new stages that are measurable (e.g. start pod). \nThe new phenology follows the V/R staging system.\n\n![Alt Text](Diagrams_soy_phenology_comparison.png)\n\n**Figure [FigureNumber]:** Comparison of soybean phenological stages for APSIM Classic and APSIM Next Generation.",
                  "Name": "Introduction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Memo, Models",
                  "Text": "The key differences for cultivars are mostly phenological parameters (e.g., Vegetative.Target, ReproductivePhotoperiodModifier). In some cases, some additional parameters have been changed, see parameter values in the cultivar section. ",
                  "Name": "Memo",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.HourlyInterpolation, Models",
                  "agregationMethod": 0,
                  "Name": "VegetativeThermalTime",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.XYPairs, Models",
                      "X": [
                        10.0,
                        20.0,
                        30.0,
                        40.0
                      ],
                      "Y": [
                        0.0,
                        10.0,
                        20.0,
                        0.0
                      ],
                      "XVariableName": None,
                      "Name": "Response",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.ThreeHourSin, Models",
                      "TempRangeFactors": None,
                      "Name": "InterpolationMethod",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                  "Name": "VegetativePhotoperiodModifier",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.XYPairs, Models",
                      "X": [
                        14.43,
                        21.19
                      ],
                      "Y": [
                        1.0,
                        0.0
                      ],
                      "XVariableName": None,
                      "Name": "XYPairs",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "These parameters are overridden by cultivar parameter descriptions to capture the difference between individual cultivars in terms of their phenological development.  Responses shown here are only indicative.  See cultivar descriptions for more information.",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].Photoperiod",
                      "Name": "XValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.HourlyInterpolation, Models",
                  "agregationMethod": 0,
                  "Name": "ReproductiveThermalTime",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.XYPairs, Models",
                      "X": [
                        10.0,
                        15.0,
                        30.0,
                        40.0
                      ],
                      "Y": [
                        5.0,
                        5.0,
                        20.0,
                        0.0
                      ],
                      "XVariableName": None,
                      "Name": "Response",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.ThreeHourSin, Models",
                      "TempRangeFactors": None,
                      "Name": "InterpolationMethod",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "Cardinal temperatures for vegetative and reproductive thermal time are the same except the range of 0-15° C. A base temperature of 10° C is used for vegetative phase, which temperature has been used to calibrate 56 soybean cultivars in the USA ([ARCHONTOULIS2014465]) and 72 soybean cultivars in China ([WU2019125940]). The small changes in cardinal temperatures for reproductive thermal time was important for solving issues related to photoperiod response for some Australian cultivars. Such a differentiation in base temperature between V and R stages is supported by [BootChapter1998], DSSAT-CROPGRO soybean model, who showed that cardinal temperatures may differ for vegetative and reproductive development for soybean. This is something to further explore as new data from other environments are added to the APSIM NG database.\n\n",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                  "Name": "ReproductivePhotoperiodModifier",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.XYPairs, Models",
                      "X": [
                        14.43,
                        21.19
                      ],
                      "Y": [
                        1.0,
                        0.0
                      ],
                      "XVariableName": None,
                      "Name": "XYPairs",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "These parameters are overridden by cultivar parameter descriptions to capture the difference between individual cultivars in terms of their phenological development.  Responses shown here are only indicative.  See cultivar descriptions for more information.",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].Photoperiod",
                      "Name": "XValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhaseLookup, Models",
                  "Name": "ThermalTime",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.PhaseLookupValue, Models",
                      "Start": "Sowing",
                      "End": "Emergence",
                      "Name": "PreEmergence",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Phenology].VegetativeThermalTime",
                          "Name": "Base",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.PhaseLookupValue, Models",
                      "Start": "Emergence",
                      "End": "EndCanopyDevelopment",
                      "Name": "VegetativeGrowth",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.MultiplyFunction, Models",
                          "Name": "TT",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Phenology].VegetativeThermalTime",
                              "Name": "Base",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Phenology].VegetativePhotoperiodModifier",
                              "Name": "PPModifier",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.PhaseLookupValue, Models",
                      "Start": "EndCanopyDevelopment",
                      "End": "HarvestRipe",
                      "Name": "ReproductiveGrowth",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.MultiplyFunction, Models",
                          "Name": "TT",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Phenology].ReproductiveThermalTime",
                              "Name": "Base",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Phenology].ReproductivePhotoperiodModifier",
                              "Name": "PPModifier",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhotoperiodFunction, Models",
                  "Twilight": 0.0,
                  "DayLength": 0.0,
                  "Name": "Photoperiod",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.GerminatingPhase, Models",
                  "Start": "Sowing",
                  "End": "Germination",
                  "Name": "Germinating",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.EmergingPhase, Models",
                  "Start": "Germination",
                  "End": "Emergence",
                  "TTForTimeStep": 0.0,
                  "Name": "Emerging",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].VegetativeThermalTime",
                      "Name": "ThermalTime",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.AddFunction, Models",
                      "Name": "Target",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 10.0,
                          "Units": None,
                          "Name": "ShootLag",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.MultiplyFunction, Models",
                          "Name": "DepthxRate",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Plant].SowingData.Depth",
                              "Name": "SowingDepth",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 1.0,
                              "Units": None,
                              "Name": "ShootRate",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.GenericPhase, Models",
                  "Start": "Emergence",
                  "End": "StartFlowering",
                  "Name": "Vegetative",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 200.0,
                      "Units": None,
                      "Name": "Target",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "Progression",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.GenericPhase, Models",
                  "Start": "StartFlowering",
                  "End": "StartPodDevelopment",
                  "Name": "EarlyFlowering",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\r\nThe duration of the grain filling phase has been taken from the model of Robertson et al (2002).",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 200.0,
                      "Units": None,
                      "Name": "Target",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "Progression",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.GenericPhase, Models",
                  "Start": "StartPodDevelopment",
                  "End": "StartGrainFilling",
                  "Name": "EarlyPodDevelopment",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\r\nThe duration of the grain filling phase has been taken from the model of Robertson et al (2002).",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 140.0,
                      "Units": None,
                      "Name": "Target",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "Progression",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.GenericPhase, Models",
                  "Start": "StartGrainFilling",
                  "End": "EndCanopyDevelopment",
                  "Name": "EarlyGrainFilling",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Target",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 0.05,
                          "Units": None,
                          "Name": "FractionofGrainfilling",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue",
                          "Name": "EntireGrainfilling",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "Progression",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.GenericPhase, Models",
                  "Start": "EndCanopyDevelopment",
                  "End": "EndPodDevelopment",
                  "Name": "MidGrainFilling",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Target",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 0.5,
                          "Units": None,
                          "Name": "FractionofMidToLateGrainfilling",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "MidToLateGrainfilling",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue",
                              "Name": "EntireGrainfilling",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Phenology].EarlyGrainFilling.Target",
                              "Name": "EarlyGrainfilling",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "Progression",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.GenericPhase, Models",
                  "Start": "EndPodDevelopment",
                  "End": "EndGrainFill",
                  "Name": "LateGrainFilling",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.SubtractFunction, Models",
                      "Name": "Target",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 500.0,
                          "Units": None,
                          "Name": "EntireGrainfillPeriod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Phenology].EarlyGrainFilling.Target",
                          "Name": "EarlyGrainFillingTarget",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Phenology].MidGrainFilling.Target",
                          "Name": "MidGrainFillingTarget",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "Progression",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.GenericPhase, Models",
                  "Start": "EndGrainFill",
                  "End": "Maturity",
                  "Name": "Maturing",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 45.0,
                      "Units": None,
                      "Name": "Target",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "Progression",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.GenericPhase, Models",
                  "Start": "Maturity",
                  "End": "HarvestRipe",
                  "Name": "Ripening",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 45.0,
                      "Units": None,
                      "Name": "Target",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "Progression",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.EndPhase, Models",
                  "Start": "HarvestRipe",
                  "End": "Unused",
                  "Name": "ReadyForHarvesting",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "ThermalTime",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.OnEventFunction, Models",
                  "SetEvent": "Emergence",
                  "ReSetEvent": "never",
                  "Name": "EmergenceDAS",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\r\nA function is used to provide emergence date as days after sowing(DAS).",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "PreEventValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Plant].DaysAfterSowing",
                      "Name": "PostEventValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.OnEventFunction, Models",
                  "SetEvent": "StartFlowering",
                  "ReSetEvent": "never",
                  "Name": "FloweringDAS",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\r\nA function is used to provide flowering date as days after sowing(DAS).",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "PreEventValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Plant].DaysAfterSowing",
                      "Name": "PostEventValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.OnEventFunction, Models",
                  "SetEvent": "Maturity",
                  "ReSetEvent": "never",
                  "Name": "MaturityDAS",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "\r\nA function is used to provide maturity date as days after sowing(DAS).",
                      "Name": "memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "PreEventValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Plant].DaysAfterSowing",
                      "Name": "PostEventValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.Organs.SimpleLeaf, Models",
              "Albedo": 0.25,
              "Gsmax350": 0.015,
              "R50": 150.0,
              "LeafInitialisationStage": "Emergence",
              "Height": 0.0,
              "BaseHeight": 0.0,
              "Width": 0.0,
              "FRGR": 0.0,
              "PotentialEP": 0.0,
              "WaterDemand": 0.0,
              "LightProfile": None,
              "KDead": 0.2,
              "LAIDead": 0.0,
              "DMSupply": None,
              "NSupply": None,
              "DMDemand": None,
              "NDemand": None,
              "potentialDMAllocation": None,
              "Name": "Leaf",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Functions.SupplyFunctions.RUEModel, Models",
                  "Name": "Photosynthesis",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.2,
                      "Units": None,
                      "Name": "RUE",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Memo, Models",
                          "Text": "This value has been fitted to available data. In contrast to Classic, the RUE value in NG refers to the whole plant (above and below ground) and the calibrated value is 1.2 g/MJ. There are no literature references for whole plant RUE in soybean. To come up with this value we considered: RUE for above ground biomass ([SINCLAIR198929]; [VANROEKEL20141189]), biomass and root data from Iowa FACTS experiments and model fitting to available data.",
                          "Name": "Memo",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.WeightedTemperatureFunction, Models",
                      "MaximumTemperatureWeighting": 0.5,
                      "Name": "FT",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            10.0,
                            20.0,
                            30.0,
                            40.0
                          ],
                          "Y": [
                            0.0,
                            1.0,
                            1.0,
                            0.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Memo, Models",
                          "Text": "The function for temperature effect on RUE comes from APSIM classic, which seems to work well across diverse environments. We did not find any comments/suggestions in the published papers suggesting alternative ones, so for now we used the same function. As more data from different environments are added to APSIM database, it will be possible to further test the photosynthetic temperature response in the future. ",
                          "Name": "Memo",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "FVPD",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            10.0,
                            50.0
                          ],
                          "Y": [
                            1.0,
                            1.0,
                            1.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].Photosynthesis.VPD",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "FN",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            1.0,
                            1.5
                          ],
                          "Y": [
                            0.0,
                            1.0,
                            1.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].Fn",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "FW",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            1.0
                          ],
                          "Y": [
                            0.0,
                            1.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].Fw",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.SupplyFunctions.RUECO2Function, Models",
                      "PhotosyntheticPathway": "C3",
                      "Name": "FCO2",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Leaf].RadiationIntercepted",
                      "Name": "RadnInt",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.SupplyFunctions.StomatalConductanceCO2Modifier, Models",
                  "Name": "StomatalConductanceCO2Modifier",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Leaf].Photosynthesis.FCO2",
                      "Name": "PhotosynthesisCO2Modifier",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.MultiplyFunction, Models",
                  "Name": "Area",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Leaf].SpecificArea",
                      "Name": "SLA",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Leaf].Live.Wt",
                      "Name": "LeafLiveWt",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.MinimumFunction, Models",
                  "Name": "FRGR",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Leaf].Photosynthesis.FT",
                      "Name": "RUE_FT",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Leaf].Photosynthesis.FN",
                      "Name": "RUE_FN",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Leaf].Photosynthesis.FVPD",
                      "Name": "RUE_FVPD",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                  "Name": "ExtinctionCoefficient",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.XYPairs, Models",
                      "X": [
                        200.0,
                        1000.0
                      ],
                      "Y": [
                        0.6,
                        0.4
                      ],
                      "XVariableName": None,
                      "Name": "XYPairs",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Soybean].SowingData.RowSpacing",
                      "Name": "XValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                  "Name": "Tallness",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.XYPairs, Models",
                      "X": [
                        1.0,
                        2.0,
                        3.0,
                        4.0,
                        5.0,
                        6.0,
                        9.0
                      ],
                      "Y": [
                        0.0,
                        0.0,
                        50.0,
                        400.0,
                        700.0,
                        900.0,
                        1000.0
                      ],
                      "XVariableName": None,
                      "Name": "XYPairs",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].Stage",
                      "Name": "XValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.AddFunction, Models",
                  "Name": "NodeNumber",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Initial",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.AccumulateFunction, Models",
                      "StartStageName": "Emergence",
                      "EndStageName": "EndCanopyDevelopment",
                      "ResetStageName": None,
                      "FractionRemovedOnCut": 1.0,
                      "FractionRemovedOnHarvest": 1.0,
                      "FractionRemovedOnGraze": 1.0,
                      "FractionRemovedOnPrune": 0.0,
                      "Name": "Change",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.DivideFunction, Models",
                          "Name": "DeltaNode",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Phenology].VegetativeThermalTime",
                              "Name": "ThermalTime",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Leaf].Phyllochron",
                              "Name": "Phyllochron",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": None,
                  "Name": "NReallocationFactor",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "This model assumes that all metabolic and storage N is available for reallocation from senescing leaves.",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhaseLookup, Models",
                  "Name": "SenescenceRate",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.PhaseLookupValue, Models",
                      "Start": "Emergence",
                      "End": "EndCanopyDevelopment",
                      "Name": "Vegetative",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 0.0,
                          "Units": None,
                          "Name": "Rate",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.PhaseLookupValue, Models",
                      "Start": "EndCanopyDevelopment",
                      "End": "Maturity",
                      "Name": "Reproductive",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.DivideFunction, Models",
                          "Name": "Rate",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Phenology].ThermalTime",
                              "Name": "TT",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.SubtractFunction, Models",
                              "Name": "TTRemaining",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.ExpressionFunction, Models",
                                  "Expression": "[Phenology].MidGrainFilling.Target + [Phenology].LateGrainFilling.Target + [Phenology].Maturing.Target",
                                  "Name": "TTRequired",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.ExpressionFunction, Models",
                                  "Expression": "[Phenology].MidGrainFilling.ProgressThroughPhase + [Phenology].LateGrainFilling.ProgressThroughPhase + [Phenology].Maturing.ProgressThroughPhase",
                                  "Name": "TTComplete",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Memo, Models",
                          "Text": "Leaf senescence rate is calculated to ensure that all leaves are senesced by crop maturity.",
                          "Name": "Memo",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Memo, Models",
                          "Text": "There is no data on detachment rates and so a simple value was chosen to ensure that leaves were not retained in the canopy. In the future, with more data available a thermal time basis leaf detachment rate will be explored. ",
                          "Name": "Memo1",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.05,
                  "Units": "/d",
                  "Name": "DetachmentRate",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": None,
                  "Name": "InitialWt",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.MultiplyFunction, Models",
                  "Name": "LAIDead",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Leaf].SpecificArea",
                      "Name": "SLA",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Leaf].Dead.Wt",
                      "Name": "LeafDeadWt",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                  "Name": "MaximumNConc",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.XYPairs, Models",
                      "X": [
                        3.0,
                        4.0,
                        9.0
                      ],
                      "Y": [
                        0.06,
                        0.05,
                        0.03
                      ],
                      "XVariableName": None,
                      "Name": "XYPairs",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].Stage",
                      "Name": "XValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "These values have been fitted to the available data. The fitted values are within the observed maximum and minimum values obtained in the FACTS soybean experiments ([ARCHONTOULIS2020721]). \n\n",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                  "Name": "CriticalNConc",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.XYPairs, Models",
                      "X": [
                        3.0,
                        4.0,
                        9.0
                      ],
                      "Y": [
                        0.06,
                        0.05,
                        0.03
                      ],
                      "XVariableName": None,
                      "Name": "XYPairs",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].Stage",
                      "Name": "XValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.02,
                  "Units": None,
                  "Name": "MinimumNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": "0-1",
                  "Name": "MaintenanceRespiration",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": "0-1",
                  "Name": "DMConversionEfficiency",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Library.BiomassRemoval, Models",
                  "Name": "BiomassRemovalDefaults",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 1.0,
                      "FractionDeadToResidue": 1.0,
                      "Name": "Harvest",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 1.0,
                      "FractionDeadToResidue": 1.0,
                      "Name": "Cut",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.5,
                  "Units": None,
                  "Name": "NRetranslocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "DMReallocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "DMRetranslocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "RemobilisationCost",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.4,
                  "Units": None,
                  "Name": "CarbonConcentration",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientDemandFunctions, Models",
                  "Name": "DMDemands",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.DivideFunction, Models",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.PhaseLookup, Models",
                          "Name": "DeltaLAIPotential",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Memo, Models",
                              "Text": "DeltaLAI describes the LAI variation over the cycle and is calculated using the Area of Largest Leaf, relative leaf area, plant number, branch number and leaf appearance rate (thermal time/phyllochron).",
                              "Name": "Memo",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.PhaseLookupValue, Models",
                              "Start": "Emergence",
                              "End": "EndCanopyDevelopment",
                              "Name": "Vegetative",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.MultiplyFunction, Models",
                                  "Name": "Delta",
                                  "ResourceName": None,
                                  "Children": [
                                    {
                                      "$child": "Models.Functions.VariableReference, Models",
                                      "VariableName": "[Leaf].AreaLargestLeaf",
                                      "Name": "AreaLargestLeaf",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    },
                                    {
                                      "$child": "Models.Functions.PhaseLookup, Models",
                                      "Name": "RelativeArea",
                                      "ResourceName": None,
                                      "Children": [
                                        {
                                          "$child": "Models.Functions.PhaseLookupValue, Models",
                                          "Start": "Emergence",
                                          "End": "StartPodDevelopment",
                                          "Name": "BeforeEndPod",
                                          "ResourceName": None,
                                          "Children": [
                                            {
                                              "$child": "Models.Functions.LinearInterpolationFunction, Models",
                                              "Name": "RelativeArea",
                                              "ResourceName": None,
                                              "Children": [
                                                {
                                                  "$child": "Models.Memo, Models",
                                                  "Text": "",
                                                  "Name": "Memo",
                                                  "ResourceName": None,
                                                  "Children": [],
                                                  "Enabled": True,
                                                  "ReadOnly": False
                                                },
                                                {
                                                  "$child": "Models.Functions.XYPairs, Models",
                                                  "X": [
                                                    2.0,
                                                    7.0,
                                                    10.0,
                                                    18.0
                                                  ],
                                                  "Y": [
                                                    0.6666,
                                                    1.0,
                                                    1.0,
                                                    0.83
                                                  ],
                                                  "XVariableName": None,
                                                  "Name": "XYPairs",
                                                  "ResourceName": None,
                                                  "Children": [],
                                                  "Enabled": True,
                                                  "ReadOnly": False
                                                },
                                                {
                                                  "$child": "Models.Functions.VariableReference, Models",
                                                  "VariableName": "[Leaf].NodeNumber",
                                                  "Name": "XValue",
                                                  "ResourceName": None,
                                                  "Children": [],
                                                  "Enabled": True,
                                                  "ReadOnly": False
                                                }
                                              ],
                                              "Enabled": True,
                                              "ReadOnly": False
                                            }
                                          ],
                                          "Enabled": True,
                                          "ReadOnly": False
                                        },
                                        {
                                          "$child": "Models.Functions.PhaseLookupValue, Models",
                                          "Start": "StartPodDevelopment",
                                          "End": "EndCanopyDevelopment",
                                          "Name": "AfterEndPod",
                                          "ResourceName": None,
                                          "Children": [
                                            {
                                              "$child": "Models.Functions.Constant, Models",
                                              "FixedValue": 0.4,
                                              "Units": None,
                                              "Name": "RelativeArea",
                                              "ResourceName": None,
                                              "Children": [],
                                              "Enabled": True,
                                              "ReadOnly": False
                                            }
                                          ],
                                          "Enabled": True,
                                          "ReadOnly": False
                                        }
                                      ],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    },
                                    {
                                      "$child": "Models.Functions.VariableReference, Models",
                                      "VariableName": "[Soybean].SowingData.Population",
                                      "Name": "PlantNumber",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    },
                                    {
                                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                                      "Name": "Branches",
                                      "ResourceName": None,
                                      "Children": [
                                        {
                                          "$child": "Models.Functions.XYPairs, Models",
                                          "X": [
                                            0.0,
                                            6.0,
                                            7.0,
                                            8.0,
                                            9.0
                                          ],
                                          "Y": [
                                            1.0,
                                            1.0,
                                            2.0,
                                            2.0,
                                            2.0
                                          ],
                                          "XVariableName": None,
                                          "Name": "XYPairs",
                                          "ResourceName": None,
                                          "Children": [],
                                          "Enabled": True,
                                          "ReadOnly": False
                                        },
                                        {
                                          "$child": "Models.Functions.VariableReference, Models",
                                          "VariableName": "[Leaf].NodeNumber",
                                          "Name": "XValue",
                                          "ResourceName": None,
                                          "Children": [],
                                          "Enabled": True,
                                          "ReadOnly": False
                                        }
                                      ],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    },
                                    {
                                      "$child": "Models.Functions.DivideFunction, Models",
                                      "Name": "LAR",
                                      "ResourceName": None,
                                      "Children": [
                                        {
                                          "$child": "Models.Functions.VariableReference, Models",
                                          "VariableName": "[Phenology].VegetativeThermalTime",
                                          "Name": "TT",
                                          "ResourceName": None,
                                          "Children": [],
                                          "Enabled": True,
                                          "ReadOnly": False
                                        },
                                        {
                                          "$child": "Models.Functions.VariableReference, Models",
                                          "VariableName": "[Leaf].Phyllochron",
                                          "Name": "Phyllochron",
                                          "ResourceName": None,
                                          "Children": [],
                                          "Enabled": True,
                                          "ReadOnly": False
                                        }
                                      ],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    }
                                  ],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.PhaseLookupValue, Models",
                              "Start": "EndCanopyDevelopment",
                              "End": "Maturity",
                              "Name": "Reproductive",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.Constant, Models",
                                  "FixedValue": 0.0,
                                  "Units": None,
                                  "Name": "Constant",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].SpecificArea",
                          "Name": "SLA",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Memo, Models",
                          "Text": "Growth in leaf area is used to calculate the leaf dry matter demand. The same was assumed in the original crop models in APSIM. Further model development can be taken once new experimental data for metabolic, structural and storage forms are provided.",
                          "Name": "Memo",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Memo, Models",
                          "Text": "DeltaLAI describes the LAI variation over the cycle and is calculated using the Area of Largest Leaf, relative leaf area, plant number, branch number and leaf appearance rate (thermal time/phyllochron).",
                          "Name": "Memo1",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStructuralPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QMetabolicPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStoragePriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientDemandFunctions, Models",
                  "Name": "NDemands",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].minimumNconc",
                          "Name": "MinNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].potentialDMAllocation.Structural",
                          "Name": "PotentialDMAllocation",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "MetabolicNconc",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Leaf].criticalNConc",
                              "Name": "CritNconc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Leaf].minimumNconc",
                              "Name": "MinNconc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].potentialDMAllocation.Structural",
                          "Name": "PotentialDMAllocation",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.DemandFunctions.StorageNDemandFunction, Models",
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 1.0,
                          "Units": None,
                          "Name": "NitrogenDemandSwitch",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].maximumNconc",
                          "Name": "MaxNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStructuralPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QMetabolicPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStoragePriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                  "Name": "FrostFraction",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.XYPairs, Models",
                      "X": [
                        -4.0,
                        4.0
                      ],
                      "Y": [
                        1.0,
                        0.0
                      ],
                      "XVariableName": None,
                      "Name": "XYPairs",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Weather].MinT",
                      "Name": "XValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                  "Name": "SpecificArea",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.XYPairs, Models",
                      "X": [
                        0.1,
                        6.0
                      ],
                      "Y": [
                        0.03,
                        0.03
                      ],
                      "XVariableName": None,
                      "Name": "XYPairs",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Leaf].LAI",
                      "Name": "XValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.006,
                  "Units": None,
                  "Name": "AreaLargestLeaf",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 45.0,
                  "Units": None,
                  "Name": "Phyllochron",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.Organs.ReproductiveOrgan, Models",
              "GrowthRespiration": 0.0,
              "MaintenanceRespiration": 0.0,
              "DMDemand": None,
              "NDemand": None,
              "DMSupply": None,
              "NSupply": None,
              "RipeStage": None,
              "potentialDMAllocation": None,
              "Live": {
                "$child": "Models.PMF.Biomass, Models",
                "Name": "Biomass",
                "ResourceName": None,
                "Children": [],
                "Enabled": True,
                "ReadOnly": False
              },
              "Dead": {
                "$child": "Models.PMF.Biomass, Models",
                "Name": "Biomass",
                "ResourceName": None,
                "Children": [],
                "Enabled": True,
                "ReadOnly": False
              },
              "Name": "Grain",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.5,
                  "Units": None,
                  "Name": "PotentialHarvestIndex",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.AddFunction, Models",
                  "Name": "FillingDuration",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].EarlyGrainFilling.Target",
                      "Name": "EarlyGrainFillingDuration",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].MidGrainFilling.Target",
                      "Name": "MidGrainFillingDuration",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].LateGrainFilling.Target",
                      "Name": "LateGrainFillingDuration",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhaseLookup, Models",
                  "Name": "DMDemandFunction",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.PhaseLookupValue, Models",
                      "Start": "StartGrainFilling",
                      "End": "EndGrainFill",
                      "Name": "GrainFilling",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.MultiplyFunction, Models",
                          "Name": "HIGrainDemand",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[AboveGround].Wt",
                              "Name": "AboveGroundWt",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.DivideFunction, Models",
                              "Name": "HarvestIndexIncrease",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Grain].PotentialHarvestIndex",
                                  "Name": "PotentialHarvestIndex",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Grain].FillingDuration",
                                  "Name": "FillingDuration",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Phenology].ThermalTime",
                              "Name": "ThermalTime",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.05,
                  "Units": None,
                  "Name": "MinimumNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.058,
                  "Units": None,
                  "Name": "MaximumNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.12,
                  "Units": None,
                  "Name": "WaterContent",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.MultiplyFunction, Models",
                  "Name": "NFillingRate",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Grain].MaximumNConc",
                      "Name": "NConc",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Grain].DMDemandFunction",
                      "Name": "DMDemand",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Grain].DMConversionEfficiency",
                      "Name": "DMConversionEfficiency",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Library.BiomassRemoval, Models",
                  "Name": "BiomassRemovalDefaults",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 1.0,
                      "FractionDeadToRemove": 1.0,
                      "FractionLiveToResidue": 0.0,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Harvest",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 1.0,
                      "FractionDeadToResidue": 1.0,
                      "Name": "Cut",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.89,
                  "Units": None,
                  "Name": "DMConversionEfficiency",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "RemobilisationCost",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.4,
                  "Units": None,
                  "Name": "CarbonConcentration",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.DivideFunction, Models",
                  "Name": "HarvestIndex",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Grain].Wt",
                      "Name": "GrainWt",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[AboveGround].Wt",
                      "Name": "AboveGroundWt",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "MaximumPotentialGrainSize",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "This parameter is not used.  Grain growth is calculated from harvest index increase.",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "NumberFunction",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "Grain number is not estimated by this model.",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientPoolFunctions, Models",
                  "Name": "DMDemandPriorityFactors",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientPoolFunctions, Models",
                  "Name": "NDemandPriorityFactors",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.Organs.Root, Models",
              "DMSupply": None,
              "NSupply": None,
              "DMDemand": None,
              "NDemand": None,
              "potentialDMAllocation": None,
              "GrowthRespiration": 0.0,
              "MaintenanceRespiration": 0.0,
              "Name": "Root",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Functions.MultiplyFunction, Models",
                  "Name": "RootFrontVelocity",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.PhaseLookup, Models",
                      "Name": "PotentialRootFrontVelocity",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.PhaseLookupValue, Models",
                          "Start": "Germination",
                          "End": "Emergence",
                          "Name": "PreEmergence",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 5.0,
                              "Units": None,
                              "Name": "Function",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.PhaseLookupValue, Models",
                          "Start": "Emergence",
                          "End": "StartGrainFilling",
                          "Name": "early",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 30.0,
                              "Units": None,
                              "Name": "Function",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.PhaseLookupValue, Models",
                          "Start": "StartGrainFilling",
                          "End": "Maturity",
                          "Name": "late",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 5.0,
                              "Units": None,
                              "Name": "Function",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "WaterFactor",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SoilWaterScale, Models",
                          "Name": "SoilWaterScale",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            0.25,
                            1.0
                          ],
                          "Y": [
                            1.0,
                            1.0,
                            1.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[SoilWaterScale]",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "AFPSFactor",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            0.03,
                            1.0
                          ],
                          "Y": [
                            1.0,
                            1.0,
                            1.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Root].AirFilledPoreSpace",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1000000.0,
                  "Units": None,
                  "Name": "MaximumRootDepth",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "There is no intrinsic limit to maximum rooting depth.  Roots will be limited by the depth of the soil profile, or via values of XF in each layer.",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.RootShape.RootShapeCylinder, Models",
                  "Name": "RootShape",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": "0-1",
                  "Name": "KLModifier",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "No modification to water extraction rate is assumed in this model.\r\n",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                  "Name": "MaxDailyNUptake",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.XYPairs, Models",
                      "X": [
                        0.0,
                        5.0
                      ],
                      "Y": [
                        1.0,
                        3.0
                      ],
                      "XVariableName": None,
                      "Name": "XYPairs",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Leaf].LAI",
                      "Name": "XValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.005,
                  "Units": "/d",
                  "Name": "SenescenceRate",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "Roots are assumed to simply senesce at a small and constant rate during the entire growth period.",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.01,
                  "Units": None,
                  "Name": "MaximumNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.005,
                  "Units": None,
                  "Name": "MinimumNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhaseLookupValue, Models",
                  "Start": "Emergence",
                  "End": "EndGrainFill",
                  "Name": "NitrogenDemandSwitch",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Constant",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.02,
                  "Units": None,
                  "Name": "KNO3",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "KNH4",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "The current model version considers only NO3 uptake but NH4 option can be activated by the user . We do not expect activation of NH4 to change total plant N uptake because the nitrification rate (conversion of NH4 to NO3) is very fast in the model. In the future we will re-valuate this. ",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Library.BiomassRemoval, Models",
                  "Name": "BiomassRemovalDefaults",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "Default values for biomass removals for Root are only estimates with no data behind them.\r\nThey are just to consider potential root senescence in response to damage.",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 1.0,
                      "FractionDeadToResidue": 1.0,
                      "Name": "Harvest",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 1.0,
                      "FractionDeadToResidue": 1.0,
                      "Name": "Cut",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                  "Name": "NUptakeSWFactor",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.XYPairs, Models",
                      "X": [
                        0.0,
                        1.0
                      ],
                      "Y": [
                        0.0,
                        1.0
                      ],
                      "XVariableName": None,
                      "Name": "XYPairs",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Root].RWC",
                      "Name": "XValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 40.0,
                  "Units": "m/g",
                  "Name": "SpecificRootLength",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": None,
                  "Name": "DMConversionEfficiency",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": None,
                  "Name": "MaintenanceRespirationFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "RemobilisationCost",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.4,
                  "Units": None,
                  "Name": "CarbonConcentration",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientDemandFunctions, Models",
                  "Name": "DMDemands",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.DemandFunctions.PartitionFractionDemandFunction, Models",
                          "Name": "DMDemandFunction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.PhaseLookup, Models",
                              "Name": "PartitionFraction",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.PhaseLookupValue, Models",
                                  "Start": "Emergence",
                                  "End": "StartFlowering",
                                  "Name": "Early",
                                  "ResourceName": None,
                                  "Children": [
                                    {
                                      "$child": "Models.Functions.Constant, Models",
                                      "FixedValue": 0.2,
                                      "Units": None,
                                      "Name": "Function",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    }
                                  ],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.PhaseLookupValue, Models",
                                  "Start": "StartFlowering",
                                  "End": "StartGrainFilling",
                                  "Name": "Middle",
                                  "ResourceName": None,
                                  "Children": [
                                    {
                                      "$child": "Models.Functions.Constant, Models",
                                      "FixedValue": 0.2,
                                      "Units": None,
                                      "Name": "Function",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    }
                                  ],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.PhaseLookupValue, Models",
                                  "Start": "StartGrainFilling",
                                  "End": "Maturity",
                                  "Name": "Late",
                                  "ResourceName": None,
                                  "Children": [
                                    {
                                      "$child": "Models.Functions.Constant, Models",
                                      "FixedValue": 0.05,
                                      "Units": None,
                                      "Name": "Function",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    }
                                  ],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 1.0,
                          "Units": None,
                          "Name": "StructuralFraction",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.DemandFunctions.StorageDMDemandFunction, Models",
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "StorageFraction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 1.0,
                              "Units": None,
                              "Name": "One",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Root].DMDemands.Structural.StructuralFraction",
                              "Name": "StructuralFraction",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStructuralPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QMetabolicPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStoragePriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientDemandFunctions, Models",
                  "Name": "NDemands",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Root].minimumNconc",
                          "Name": "MinNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Root].potentialDMAllocation.Structural",
                          "Name": "PotentialDMAllocation",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "MetabolicNconc",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Root].criticalNConc",
                              "Name": "CritNconc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Root].minimumNconc",
                              "Name": "MinNconc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Root].potentialDMAllocation.Structural",
                          "Name": "PotentialDMAllocation",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.DemandFunctions.StorageNDemandFunction, Models",
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Root].nitrogenDemandSwitch",
                          "Name": "NitrogenDemandSwitch",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Root].maximumNconc",
                          "Name": "MaxNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStructuralPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QMetabolicPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStoragePriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.VariableReference, Models",
                  "VariableName": "[Root].MinimumNConc",
                  "Name": "CriticalNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientPoolFunctions, Models",
                  "Name": "InitialWt",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.2,
                      "Units": "g/plant",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.Organs.Nodule, Models",
              "DMSupply": None,
              "NSupply": None,
              "DMDemand": None,
              "NDemand": None,
              "potentialDMAllocation": None,
              "Name": "Nodule",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Memo, Models",
                  "Text": "",
                  "Name": "Memo",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "FixationMetabolicCost",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.02,
                  "Units": None,
                  "Name": "MaximumNConc",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.01,
                  "Units": None,
                  "Name": "MinimumNConc",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "SenescenceRate",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "NReallocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhaseLookupValue, Models",
                  "Start": "Emergence",
                  "End": "EndGrainFill",
                  "Name": "NitrogenDemandSwitch",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Constant",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Library.BiomassRemoval, Models",
                  "Name": "BiomassRemovalDefaults",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "Default values for biomass removals for Root are only estimates with no data behind them.\r\nThey are just to consider potential root senescence in response to damage.",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.2,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Harvest",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.2,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Cut",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": "g/m^2",
                  "Name": "InitialWtFunction",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": "/d",
                  "Name": "DetachmentRateFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": "0-1",
                  "Name": "MaintenanceRespirationFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": "0-1",
                  "Name": "DMConversionEfficiency",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "NRetranslocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "DMReallocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "DMRetranslocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.VariableReference, Models",
                  "VariableName": "[Nodule].MinimumNConc",
                  "Name": "CriticalNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "RemobilisationCost",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.4,
                  "Units": None,
                  "Name": "CarbonConcentration",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.HoldFunction, Models",
                  "WhenToHold": "StartGrainFilling",
                  "Name": "FixationRate",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MinimumFunction, Models",
                      "Name": "DailyPotentialFixationRate",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.MultiplyFunction, Models",
                          "Name": "PotentialFixationRate",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[AboveGroundLive].Wt",
                              "Name": "LiveBiomass",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.PhaseLookup, Models",
                              "Name": "SpecificFixationRate",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.PhaseLookupValue, Models",
                                  "Start": "Emergence",
                                  "End": "StartFlowering",
                                  "Name": "VegetativeGrowth",
                                  "ResourceName": None,
                                  "Children": [
                                    {
                                      "$child": "Models.Functions.Constant, Models",
                                      "FixedValue": 0.004,
                                      "Units": None,
                                      "Name": "Rate",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    }
                                  ],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.PhaseLookupValue, Models",
                                  "Start": "StartFlowering",
                                  "End": "EndGrainFill",
                                  "Name": "ReproductiveGrowth",
                                  "ResourceName": None,
                                  "Children": [
                                    {
                                      "$child": "Models.Functions.Constant, Models",
                                      "FixedValue": 0.002,
                                      "Units": None,
                                      "Name": "Rate",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    }
                                  ],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 0.5,
                          "Units": "g/g",
                          "Name": "MaximumFixationRate",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientPoolFunctions, Models",
                  "Name": "DMDemands",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.DemandFunctions.PartitionFractionDemandFunction, Models",
                          "Name": "DMDemandFunction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.PhaseLookup, Models",
                              "Name": "PartitionFraction",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.PhaseLookupValue, Models",
                                  "Start": "Emergence",
                                  "End": "StartFlowering",
                                  "Name": "Early",
                                  "ResourceName": None,
                                  "Children": [
                                    {
                                      "$child": "Models.Functions.Constant, Models",
                                      "FixedValue": 0.03,
                                      "Units": None,
                                      "Name": "Function",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    }
                                  ],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.PhaseLookupValue, Models",
                                  "Start": "StartFlowering",
                                  "End": "StartGrainFilling",
                                  "Name": "Middle",
                                  "ResourceName": None,
                                  "Children": [
                                    {
                                      "$child": "Models.Functions.Constant, Models",
                                      "FixedValue": 0.02,
                                      "Units": None,
                                      "Name": "Function",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    }
                                  ],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.PhaseLookupValue, Models",
                                  "Start": "StartGrainFilling",
                                  "End": "Maturity",
                                  "Name": "Late",
                                  "ResourceName": None,
                                  "Children": [
                                    {
                                      "$child": "Models.Functions.Constant, Models",
                                      "FixedValue": 0.0,
                                      "Units": None,
                                      "Name": "Function",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    }
                                  ],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Memo, Models",
                                  "Text": "",
                                  "Name": "Memo",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 1.0,
                          "Units": "g/g",
                          "Name": "StructuralFraction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Memo, Models",
                              "Text": "",
                              "Name": "Memo",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.DemandFunctions.StorageDMDemandFunction, Models",
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "StorageFraction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 1.0,
                              "Units": None,
                              "Name": "One",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Nodule].DMDemands.Structural.StructuralFraction",
                              "Name": "StructuralFraction",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientPoolFunctions, Models",
                  "Name": "NDemands",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Nodule].minimumNconc",
                          "Name": "MinNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Nodule].potentialDMAllocation.Structural",
                          "Name": "PotentialDMAllocation",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "MetabolicNconc",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Nodule].criticalNConc",
                              "Name": "CritNconc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Nodule].minimumNconc",
                              "Name": "MinNconc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Nodule].potentialDMAllocation.Structural",
                          "Name": "PotentialDMAllocation",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.DemandFunctions.StorageNDemandFunction, Models",
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Nodule].nitrogenDemandSwitch",
                          "Name": "NitrogenDemandSwitch",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Nodule].maximumNconc",
                          "Name": "MaxNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.Organs.GenericOrgan, Models",
              "StartLive": None,
              "DMSupply": None,
              "NSupply": None,
              "DMDemand": None,
              "NDemand": None,
              "potentialDMAllocation": None,
              "IsAboveGround": True,
              "Name": "Shell",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": None,
                  "Name": "NReallocationFactor",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.5,
                  "Units": None,
                  "Name": "NRetranslocationFactor",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.045,
                  "Units": None,
                  "Name": "MaximumNConc",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.VariableReference, Models",
                  "VariableName": "[Shell].MinimumNConc",
                  "Name": "CriticalNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.01,
                  "Units": None,
                  "Name": "MinimumNConc",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.1,
                  "Units": None,
                  "Name": "DMRetranslocationFactor",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhaseLookupValue, Models",
                  "Start": "StartPodDevelopment",
                  "End": "EndPodDevelopment",
                  "Name": "NitrogenDemandSwitch",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Constant",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Library.BiomassRemoval, Models",
                  "Name": "BiomassRemovalDefaults",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 1.0,
                      "FractionDeadToResidue": 1.0,
                      "Name": "Harvest",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 1.0,
                      "FractionDeadToResidue": 1.0,
                      "Name": "Cut",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhaseLookup, Models",
                  "Name": "SenescenceRate",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.PhaseLookupValue, Models",
                      "Start": "StartPodDevelopment",
                      "End": "Maturity",
                      "Name": "ReproductivePhase",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.MinimumFunction, Models",
                          "Name": "Rate",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.DivideFunction, Models",
                              "Name": "Fraction",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Phenology].ThermalTime",
                                  "Name": "TT",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.SubtractFunction, Models",
                                  "Name": "TTRemaining",
                                  "ResourceName": None,
                                  "Children": [
                                    {
                                      "$child": "Models.Functions.ExpressionFunction, Models",
                                      "Expression": "[Phenology].EarlyPodDevelopment.Target+[Phenology].EarlyGrainFilling.Target+[Phenology].MidGrainFilling.Target+[Phenology].LateGrainFilling.Target",
                                      "Name": "ReproductiveTT",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    },
                                    {
                                      "$child": "Models.Functions.ExpressionFunction, Models",
                                      "Expression": "[Phenology].EarlyPodDevelopment.ProgressThroughPhase+[Phenology].EarlyGrainFilling.ProgressThroughPhase+[Phenology].MidGrainFilling.ProgressThroughPhase+[Phenology].LateGrainFilling.ProgressThroughPhase",
                                      "Name": "ReproductiveTT1Complete",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    }
                                  ],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 1.0,
                              "Units": None,
                              "Name": "One",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": "/d",
                  "Name": "DetachmentRateFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": "0-1",
                  "Name": "MaintenanceRespirationFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": "0-1",
                  "Name": "DMConversionEfficiency",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "DMReallocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "RemobilisationCost",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.4,
                  "Units": None,
                  "Name": "CarbonConcentration",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.2,
                  "Units": None,
                  "Name": "PotentialHarvestIndex",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.AddFunction, Models",
                  "Name": "FillingDuration",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].EarlyPodDevelopment.Target",
                      "Name": "EarlyPodDevelopmentDuration",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].EarlyGrainFilling.Target",
                      "Name": "EarlyGrainFillingDuration",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].MidGrainFilling.Target",
                      "Name": "MidGrainFillingDuration",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientDemandFunctions, Models",
                  "Name": "DMDemands",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.PhaseLookup, Models",
                          "Name": "DMDemandFunction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.PhaseLookupValue, Models",
                              "Start": "StartPodDevelopment",
                              "End": "EndPodDevelopment",
                              "Name": "PodDevelopment",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.MultiplyFunction, Models",
                                  "Name": "HIPodDemand",
                                  "ResourceName": None,
                                  "Children": [
                                    {
                                      "$child": "Models.Functions.VariableReference, Models",
                                      "VariableName": "[AboveGround].Wt",
                                      "Name": "AboveGroundWt",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    },
                                    {
                                      "$child": "Models.Functions.DivideFunction, Models",
                                      "Name": "HarvestIndexIncrease",
                                      "ResourceName": None,
                                      "Children": [
                                        {
                                          "$child": "Models.Functions.VariableReference, Models",
                                          "VariableName": "[Shell].PotentialHarvestIndex",
                                          "Name": "PotentialHarvestIndex",
                                          "ResourceName": None,
                                          "Children": [],
                                          "Enabled": True,
                                          "ReadOnly": False
                                        },
                                        {
                                          "$child": "Models.Functions.VariableReference, Models",
                                          "VariableName": "[Shell].FillingDuration",
                                          "Name": "FillingDuration",
                                          "ResourceName": None,
                                          "Children": [],
                                          "Enabled": True,
                                          "ReadOnly": False
                                        }
                                      ],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    },
                                    {
                                      "$child": "Models.Functions.VariableReference, Models",
                                      "VariableName": "[Phenology].ThermalTime",
                                      "Name": "ThermalTime",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    }
                                  ],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 1.0,
                          "Units": None,
                          "Name": "StructuralFraction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Memo, Models",
                              "Text": "",
                              "Name": "Memo",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Memo, Models",
                          "Text": "Shell harvest index is used to calculate shell biomass demand, which follows the same approach than grain harvest index. It represents the fraction of shell biomass in relation to total aboveground biomass.",
                          "Name": "Memo",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.DemandFunctions.StorageDMDemandFunction, Models",
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "StorageFraction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 1.0,
                              "Units": None,
                              "Name": "One",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Shell].DMDemands.Structural.StructuralFraction",
                              "Name": "StructuralFraction",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStructuralPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QMetabolicPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStoragePriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientDemandFunctions, Models",
                  "Name": "NDemands",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Shell].minimumNconc",
                          "Name": "MinNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Shell].potentialDMAllocation.Structural",
                          "Name": "PotentialDMAllocation",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.DemandFunctions.StorageNDemandFunction, Models",
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Shell].nitrogenDemandSwitch",
                          "Name": "NitrogenDemandSwitch",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Shell].maximumNconc",
                          "Name": "MaxNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStructuralPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QMetabolicPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStoragePriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.RetranslocateNonStructural, Models",
                  "Name": "RetranslocateNitrogen",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientPoolFunctions, Models",
                  "Name": "InitialWt",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": "g/m^2",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "Photosynthesis",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.VariableReference, Models",
                  "VariableName": "[Shell].MinimumNConc",
                  "Name": "initialNConcFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.Organs.GenericOrgan, Models",
              "StartLive": None,
              "DMSupply": None,
              "NSupply": None,
              "DMDemand": None,
              "NDemand": None,
              "potentialDMAllocation": None,
              "IsAboveGround": True,
              "Name": "Stem",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Memo, Models",
                  "Text": "",
                  "Name": "Memo",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": None,
                  "Name": "NReallocationFactor",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.5,
                  "Units": None,
                  "Name": "NRetranslocationFactor",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhaseLookup, Models",
                  "Name": "DMRetranslocationFactor",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.PhaseLookupValue, Models",
                      "Start": "EndCanopyDevelopment",
                      "End": "EndGrainFill",
                      "Name": "GrainFilling",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 0.2,
                          "Units": None,
                          "Name": "RetranslocationFactor",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                  "Name": "MaximumNConc",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.XYPairs, Models",
                      "X": [
                        3.0,
                        4.0,
                        5.0
                      ],
                      "Y": [
                        0.025,
                        0.02,
                        0.015
                      ],
                      "XVariableName": None,
                      "Name": "XYPairs",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].Stage",
                      "Name": "XValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.006,
                  "Units": None,
                  "Name": "MinimumNConc",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhaseLookupValue, Models",
                  "Start": "Emergence",
                  "End": "StartGrainFilling",
                  "Name": "NitrogenDemandSwitch",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "There is no evidence of stem N storage during grain filling. Also, there is no stem structural growth during this time.",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "Constant",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Library.BiomassRemoval, Models",
                  "Name": "BiomassRemovalDefaults",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 1.0,
                      "FractionDeadToResidue": 1.0,
                      "Name": "Harvest",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 1.0,
                      "FractionDeadToResidue": 1.0,
                      "Name": "Cut",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhaseLookup, Models",
                  "Name": "SenescenceRate",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.PhaseLookupValue, Models",
                      "Start": "EndCanopyDevelopment",
                      "End": "Maturity",
                      "Name": "PostCanopyPhase",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.MinimumFunction, Models",
                          "Name": "Rate",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.DivideFunction, Models",
                              "Name": "Fraction",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[Phenology].ThermalTime",
                                  "Name": "TT",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.SubtractFunction, Models",
                                  "Name": "TTRemaining",
                                  "ResourceName": None,
                                  "Children": [
                                    {
                                      "$child": "Models.Functions.ExpressionFunction, Models",
                                      "Expression": "[Phenology].EarlyPodDevelopment.Target+[Phenology].EarlyGrainFilling.Target+[Phenology].MidGrainFilling.Target+[Phenology].LateGrainFilling.Target",
                                      "Name": "PostCanopyTT",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    },
                                    {
                                      "$child": "Models.Functions.ExpressionFunction, Models",
                                      "Expression": "[Phenology].MidGrainFilling.ProgressThroughPhase+[Phenology].LateGrainFilling.ProgressThroughPhase",
                                      "Name": "PostCanopyTTComplete",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    }
                                  ],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 1.0,
                              "Units": None,
                              "Name": "One",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": "/d",
                  "Name": "DetachmentRateFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": "0-1",
                  "Name": "MaintenanceRespirationFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": "0-1",
                  "Name": "DMConversionEfficiency",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "DMReallocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.VariableReference, Models",
                  "VariableName": "[Stem].MinimumNConc",
                  "Name": "CriticalNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "RemobilisationCost",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.4,
                  "Units": None,
                  "Name": "CarbonConcentration",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientDemandFunctions, Models",
                  "Name": "DMDemands",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.PhaseLookup, Models",
                          "Name": "Fraction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Memo, Models",
                              "Text": "",
                              "Name": "memo",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.PhaseLookupValue, Models",
                              "Start": "Emergence",
                              "End": "StartPodDevelopment",
                              "Name": "StemGrowthPhase",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.Constant, Models",
                                  "FixedValue": 0.75,
                                  "Units": None,
                                  "Name": "Fraction",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.PhaseLookupValue, Models",
                              "Start": "StartPodDevelopment",
                              "End": "EndGrainFill",
                              "Name": "PodAndGrainGrowth",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.Constant, Models",
                                  "FixedValue": 0.0,
                                  "Units": None,
                                  "Name": "Fraction",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Stem].PotentialGrowth",
                          "Name": "DMDemand",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Stem].PotentialGrowth",
                          "Name": "DMDemand",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "StorageFraction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 1.0,
                              "Units": None,
                              "Name": "One",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Stem].DMDemands.Structural.Fraction",
                              "Name": "StructuralFraction",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStructuralPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QMetabolicPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStoragePriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientDemandFunctions, Models",
                  "Name": "NDemands",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Stem].minimumNconc",
                          "Name": "MinNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Stem].potentialDMAllocation.Structural",
                          "Name": "PotentialDMAllocation",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "MetabolicNconc",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Stem].criticalNConc",
                              "Name": "CritNconc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Stem].minimumNconc",
                              "Name": "MinNconc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Stem].potentialDMAllocation.Structural",
                          "Name": "PotentialDMAllocation",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.DemandFunctions.StorageNDemandFunction, Models",
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Stem].nitrogenDemandSwitch",
                          "Name": "NitrogenDemandSwitch",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Stem].maximumNconc",
                          "Name": "MaxNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStructuralPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QMetabolicPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStoragePriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.DemandFunctions.PartitionFractionDemandFunction, Models",
                  "Name": "PotentialGrowth",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.PhaseLookup, Models",
                      "Name": "PartitionFraction",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Memo, Models",
                          "Text": "",
                          "Name": "memo",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.PhaseLookupValue, Models",
                          "Start": "Emergence",
                          "End": "EndPodDevelopment",
                          "Name": "StemGrowthPhase",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 0.4,
                              "Units": None,
                              "Name": "StemFraction",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.PhaseLookupValue, Models",
                          "Start": "EndPodDevelopment",
                          "End": "EndGrainFill",
                          "Name": "GrainFilling",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 0.0,
                              "Units": None,
                              "Name": "StemFraction",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.RetranslocateNonStructural, Models",
                  "Name": "RetranslocateNitrogen",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientPoolFunctions, Models",
                  "Name": "InitialWt",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": "g/m^2",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "Photosynthesis",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.VariableReference, Models",
                  "VariableName": "[Stem].MinimumNConc",
                  "Name": "initialNConcFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.CompositeBiomass, Models",
              "OrganNames": [
                "Leaf",
                "Stem",
                "Grain",
                "Shell"
              ],
              "IncludeLive": True,
              "IncludeDead": True,
              "Name": "AboveGround",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.CompositeBiomass, Models",
              "OrganNames": [
                "Root",
                "Nodule"
              ],
              "IncludeLive": True,
              "IncludeDead": True,
              "Name": "BelowGround",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.CompositeBiomass, Models",
              "OrganNames": [
                "Leaf",
                "Stem",
                "Grain",
                "Shell"
              ],
              "IncludeLive": True,
              "IncludeDead": False,
              "Name": "AboveGroundLive",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.CompositeBiomass, Models",
              "OrganNames": [
                "Leaf",
                "Stem",
                "Grain",
                "Shell",
                "Root",
                "Nodule"
              ],
              "IncludeLive": True,
              "IncludeDead": True,
              "Name": "Total",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.CompositeBiomass, Models",
              "OrganNames": [
                "Leaf",
                "Stem",
                "Grain",
                "Shell",
                "Root",
                "Nodule"
              ],
              "IncludeLive": True,
              "IncludeDead": False,
              "Name": "TotalLive",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.CompositeBiomass, Models",
              "OrganNames": [
                "Leaf",
                "Stem",
                "Grain",
                "Shell",
                "Root",
                "Nodule"
              ],
              "IncludeLive": False,
              "IncludeDead": True,
              "Name": "TotalDead",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.CompositeBiomass, Models",
              "OrganNames": [
                "Shell",
                "Grain"
              ],
              "IncludeLive": True,
              "IncludeDead": True,
              "Name": "Pod",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.Functions.Constant, Models",
              "FixedValue": 0.0,
              "Units": None,
              "Name": "MortalityRate",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.Functions.Constant, Models",
              "FixedValue": 0.0,
              "Units": None,
              "Name": "SeedMortalityRate",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.Core.Folder, Models",
              "ShowInDocs": False,
              "GraphsPerPage": 6,
              "Name": "Cultivars",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Core.Folder, Models",
                  "ShowInDocs": False,
                  "GraphsPerPage": 6,
                  "Name": "Australia",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 480",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 100",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.015",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 467",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 12.39, 15.61",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 12.39, 15.61",
                        "[Phenology].MidGrainFilling.Target.FractionofMidToLateGrainfilling.FixedValue=0.33",
                        "[Grain].MaximumNConc.FixedValue = 0.067",
                        "[Stem].PotentialGrowth.PartitionFraction.StemGrowthPhase.StemFraction.FixedValue = 0.5",
                        "[Leaf].AreaLargestLeaf.FixedValue=0.009"
                      ],
                      "Name": "Buchanan",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 380",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 100",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.05",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 405",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 13.75, 17",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 11.9, 15.5",
                        "[Leaf].AreaLargestLeaf.FixedValue=0.009"
                      ],
                      "Name": "Djakal",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 480",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 100",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.015",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 467",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 12.39, 15.61",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 12.39, 15.61",
                        "[Phenology].MidGrainFilling.Target.FractionofMidToLateGrainfilling.FixedValue=0.33",
                        "[Grain].MaximumNConc.FixedValue = 0.067",
                        "[Stem].PotentialGrowth.PartitionFraction.StemGrowthPhase.StemFraction.FixedValue = 0.5",
                        "[Leaf].AreaLargestLeaf.FixedValue=0.012"
                      ],
                      "Name": "Davis",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 380",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 80",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.05",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 440",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 13.26, 16.67",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 13.26, 16.67",
                        "[Leaf].AreaLargestLeaf.FixedValue=0.009"
                      ],
                      "Name": "Hooper_MG40",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 360",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 80",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.05",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 440",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 13.26, 16.67",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 13.26, 16.67",
                        "[Leaf].AreaLargestLeaf.FixedValue=0.009"
                      ],
                      "Name": "Stephens_MG40",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 285",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 80",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.05",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 405",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 13.75, 18",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 13.75, 18"
                      ],
                      "Name": "FiskebyV",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Memo, Models",
                          "Text": "Early-maturing, photoperiod-insensitive cultivar developed in Sweden for high latitude environments",
                          "Name": "Memo",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 314",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 80",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.05",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 485",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 13.75, 17",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 12, 18",
                        "[Leaf].AreaLargestLeaf.FixedValue=0.009"
                      ],
                      "Name": "F148_7",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Memo, Models",
                          "Text": "Indeterminate breeding selection from the national soybean breeding program combining phytophthora resistance, earliness,\nculinary quality, and yield",
                          "Name": "Memo",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 377",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 80",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.05",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 405",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 13.75, 15.5",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 11.9, 15.5"
                      ],
                      "Name": "Bowyer",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Memo, Models",
                          "Text": "Brown hilum cultivar, adapted to southern New South Wales, released by K McWhirter in 1982",
                          "Name": "Memo",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 550",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 80",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.05",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 405",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 13.75, 18",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 10.9, 15.5",
                        "[Leaf].AreaLargestLeaf.FixedValue=0.009"
                      ],
                      "Name": "Bunya",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Memo, Models",
                          "Text": "Culinary cultivar released in 2006 by CSIRO with broad adaptation from northern New South Wales to northern Queensland",
                          "Name": "Memo",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 390",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 80",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.05",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 425",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 12.75, 16",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 12.75, 16",
                        "[Leaf].AreaLargestLeaf.FixedValue=0.009"
                      ],
                      "Name": "Cowrie",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Memo, Models",
                          "Text": "Light hilum cultivar released by theNewSouth Wales Department of Agriculture for the coastal Northern Rivers District inNewSouth Wales",
                          "Name": "Memo",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 430",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 80",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.05",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 425",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 12.75, 15.3",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 12.75, 15.3",
                        "[Leaf].AreaLargestLeaf.FixedValue=0.009"
                      ],
                      "Name": "Soya791",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 440",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 80",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.05",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 425",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 12.75, 15",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 12.75, 15",
                        "[Leaf].AreaLargestLeaf.FixedValue=0.009"
                      ],
                      "Name": "Manark",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 490",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 80",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.05",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 425",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 12.75, 15",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 12.75, 15",
                        "[Leaf].AreaLargestLeaf.FixedValue=0.009"
                      ],
                      "Name": "Warrigal",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 806",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 80",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.05",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 425",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 12.75, 14.7",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 12.75, 14.7",
                        "[Leaf].AreaLargestLeaf.FixedValue=0.009"
                      ],
                      "Name": "Leichhardt",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 332",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 80",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.05",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 440",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 13.26, 16.67",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 13.26, 16.67",
                        "[Leaf].AreaLargestLeaf.FixedValue=0.009"
                      ],
                      "Name": "Djakal1",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Core.Folder, Models",
                  "ShowInDocs": False,
                  "GraphsPerPage": 6,
                  "Name": "USA",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 268",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 76",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.05",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 432",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 14.43, 21.19",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 14.43, 21.19"
                      ],
                      "Name": "Trial_MG00",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 294",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 81",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.05",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 442",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 14.29, 20.14",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 14.29, 20.14"
                      ],
                      "Name": "Lambert_MG0",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 336",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 118",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.05",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 518",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 14.37, 19.30",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 14.37, 19.30"
                      ],
                      "Name": "IA1006_MG10",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 329",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 150",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.05",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 580",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 13.92, 17.93",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 13.92, 17.93"
                      ],
                      "Name": "IA2008_MG20",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 328",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 106",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.3",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 499",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 13.59, 17.6",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 13.59, 17.6"
                      ],
                      "Name": "PioneerP22T61_MG22",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 328",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 106",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.35",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 580",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 13.59, 17.6",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 13.59, 17.6"
                      ],
                      "Name": "PioneerP22T69R_MG22",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 271",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 120",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.125",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 529",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 13.77, 17.57",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 13.77, 17.57"
                      ],
                      "Name": "Pioneer92MGI_MG26",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 328",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 106",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.14",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 499",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 13.59, 17.6",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 13.59, 17.6"
                      ],
                      "Name": "PioneerP92Y75_MG27",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 367",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 163",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.175",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 607",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 14.05, 17.56",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 14.05, 17.56"
                      ],
                      "Name": "Macon_MG30",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 250",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 110",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.30",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 457",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 13.4, 16.0",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 13.4, 16.0",
                        "[Grain].MaximumNConc.FixedValue = 0.065",
                        "[Shell].PotentialHarvestIndex.FixedValue = 0.5",
                        "[Leaf].AreaLargestLeaf.FixedValue=0.013"
                      ],
                      "Name": "PioneerP93M11_MG31",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 272",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 117",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.20",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 522",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 13.45, 18.0",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 13.45, 18.0"
                      ],
                      "Name": "Becks321NRR_MG32",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 328",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 106",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.40",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 499",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 13.59, 17.6",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 13.59, 17.6"
                      ],
                      "Name": "PioneerP932T16R_MG32",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 267",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 112",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.26",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 506",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 13.27, 17.2",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 13.27, 17.2"
                      ],
                      "Name": "Becks367NRR_MG37",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 292",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 123",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.30",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 535",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 13.09, 16.49",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 13.09, 16.49"
                      ],
                      "Name": "AsgrowAG4403_MG40",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 292",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 123",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.30",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 535",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 13.09, 16.49",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 13.09, 16.49"
                      ],
                      "Name": "HornbeckHBK4891_MG40",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 322",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 143",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.30",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 567",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 13.05, 16.46",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 13.05, 16.46"
                      ],
                      "Name": "Pioneer94B01_MG40",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 404",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 112",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.30",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 505",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 12.83, 16.13",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 12.83, 16.13"
                      ],
                      "Name": "PioneerP9504_MG50",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 404",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 131",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.30",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 549",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 12.83, 16.13",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 12.83, 16.13"
                      ],
                      "Name": "AsgrowAG5701_MG50",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 438",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 112",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.30",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 505",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 12.58, 15.88",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 12.58, 15.88"
                      ],
                      "Name": "Hutcheson_MG50",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 436",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 93",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.30",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 467",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 12.39, 15.61",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 12.39, 15.61"
                      ],
                      "Name": "NK622_MG60",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Core.Folder, Models",
                  "ShowInDocs": False,
                  "GraphsPerPage": 6,
                  "Name": "China",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 500",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 120",
                        "[Phenology].EarlyPodDevelopment.Target.FixedValue = 180",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.2",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 440",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 12.63, 15.13",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 12.63, 15.13",
                        "[Phenology].MidGrainFilling.Target.FractionofMidToLateGrainfilling.FixedValue=0.45",
                        "[Leaf].AreaLargestLeaf.FixedValue=0.013",
                        "[Grain].PotentialHarvestIndex.FixedValue=0.35"
                      ],
                      "Name": "Nandou12",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 500",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 200",
                        "[Phenology].EarlyPodDevelopment.Target.FixedValue = 170",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.2",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 460",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 12.63, 15.13",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 12.63, 15.13",
                        "[Phenology].MidGrainFilling.Target.FractionofMidToLateGrainfilling.FixedValue=0.45",
                        "[Leaf].AreaLargestLeaf.FixedValue=0.013",
                        "[Grain].PotentialHarvestIndex.FixedValue=0.35"
                      ],
                      "Name": "Texuan13",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 540",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 140",
                        "[Phenology].EarlyPodDevelopment.Target.FixedValue = 140",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.2",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 420",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 12.94, 15.44",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 12.94, 15.44",
                        "[Phenology].MidGrainFilling.Target.FractionofMidToLateGrainfilling.FixedValue=0.45",
                        "[Leaf].AreaLargestLeaf.FixedValue=0.013",
                        "[Grain].PotentialHarvestIndex.FixedValue=0.25"
                      ],
                      "Name": "Jiuyuehuang",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 370",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 120",
                        "[Phenology].EarlyPodDevelopment.Target.FixedValue = 180",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.361",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 600",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 13.57, 16.07",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 13.57, 16.07",
                        "[Phenology].MidGrainFilling.Target.FractionofMidToLateGrainfilling.FixedValue=0.45",
                        "[Leaf].AreaLargestLeaf.FixedValue=0.02",
                        "[Leaf].Phyllochron.FixedValue=45"
                      ],
                      "Name": "Hedou19",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Core.Folder, Models",
                  "ShowInDocs": False,
                  "GraphsPerPage": 6,
                  "Name": "Generic",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 310",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 100",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.475",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 590",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 14.6, 22.35",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 14.6, 22.35"
                      ],
                      "Name": "Generic_MG000",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 320",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 100",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.467",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 600",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 14.35, 21.11",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 14.35, 21.11"
                      ],
                      "Name": "Generic_MG00",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 336",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 120",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.422",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 616",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 14.1, 19.95",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 14.1, 19.95"
                      ],
                      "Name": "Generic_MG0",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 340",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 120",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.411",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 632",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 13.84, 18.77",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 13.84, 18.77"
                      ],
                      "Name": "Generic_MG1",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 348",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 120",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.386",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 648",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 13.59, 17.61",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 13.59, 17.61"
                      ],
                      "Name": "Generic_MG2",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 380",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 120",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.361",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 664",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 13.4, 16.91",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 13.4, 16.91"
                      ],
                      "Name": "Generic_MG3",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 388",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 140",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.324",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 664",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 13.1, 16.49",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 13.1, 16.49"
                      ],
                      "Name": "Generic_MG4",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 396",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 160",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.072",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 696",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 12.83, 16.13",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 12.83, 16.13"
                      ],
                      "Name": "Generic_MG5",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 404",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 180",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.056",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 712",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 12.58, 15.8",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 12.58, 15.8"
                      ],
                      "Name": "Generic_MG6",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 416",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 200",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.055",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 728",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 12.33, 15.46",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 12.33, 15.46"
                      ],
                      "Name": "Generic_MG7",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 430",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 200",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.054",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 744",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 12.07, 15.1",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 12.07, 15.1"
                      ],
                      "Name": "Generic_MG8",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 460",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 200",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.053",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 748",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 11.88, 14.82",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 11.88, 14.82"
                      ],
                      "Name": "Generic_MG9",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Cultivar, Models",
                      "Command": [
                        "[Phenology].Vegetative.Target.FixedValue = 470",
                        "[Phenology].EarlyFlowering.Target.FixedValue = 200",
                        "[Phenology].EarlyGrainFilling.Target.FractionofGrainfilling.FixedValue = 0.053",
                        "[Phenology].LateGrainFilling.Target.EntireGrainfillPeriod.FixedValue = 748",
                        "[Phenology].VegetativePhotoperiodModifier.XYPairs.X = 11.78, 14.65",
                        "[Phenology].ReproductivePhotoperiodModifier.XYPairs.X = 11.78, 14.65"
                      ],
                      "Name": "Generic_MG10",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            }
          ],
          "Enabled": True,
          "ReadOnly": False
        },
        {
          "$child": "Models.PMF.Plant, Models",
          "PlantType": "TropicalPasture",
          "IsEnding": False,
          "DaysAfterEnding": 0,
          "Name": "TropicalPasture",
          "ResourceName": None,
          "Children": [
            {
              "$child": "Models.Functions.AccumulateFunction, Models",
              "StartStageName": "Germination",
              "EndStageName": "Emergence",
              "ResetStageName": None,
              "FractionRemovedOnCut": 0.0,
              "FractionRemovedOnHarvest": 0.0,
              "FractionRemovedOnGraze": 0.0,
              "FractionRemovedOnPrune": 0.0,
              "Name": "Age",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 49.0,
                  "Units": None,
                  "Name": "day",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.OrganArbitrator, Models",
              "Name": "Arbitrator",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.PMF.BiomassTypeArbitrator, Models",
                  "Name": "DMArbitration",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Core.Folder, Models",
                      "ShowInDocs": False,
                      "GraphsPerPage": 6,
                      "Name": "PotentialPartitioningMethods",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.PMF.Arbitrator.ReallocationMethod, Models",
                          "Name": "ReallocationMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.PMF.Arbitrator.AllocateFixationMethod, Models",
                          "Name": "AllocateFixationMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.PMF.Arbitrator.RetranslocationMethod, Models",
                          "Name": "RetranslocationMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.PMF.Arbitrator.SendPotentialDMAllocationsMethod, Models",
                          "Name": "SendPotentialDMAllocationsMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Core.Folder, Models",
                      "ShowInDocs": False,
                      "GraphsPerPage": 6,
                      "Name": "AllocationMethods",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.PMF.Arbitrator.NutrientConstrainedAllocationMethod, Models",
                          "Name": "NutrientConstrainedAllocationMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.PMF.Arbitrator.DryMatterAllocationsMethod, Models",
                          "Name": "DryMatterAllocationsMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.RelativeAllocation, Models",
                      "Name": "ArbitrationMethod",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.BiomassTypeArbitrator, Models",
                  "Name": "NArbitration",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Core.Folder, Models",
                      "ShowInDocs": False,
                      "GraphsPerPage": 6,
                      "Name": "PotentialPartitioningMethods",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.PMF.Arbitrator.ReallocationMethod, Models",
                          "Name": "ReallocationMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Core.Folder, Models",
                      "ShowInDocs": False,
                      "GraphsPerPage": 6,
                      "Name": "ActualPartitioningMethods",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.PMF.Arbitrator.AllocateFixationMethod, Models",
                          "Name": "AllocateFixationMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.PMF.Arbitrator.RetranslocationMethod, Models",
                          "Name": "RetranslocationMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Core.Folder, Models",
                      "ShowInDocs": False,
                      "GraphsPerPage": 6,
                      "Name": "AllocationMethods",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.PMF.Arbitrator.NitrogenAllocationsMethod, Models",
                          "Name": "NitrogenAllocationsMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.RelativeAllocation, Models",
                      "Name": "ArbitrationMethod",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.Arbitrator.AllocateUptakesMethod, Models",
                      "Name": "AllocateUptakesMethod",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Arbitrator.WaterUptakeMethod, Models",
                  "Name": "WaterUptakeMethod",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Arbitrator.NitrogenUptakeMethod, Models",
                  "Name": "NitrogenUptakeMethod",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.Functions.PhotoperiodFunction, Models",
              "Twilight": -6.0,
              "DayLength": 0.0,
              "Name": "Photoperiod",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.Functions.PhotoperiodDeltaFunction, Models",
              "Twilight": -6.0,
              "Name": "PhotoperiodVariation",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.Phen.Phenology, Models",
              "Name": "Phenology",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Functions.MultiplyFunction, Models",
                  "Name": "ThermalTime",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.HourlyInterpolation, Models",
                      "agregationMethod": 0,
                      "Name": "BaseThermalTime",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            25.0,
                            45.0
                          ],
                          "Y": [
                            0.0,
                            13.0,
                            0.0
                          ],
                          "XVariableName": None,
                          "Name": "Response",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.ThreeHourSin, Models",
                          "TempRangeFactors": None,
                          "Name": "InterpolationMethod",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Memo, Models",
                  "Text": "The duration of each phenological phase in white clover is controlled either by the accumulation of thermal time or photoperiod.",
                  "Name": "Description",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.GerminatingPhase, Models",
                  "Start": "Sowing",
                  "End": "Germination",
                  "Name": "Germinating",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "ThermalTime",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "Germination in most clover species is very variable (e.g. [HamptonEtAl1987]; [MurrayThesis2012]), with about 70-90% of seeds germinating for white clover. Temperatures between 5 and 20^o^C affect the speed of germination but not the final percentage ([HamptonEtAl1987]).  The model does not consider variable germination rate, thus the population should be adjusted by the user when sowing to account for less than 100% germination.  \r\n\r\n",
                      "Name": "Description",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.EmergingPhase, Models",
                  "Start": "Germination",
                  "End": "Emergence",
                  "TTForTimeStep": 0.0,
                  "Name": "Emerging",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "White clover has small seeds and these need to be placed near the soil surface to emerge successfully.  Best emergence rates occur when seeds are sown at depth of 5mm ([Hyde1950]; [MurrayAndClements1993]), and few plants will emerge from seeds deeper than 40mm.  Variations in germination rate due to depth are not currently simulated by the model, however the sowing depth does affect timing of emergence, in combination with temperature.  This is controlled by the parameter ShootRate, which is the shoot elongation rate as function of thermal time (mm/oCd).  \r\nEmergence of white clover occurs a few days after germination ([Hyde1950]; [MootEtAll2000]), but this is quite variable (taking 5-20 days).  Both temperature and depth of seed are important factors.  Thermal time for the emergence of white clover (cv. Huia and Nomad) has been reported as being approximately 150oCd ([MootEtAll2000]; [MurrayThesis2012]), and there seem to be no significant variations between NZ cultivars ([HamptonEtAl1987]).  \r\n",
                      "Name": "Description",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "ThermalTime",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 400.0,
                      "Units": "oCd",
                      "Name": "Target",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.GenericPhase, Models",
                  "Start": "Emergence",
                  "End": "Induction",
                  "Name": "Vegetative",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 3200.0,
                      "Units": "oCd",
                      "Name": "MaxThermalTimeToBolt",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "During this phase, white clover plants partition biomass to leaf+petiole, stolon, and root+taproot organs. Initially, this phase starts when plants emerge, but it is also triggered as phenology resets after the end of the reproductive phase.  The vegetative phase ends when the inductive phase starts.  \r\nWhite clover is typically described as either short day or short-long day species ([Thomas1961]; [ChapterThomas1980]), which means induction is caused by exposure to short daylight lengths but reproductive growth may also be triggered by a period with long photoperiod.  Exposure to cold temperature can also lead to induction with subsequent initiation of inflorescence development when days become warmer and longer ([Thomas1981_SLD]; [BeattyThesis1959]). The variability in response of induction to environmental conditions are considerably large in white clover. In most stands at least some plants will show reproductive growth regardless of environmental conditions. The variations show some trend when contrasting large- to short-leaved cultivars or for cultivars from different origins (especially latitude), but the effect of genotypes within each group was not found to be significant (e.g. [ChapterThomas1980]; [Thomas1982_Comparison]; [Norris1984_daylength]; [BeattyThesis1959]).  \r\nInducing daylight length seem to be 12 hours or less, while temperatures need to be below about 10^o^C ([Thomas1979_Natural]; [Thomas1981_SLD]; [Thomas1982_Comparison]; [Norris1984_daylength]; [Norris1985_Temperature]). The base white clover model uses the accumulation of days under either low temperature or short photoperiod to trigger the end of the vegetative phase.\r\n",
                      "Name": "Description",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 25.0,
                      "Units": "days",
                      "Name": "Target",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.AddFunction, Models",
                      "Name": "Progression",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.LinearInterpolationFunction, Models",
                          "Name": "ShortDays",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.XYPairs, Models",
                              "X": [
                                13.0,
                                16.0
                              ],
                              "Y": [
                                1.0,
                                1.0
                              ],
                              "XVariableName": None,
                              "Name": "XYPairs",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                                  "Name": "ShortDays",
                                  "ResourceName": None,
                                  "Children": [
                                    {
                                      "$child": "Models.Functions.XYPairs, Models",
                                      "X": [
                                        13.0,
                                        16.0
                                      ],
                                      "Y": [
                                        1.0,
                                        1.0
                                      ],
                                      "XVariableName": None,
                                      "Name": "XYPairs",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    },
                                    {
                                      "$child": "Models.Functions.VariableReference, Models",
                                      "VariableName": "[Phenology].Photoperiod",
                                      "Name": "XValue",
                                      "ResourceName": None,
                                      "Children": [],
                                      "Enabled": True,
                                      "ReadOnly": False
                                    }
                                  ],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Phenology].Photoperiod",
                              "Name": "XValue",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.LinearInterpolationFunction, Models",
                          "Name": "LowTemperatures",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.XYPairs, Models",
                              "X": [
                                0.0,
                                2.0,
                                6.5,
                                8.0,
                                11.0,
                                15.0
                              ],
                              "Y": [
                                0.0,
                                0.0,
                                1.0,
                                1.0,
                                0.0,
                                0.0
                              ],
                              "XVariableName": None,
                              "Name": "XYPairs",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Phenology].ThermalTime",
                              "Name": "XValue",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "ThermalTime",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.GenericPhase, Models",
                  "Start": "Induction",
                  "End": "Flowering",
                  "Name": "Inductive",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "During this phase the switch plants await for the environmental cue to start the reproductive growth. Physiological and metabolic processes remain the same as for the vegetative phase; plants partition biomass to leaf+petiole, stolon, and root+taproot organs. The inductive phase starts when plants accumulate a given number of days under inductive conditions, i.e. low temperatures and/or short daylight lengths ([Thomas1979_Natural]; [Norris1984_daylength]; [Norris1985_Temperature]).  The phase ends when the plants start reproduction growth, with the emergence of flower buds.  \r\nThere is a large level of variation in white clover plants and their flowering response to environmental conditions, with some plants reaching reproductive growth in almost any stand ([Thomas1979_Natural]; [BeattyThesis1959]). Nonetheless, in natural conditions reproductive growth is typically triggered by long daylight lengths, with reported values around 14 hours ([ChapterThomas1980]; [Thomas1981_SLD];[Norris1984_daylength]).  The accumulation of days under long photoperiod is used by the model to trigger the end of the inductive phase.  \r\n",
                      "Name": "Description",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 60.0,
                      "Units": "days",
                      "Name": "Target",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "Progression",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            10.0,
                            11.5,
                            14.0,
                            15.0,
                            16.0
                          ],
                          "Y": [
                            0.0,
                            0.0,
                            1.0,
                            1.0,
                            1.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Phenology].Photoperiod",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.GenericPhase, Models",
                  "Start": "Flowering",
                  "End": "Ripening",
                  "Name": "Reproductive",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "During this phase the plant is partitioning biomass to all organs, including reproductive organs (peduncles, inflorescence, and seeds). The current model does not describe the flower and seeds development explicitly, only a generic organ referred to as *inflorescence* is defined. The main reason for this is the variability and complex phenology of reproduction of white clover; with seeds maturing while new flowers are still being initiated over summer and early autumn (e.g. [HydeEtAl1959]; [Thomas1987Chapter]; [MedeirosAndSteiner2000]). Also due to their similarities and the fact that a population is described in the model (instead of single plants), peduncles and petioles are simulated as a single organ.  \r\nFlower emergence typically starts in spring, peaks at the beginning of summer and continues until autumn ([ChapterThomas1980]; [MacfarlaneAndSheath1984]; [MarshalEtAl1993_Production]). The actual periods vary for different cultivars, and some have flowers nearly all year (albeit with low frequency over winter).  A drop in flower numbers in mid-summer and subsequent increase in early autumn has been shown in some cultivars and has been linked to water deficit or inhibition due to long days ([ChapterThomas1980]; [MacfarlaneAndSheath1984]).  Management practices, such as applying N fertiliser, irrigation and defoliations, have, in general, a positive impact on flower and seed production ([ChapterThomas1980]; [MedeirosAndSteiner2000]; [BissuelBelaygueEtAl2002_flower]), but there is large variation in the responses, depending on the cultivars as well as the extent and timing of the intervention. Currently, the model does not account for the effect of this practices explicitly, although it is expected that some effects are captured by changes in biomass allocation caused by actions such as irrigation of application of fertilisers.\r\n",
                      "Name": "Description",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": "days",
                      "Name": "Target",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Phenology].ThermalTime",
                      "Name": "ThermalTime",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "Progression",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            9.0,
                            11.5,
                            14.0,
                            15.0
                          ],
                          "Y": [
                            1.0,
                            1.0,
                            0.0,
                            0.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Phenology].Photoperiod",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Phen.GotoPhase, Models",
                  "Start": "Ripening",
                  "PhaseNameToGoto": "Vegetative",
                  "Name": "PhenologyRewind",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhotoperiodDeltaFunction, Models",
                  "Twilight": -6.0,
                  "Name": "PhotoperiodVariation",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.PhotoperiodFunction, Models",
                  "Twilight": -6.0,
                  "DayLength": 0.0,
                  "Name": "Photoperiod",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.Organs.PerennialLeaf, Models",
              "DMSupply": None,
              "NSupply": None,
              "DMDemand": None,
              "NDemand": None,
              "MaintenanceRespiration": 0.0,
              "GrowthRespiration": 0.0,
              "Albedo": 0.26,
              "Gsmax350": 0.006,
              "R50": 200.0,
              "LAI": 0.0,
              "Height": 0.0,
              "FRGR": 0.0,
              "PotentialEP": 0.0,
              "WaterDemand": 0.0,
              "LightProfile": None,
              "potentialDMAllocation": None,
              "Name": "Leaf",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Memo, Models",
                  "Text": "",
                  "Name": "Description",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.MultiplyFunction, Models",
                  "Name": "InitialWtFunction",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.8,
                      "Units": "g/plant",
                      "Name": "InitialPlantWt",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Plant].Population",
                      "Name": "Population",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.65,
                  "Units": None,
                  "Name": "ExtinctionCoefficient",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "For the Brazilian Brachiaria data (Brachiaria brizantha cv. BRS Piatã), the extinction coefficient was 0.65 [Bosi_et_al_2020a].",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.4,
                  "Units": None,
                  "Name": "ExtinctionCoefficientDead",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "Value of 0.57 from \r\nRadiation Interception and Growth of Planted and Coppice Stands of Four Fast-GrowingTropical Trees\r\nAuthor(s): Robin A. Harrington and James H. Fownes\r\nSource: Journal of Applied Ecology, Vol. 32, No. 1 (Feb., 1995), pp. 1-8\r\n\r\nBut their data suggests a value of 0.3",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                  "Name": "HeightFunction",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.XYPairs, Models",
                      "X": [
                        0.0,
                        100.0,
                        1000.0
                      ],
                      "Y": [
                        100.0,
                        300.0,
                        1400.0
                      ],
                      "XVariableName": None,
                      "Name": "XYPairs",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[TotalLive].Wt",
                      "Name": "XValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.021,
                  "Units": None,
                  "Name": "MaximumNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.005,
                  "Units": None,
                  "Name": "MinimumNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Library.BiomassRemoval, Models",
                  "Name": "BiomassRemovalDefaults",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.0,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Harvest",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.0,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Cut",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.0,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Prune",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.0,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Graze",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.MultiplyFunction, Models",
                  "Name": "LeafResidenceTime",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 80.0,
                      "Units": "days",
                      "Name": "Potential",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "LAIFactor",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            5.0,
                            6.0,
                            7.0,
                            8.0,
                            9.0
                          ],
                          "Y": [
                            1.0,
                            1.0,
                            0.8,
                            0.7,
                            0.6,
                            0.2
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].LAI",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Memo, Models",
                          "Text": "This parameter calculates the effect of self-shading on senescence of leaves.",
                          "Name": "Memo",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "FloweringFactor",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            5.0,
                            6.0,
                            7.0
                          ],
                          "Y": [
                            1.0,
                            1.0,
                            0.7,
                            0.7
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[TropicalPasture].Phenology.Stage",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Memo, Models",
                          "Text": "This parameter calculates the effect of phenology on senescence of leaves.",
                          "Name": "Memo",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "WaterDeficitFactor",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            0.9,
                            1.0
                          ],
                          "Y": [
                            1.0,
                            1.0,
                            1.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].Fw",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Memo, Models",
                          "Text": "This parameter calculates the effect of water stress on senescence of leaves.",
                          "Name": "Memo",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                  "Name": "LeafDetachmentTime",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.XYPairs, Models",
                      "X": [
                        0.0,
                        1.0,
                        2.0,
                        3.0,
                        4.0,
                        5.0
                      ],
                      "Y": [
                        30.0,
                        30.0,
                        30.0,
                        30.0,
                        30.0,
                        30.0
                      ],
                      "XVariableName": None,
                      "Name": "XYPairs",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[TropicalPasture].Phenology.Stage",
                      "Name": "XValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "SoilMoistureFactor",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            0.5,
                            0.75,
                            1.0
                          ],
                          "Y": [
                            0.1,
                            0.1,
                            0.25,
                            1.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].Fw",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.5,
                  "Units": None,
                  "Name": "NReallocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "DMRetranslocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "NRetranslocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.SupplyFunctions.RUEModel, Models",
                  "Name": "Photosynthesis",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.44,
                      "Units": None,
                      "Name": "RUE",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Memo, Models",
                          "Text": "[Bosi_et_al_2020a] found a RUE of 2.31 g/MJ PAR for Brachiaria brizantha cv. BRS Piata (C4 grass) in irrigated experiment, applying a correction factor for limiting temperature and not considering data in flowering period (it changes the partitioning). Converting this to total radiation was obtained 1.16 g/MJ and adding 20% to account for root production gives 1.44 g/MJ.\r\n",
                          "Name": "Memo",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.WeightedTemperatureFunction, Models",
                      "MaximumTemperatureWeighting": 0.75,
                      "Name": "FT",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            30.0,
                            40.0,
                            43.0
                          ],
                          "Y": [
                            0.0,
                            1.0,
                            1.0,
                            0.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Memo, Models",
                          "Text": "[Bosi_et_al_2020b] found a base temperature value of 8.9oC for Brachiaria brizantha cv. BRS Piata (C4 grass). Values of the two optimum temperatures and of maximum temperature were obtained from [Pequeno_et_al_2014].",
                          "Name": "Memo",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "FN",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].Fn",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            1.0,
                            1.5
                          ],
                          "Y": [
                            0.0,
                            1.0,
                            1.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.SupplyFunctions.RUECO2Function, Models",
                      "PhotosyntheticPathway": "C4",
                      "Name": "FCO2",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "FVPD",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            10.0,
                            12.5,
                            40.0
                          ],
                          "Y": [
                            1.0,
                            1.0,
                            0.95,
                            0.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].Photosynthesis.VPD",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "FW",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].Fw",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            0.3
                          ],
                          "Y": [
                            0.0,
                            1.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Leaf].RadiationIntercepted",
                      "Name": "RadnInt",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.MinimumFunction, Models",
                  "Name": "FRGRFunction",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Leaf].Photosynthesis.FT",
                      "Name": "FT",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Leaf].Photosynthesis.FN",
                      "Name": "FN",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Leaf].Photosynthesis.FVPD",
                      "Name": "FVPD",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "PotentialPhotosynthesis",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].Photosynthesis.RUE",
                          "Name": "RUE",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.MaximumFunction, Models",
                          "Name": "Radiation",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Leaf].RadiationIntercepted",
                              "Name": "RadnInt",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 1E-05,
                              "Units": None,
                              "Name": "MinRadn",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.MultiplyFunction, Models",
                  "Name": "SpecificLeafAreaFunction",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "StandardSLA",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            0.3,
                            0.7,
                            2.0
                          ],
                          "Y": [
                            0.015,
                            0.018,
                            0.02,
                            0.02
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[TropicalPasture].Age",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Memo, Models",
                          "Text": "StandardSLA is the specific leaf area of plants that were not submitted to any stress. The values used here were obtained from [Bosi_et_al_2020b].",
                          "Name": "Memo",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "WaterFactor",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            0.8,
                            1.0
                          ],
                          "Y": [
                            0.0,
                            0.4,
                            1.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].Fw",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Memo, Models",
                          "Text": "This factor calculates the effect of water deficit on specific leaf area. The values used here were calibrated by [Bosi_et_al_2020a].",
                          "Name": "Memo",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.03,
                  "Units": "m2/m2",
                  "Name": "MinimumLAI",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": None,
                  "Name": "DMConversionEfficiency",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.4,
                  "Units": None,
                  "Name": "CarbonConcentration",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": None,
                  "Name": "LeafDevelopmentRate",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": None,
                  "Name": "StomatalConductanceCO2Modifier",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientPoolFunctions, Models",
                  "Name": "DMDemands",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.DemandFunctions.PartitionFractionDemandFunction, Models",
                          "Name": "DMDemandFunction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.LinearInterpolationFunction, Models",
                              "Name": "PartitionFraction",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.XYPairs, Models",
                                  "X": [
                                    0.0,
                                    1.0,
                                    2.0,
                                    3.0,
                                    4.0,
                                    5.0,
                                    6.0,
                                    7.0
                                  ],
                                  "Y": [
                                    0.5,
                                    0.5,
                                    0.55,
                                    0.55,
                                    0.55,
                                    0.55,
                                    0.55,
                                    0.55
                                  ],
                                  "XVariableName": None,
                                  "Name": "XYPairs",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[TropicalPasture].Phenology.Stage",
                                  "Name": "XValue",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Memo, Models",
                                  "Text": "PartitionFraction is driven by phenology. The values used here were calibrated by [Bosi_et_al_2020a] for Brachiaria brizantha cv. BRS Piatã.",
                                  "Name": "Memo",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 1.0,
                          "Units": None,
                          "Name": "StructuralFraction",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.DemandFunctions.StorageDMDemandFunction, Models",
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "StorageFraction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 1.0,
                              "Units": None,
                              "Name": "One",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Leaf].DMDemands.Structural.StructuralFraction",
                              "Name": "StructuralFraction",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientPoolFunctions, Models",
                  "Name": "NDemands",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].minimumNconc",
                          "Name": "MinNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].potentialDMAllocation.Structural",
                          "Name": "PotentialDMAllocation",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.DemandFunctions.StorageNDemandFunction, Models",
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].MaximumNconc",
                          "Name": "MaxNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 1.0,
                          "Units": None,
                          "Name": "NitrogenDemandSwitch",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.MultiplyFunction, Models",
                  "Name": "SenescenceRate",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "The reference senescence rate is adjusted using generic relationships to environmental factors, such as temperature and soil moisture.  These are based on general knowledge from other plant models.",
                      "Name": "Description",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Leaf].SenescenceRate.ReferenceRate",
                      "Name": "ReferenceRate",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "StageFactor",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            3.0,
                            4.0,
                            5.5,
                            5.8,
                            6.0
                          ],
                          "Y": [
                            0.5,
                            1.0,
                            1.0,
                            1.5,
                            1.5
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Phenology].Stage",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "CoverFactor",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            2.0,
                            4.0,
                            5.0,
                            7.0
                          ],
                          "Y": [
                            0.0,
                            0.1,
                            0.5,
                            0.5,
                            1.5
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].LAI",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.WeightedTemperatureFunction, Models",
                      "MaximumTemperatureWeighting": 0.5,
                      "Name": "TemperatureFactor",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            -20.0,
                            -5.0,
                            0.0,
                            5.0,
                            17.0,
                            28.0,
                            45.0
                          ],
                          "Y": [
                            1.5,
                            0.25,
                            0.1,
                            0.1,
                            1.0,
                            1.0,
                            1.5
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "SoilMoistureFactor",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            0.5,
                            1.0
                          ],
                          "Y": [
                            2.0,
                            1.0,
                            1.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].Fw",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                  "Name": "LeafKillFraction",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.XYPairs, Models",
                      "X": [
                        0.0,
                        2.0
                      ],
                      "Y": [
                        0.14,
                        0.0
                      ],
                      "XVariableName": None,
                      "Name": "XYPairs",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Weather].MinT",
                      "Name": "XValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "This function calculates the fraction of leaf live mass that is killed by frost. That fraction is calculated based on minimum air temperature.",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.VariableReference, Models",
                  "VariableName": "[Phenology].ThermalTime",
                  "Name": "ThermalTime",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.Organs.GenericOrgan, Models",
              "StartLive": None,
              "DMSupply": None,
              "NSupply": None,
              "DMDemand": None,
              "NDemand": None,
              "potentialDMAllocation": None,
              "IsAboveGround": True,
              "Name": "Stem",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Memo, Models",
                  "Text": "The stem organ is used to describe the growth of stem plus inflorescence.",
                  "Name": "Description",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientPoolFunctions, Models",
                  "Name": "InitialWt",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": "g/m^2",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.003,
                  "Units": None,
                  "Name": "MaximumNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.001,
                  "Units": None,
                  "Name": "MinimumNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.VariableReference, Models",
                  "VariableName": "[Stem].MinimumNConc",
                  "Name": "CriticalNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": None,
                  "Name": "NitrogenDemandSwitch",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Library.BiomassRemoval, Models",
                  "Name": "BiomassRemovalDefaults",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.0,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Harvest",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.0,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Cut",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.0,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Prune",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.0,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Graze",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.0,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Thin",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                  "Name": "SenescenceRate",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.XYPairs, Models",
                      "X": [
                        0.0,
                        1.0,
                        2.0,
                        3.0,
                        4.0,
                        5.0,
                        6.0,
                        7.0
                      ],
                      "Y": [
                        0.005,
                        0.005,
                        0.005,
                        0.005,
                        0.005,
                        0.005,
                        0.015,
                        0.015
                      ],
                      "XVariableName": None,
                      "Name": "XYPairs",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[TropicalPasture].Phenology.Stage",
                      "Name": "XValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "Stem senescence rate is driven by phenology. The values used here were calibrated by [Bosi_et_al_2020a] for Brachiaria brizantha cv. BRS Piatã.",
                      "Name": "Memo",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                  "Name": "DetachmentRateFunction",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.XYPairs, Models",
                      "X": [
                        0.0,
                        1.0,
                        2.0,
                        3.0,
                        4.0,
                        5.0,
                        6.0,
                        7.0
                      ],
                      "Y": [
                        0.01,
                        0.01,
                        0.01,
                        0.01,
                        0.01,
                        0.01,
                        0.01,
                        0.01
                      ],
                      "XVariableName": None,
                      "Name": "XYPairs",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[TropicalPasture].Phenology.Stage",
                      "Name": "XValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "DMReallocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "NReallocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "DMRetranslocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "NRetranslocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": "0-1",
                  "Name": "DMConversionEfficiency",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": "0-1",
                  "Name": "MaintenanceRespirationFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "RemobilisationCost",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.4,
                  "Units": None,
                  "Name": "CarbonConcentration",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientDemandFunctions, Models",
                  "Name": "DMDemands",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.DemandFunctions.PartitionFractionDemandFunction, Models",
                          "Name": "DMDemandFunction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.LinearInterpolationFunction, Models",
                              "Name": "PartitionFraction",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.XYPairs, Models",
                                  "X": [
                                    0.0,
                                    1.0,
                                    2.0,
                                    3.0,
                                    4.0,
                                    5.0,
                                    6.0,
                                    7.0
                                  ],
                                  "Y": [
                                    0.17,
                                    0.17,
                                    0.2,
                                    0.25,
                                    0.45,
                                    0.45,
                                    0.45,
                                    0.25
                                  ],
                                  "XVariableName": None,
                                  "Name": "XYPairs",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[TropicalPasture].Phenology.Stage",
                                  "Name": "XValue",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Memo, Models",
                                  "Text": "PartitionFraction is driven by phenology. The values used here were calibrated by [Bosi_et_al_2020a] for Brachiaria brizantha cv. BRS Piatã.",
                                  "Name": "Memo",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 1.0,
                          "Units": "g/g",
                          "Name": "StructuralFraction",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.DemandFunctions.StorageDMDemandFunction, Models",
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "StorageFraction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 1.0,
                              "Units": None,
                              "Name": "One",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Stem].DMDemands.Structural.StructuralFraction",
                              "Name": "StructuralFraction",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStructuralPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QMetabolicPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStoragePriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientDemandFunctions, Models",
                  "Name": "NDemands",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Stem].minimumNconc",
                          "Name": "MinNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Stem].potentialDMAllocation.Structural",
                          "Name": "PotentialDMAllocation",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "MetabolicNconc",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Stem].criticalNConc",
                              "Name": "CritNconc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Stem].minimumNconc",
                              "Name": "MinNconc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Stem].potentialDMAllocation.Structural",
                          "Name": "PotentialDMAllocation",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.DemandFunctions.StorageNDemandFunction, Models",
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Stem].nitrogenDemandSwitch",
                          "Name": "NitrogenDemandSwitch",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Stem].maximumNconc",
                          "Name": "MaxNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStructuralPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QMetabolicPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStoragePriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.RetranslocateNonStructural, Models",
                  "Name": "RetranslocateNitrogen",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "Photosynthesis",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.VariableReference, Models",
                  "VariableName": "[Stem].MinimumNConc",
                  "Name": "initialNConcFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.Organs.Root, Models",
              "DMSupply": None,
              "NSupply": None,
              "DMDemand": None,
              "NDemand": None,
              "potentialDMAllocation": None,
              "GrowthRespiration": 0.0,
              "MaintenanceRespiration": 0.0,
              "Name": "Root",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Functions.RootShape.RootShapeCylinder, Models",
                  "Name": "RootShape",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientPoolFunctions, Models",
                  "Name": "InitialWt",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.02,
                      "Units": "g/plant",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 20.0,
                  "Units": None,
                  "Name": "RootFrontVelocity",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1000000.0,
                  "Units": None,
                  "Name": "MaximumRootDepth",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 100.0,
                  "Units": "m/g",
                  "Name": "SpecificRootLength",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.003,
                  "Units": None,
                  "Name": "MaximumNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.001,
                  "Units": None,
                  "Name": "MinimumNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.VariableReference, Models",
                  "VariableName": "[Root].MinimumNConc",
                  "Name": "CriticalNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": None,
                  "Name": "NitrogenDemandSwitch",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 8.0,
                  "Units": None,
                  "Name": "MaxDailyNUptake",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                  "Name": "KNO3",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.XYPairs, Models",
                      "X": [
                        0.0,
                        0.003
                      ],
                      "Y": [
                        0.03,
                        0.03
                      ],
                      "XVariableName": None,
                      "Name": "XYPairs",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Root].LengthDensity",
                      "Name": "XValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                  "Name": "KNH4",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.XYPairs, Models",
                      "X": [
                        0.0,
                        0.003
                      ],
                      "Y": [
                        0.0,
                        0.0
                      ],
                      "XVariableName": None,
                      "Name": "XYPairs",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Root].LengthDensity",
                      "Name": "XValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                  "Name": "NUptakeSWFactor",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.XYPairs, Models",
                      "X": [
                        0.0,
                        0.5,
                        1.0
                      ],
                      "Y": [
                        1.0,
                        1.0,
                        1.0
                      ],
                      "XVariableName": None,
                      "Name": "XYPairs",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Root].RWC",
                      "Name": "XValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.LinearInterpolationFunction, Models",
                  "Name": "KLModifier",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.XYPairs, Models",
                      "X": [
                        0.0,
                        3.0,
                        10.0
                      ],
                      "Y": [
                        1.0,
                        1.0,
                        1.0
                      ],
                      "XVariableName": None,
                      "Name": "XYPairs",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.VariableReference, Models",
                      "VariableName": "[Leaf].LAI",
                      "Name": "XValue",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Library.BiomassRemoval, Models",
                  "Name": "BiomassRemovalDefaults",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.0,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Harvest",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.0,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Cut",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.0,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Prune",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.0,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Graze",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.0,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Thin",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.001,
                  "Units": None,
                  "Name": "SenescenceRate",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "DMRetranslocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "NRetranslocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": None,
                  "Name": "DMConversionEfficiency",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": None,
                  "Name": "MaintenanceRespirationFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "RemobilisationCost",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.4,
                  "Units": None,
                  "Name": "CarbonConcentration",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientDemandFunctions, Models",
                  "Name": "DMDemands",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.DemandFunctions.PartitionFractionDemandFunction, Models",
                          "Name": "DMDemandFunction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.LinearInterpolationFunction, Models",
                              "Name": "PartitionFraction",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.XYPairs, Models",
                                  "X": [
                                    0.0,
                                    1.0,
                                    2.0,
                                    3.0,
                                    4.0,
                                    5.0,
                                    6.0,
                                    7.0
                                  ],
                                  "Y": [
                                    0.25,
                                    0.25,
                                    0.19,
                                    0.15,
                                    0.0,
                                    0.0,
                                    0.0,
                                    0.15
                                  ],
                                  "XVariableName": None,
                                  "Name": "XYPairs",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[TropicalPasture].Phenology.Stage",
                                  "Name": "XValue",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Memo, Models",
                                  "Text": "PartitionFraction is driven by phenology. The values used here were calibrated by [Bosi_et_al_2020a] for Brachiaria brizantha cv. BRS Piatã. Partitioning to root is considering 75% of the partitioning for belowground growth.",
                                  "Name": "Memo",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 1.0,
                          "Units": "g/g",
                          "Name": "StructuralFraction",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.DemandFunctions.StorageDMDemandFunction, Models",
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "StorageFraction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 1.0,
                              "Units": None,
                              "Name": "One",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Root].DMDemands.Structural.StructuralFraction",
                              "Name": "StructuralFraction",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStructuralPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QMetabolicPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStoragePriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientDemandFunctions, Models",
                  "Name": "NDemands",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Root].minimumNconc",
                          "Name": "MinNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Root].potentialDMAllocation.Structural",
                          "Name": "PotentialDMAllocation",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "MetabolicNconc",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Root].criticalNConc",
                              "Name": "CritNconc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[Root].minimumNconc",
                              "Name": "MinNconc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Root].potentialDMAllocation.Structural",
                          "Name": "PotentialDMAllocation",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.DemandFunctions.StorageNDemandFunction, Models",
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Root].nitrogenDemandSwitch",
                          "Name": "NitrogenDemandSwitch",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Root].maximumNconc",
                          "Name": "MaxNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStructuralPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QMetabolicPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStoragePriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": None,
                  "Name": "RootDepthStressFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.Organs.GenericOrgan, Models",
              "StartLive": None,
              "DMSupply": None,
              "NSupply": None,
              "DMDemand": None,
              "NDemand": None,
              "potentialDMAllocation": None,
              "IsAboveGround": False,
              "Name": "StorageOrgan",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Memo, Models",
                  "Text": "StorageOrgan is used to describe the crown of bunch grasses and it may also be used to simulate stolon or rhizome for species that have these. Currently, this organ is considered as a belowground organ. Partitioning to StorageOrgan is considered 25% of the partitioning for belowground growth.",
                  "Name": "Description",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientPoolFunctions, Models",
                  "Name": "InitialWt",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": "g/m^2",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.003,
                  "Units": None,
                  "Name": "MaximumNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.001,
                  "Units": None,
                  "Name": "MinimumNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.VariableReference, Models",
                  "VariableName": "[StorageOrgan].MinimumNConc",
                  "Name": "CriticalNConc",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": None,
                  "Name": "NitrogenDemandSwitch",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Library.BiomassRemoval, Models",
                  "Name": "BiomassRemovalDefaults",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.0,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Harvest",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.0,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Cut",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.0,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Prune",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.0,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Graze",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.PMF.OrganBiomassRemovalType, Models",
                      "FractionLiveToRemove": 0.0,
                      "FractionDeadToRemove": 0.0,
                      "FractionLiveToResidue": 0.0,
                      "FractionDeadToResidue": 0.0,
                      "Name": "Thin",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": "/d",
                  "Name": "DetachmentRateFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "DMReallocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": None,
                  "Name": "NReallocationFactor",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.MultiplyFunction, Models",
                  "Name": "DMRetranslocationFactor",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.499,
                      "Units": None,
                      "Name": "Potential",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "FloweringFactor",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            1.0,
                            2.0,
                            3.0,
                            4.0,
                            5.0,
                            6.0,
                            7.0
                          ],
                          "Y": [
                            1.0,
                            1.0,
                            1.0,
                            1.0,
                            1.0,
                            1.0,
                            1.0,
                            1.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[TropicalPasture].Phenology.Stage",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Memo, Models",
                          "Text": "This parameter calculates the effect of phenology on DM retranslocation from the storage organ to other organs.",
                          "Name": "Memo",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "LAIFactor",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            0.01,
                            0.77,
                            1.2,
                            1.21,
                            7.0,
                            8.0,
                            9.0
                          ],
                          "Y": [
                            2.0,
                            0.5,
                            0.0,
                            0.0,
                            0.0,
                            0.0,
                            0.0,
                            0.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[TropicalPasture].Leaf.LAI",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Memo, Models",
                          "Text": "This parameter calculates the effect of LAI on DM retranslocation from the storage organ to other organs.",
                          "Name": "Memo",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.MultiplyFunction, Models",
                  "Name": "NRetranslocationFactor",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Potential",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "FnFactor",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            0.5,
                            1.0
                          ],
                          "Y": [
                            1.0,
                            0.0,
                            0.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].Fn",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Memo, Models",
                          "Text": "This parameter calculates the effect of N stress on N retranslocation from the storage organ to other organs.",
                          "Name": "Memo",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "LAIFactor",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            0.5,
                            7.0,
                            8.0,
                            9.0
                          ],
                          "Y": [
                            1.0,
                            1.0,
                            0.0,
                            0.0,
                            0.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].LAI",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Memo, Models",
                          "Text": "This parameter calculates the effect of LAI on N retranslocation from the storage organ to other organs.",
                          "Name": "Memo",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 1.0,
                  "Units": "0-1",
                  "Name": "DMConversionEfficiency",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": "0-1",
                  "Name": "MaintenanceRespirationFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "RemobilisationCost",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.4,
                  "Units": None,
                  "Name": "CarbonConcentration",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientDemandFunctions, Models",
                  "Name": "DMDemands",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.DemandFunctions.PartitionFractionDemandFunction, Models",
                          "Name": "DMDemandFunction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.LinearInterpolationFunction, Models",
                              "Name": "PartitionFraction",
                              "ResourceName": None,
                              "Children": [
                                {
                                  "$child": "Models.Functions.XYPairs, Models",
                                  "X": [
                                    0.0,
                                    1.0,
                                    2.0,
                                    3.0,
                                    4.0,
                                    5.0,
                                    6.0,
                                    7.0
                                  ],
                                  "Y": [
                                    0.08,
                                    0.08,
                                    0.06,
                                    0.05,
                                    0.0,
                                    0.0,
                                    0.0,
                                    0.05
                                  ],
                                  "XVariableName": None,
                                  "Name": "XYPairs",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Functions.VariableReference, Models",
                                  "VariableName": "[TropicalPasture].Phenology.Stage",
                                  "Name": "XValue",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                },
                                {
                                  "$child": "Models.Memo, Models",
                                  "Text": "PartitionFraction is driven by phenology. The values used here were calibrated by [Bosi_et_al_2020a] for Brachiaria brizantha cv. BRS Piatã.",
                                  "Name": "Memo",
                                  "ResourceName": None,
                                  "Children": [],
                                  "Enabled": True,
                                  "ReadOnly": False
                                }
                              ],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.Constant, Models",
                          "FixedValue": 0.5,
                          "Units": "g/g",
                          "Name": "StructuralFraction",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.0,
                      "Units": None,
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.DemandFunctions.StorageDMDemandFunction, Models",
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "StorageFraction",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.Constant, Models",
                              "FixedValue": 1.0,
                              "Units": None,
                              "Name": "One",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[StorageOrgan].DMDemands.Structural.StructuralFraction",
                              "Name": "StructuralFraction",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStructuralPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QMetabolicPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStoragePriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.NutrientDemandFunctions, Models",
                  "Name": "NDemands",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Structural",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[StorageOrgan].minimumNconc",
                          "Name": "MinNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[StorageOrgan].potentialDMAllocation.Structural",
                          "Name": "PotentialDMAllocation",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.MultiplyFunction, Models",
                      "Name": "Metabolic",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.SubtractFunction, Models",
                          "Name": "MetabolicNconc",
                          "ResourceName": None,
                          "Children": [
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[StorageOrgan].criticalNConc",
                              "Name": "CritNconc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            },
                            {
                              "$child": "Models.Functions.VariableReference, Models",
                              "VariableName": "[StorageOrgan].minimumNconc",
                              "Name": "MinNconc",
                              "ResourceName": None,
                              "Children": [],
                              "Enabled": True,
                              "ReadOnly": False
                            }
                          ],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[StorageOrgan].potentialDMAllocation.Structural",
                          "Name": "PotentialDMAllocation",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.DemandFunctions.StorageNDemandFunction, Models",
                      "Name": "Storage",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[StorageOrgan].nitrogenDemandSwitch",
                          "Name": "NitrogenDemandSwitch",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[StorageOrgan].maximumNconc",
                          "Name": "MaxNconc",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStructuralPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QMetabolicPriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 1.0,
                      "Units": None,
                      "Name": "QStoragePriority",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.RetranslocateNonStructural, Models",
                  "Name": "RetranslocateNitrogen",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.Constant, Models",
                  "FixedValue": 0.0,
                  "Units": None,
                  "Name": "Photosynthesis",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.VariableReference, Models",
                  "VariableName": "[StorageOrgan].MinimumNConc",
                  "Name": "initialNConcFunction",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Functions.MultiplyFunction, Models",
                  "Name": "SenescenceRate",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Memo, Models",
                      "Text": "The reference senescence rate is adjusted using generic relationships to environmental factors, such as temperature and soil moisture.  These are based on general knowledge from other plant models.",
                      "Name": "Description",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.Constant, Models",
                      "FixedValue": 0.01,
                      "Units": "/day",
                      "Name": "ReferenceRate",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "StageFactor",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            3.0,
                            4.0,
                            6.0
                          ],
                          "Y": [
                            0.5,
                            1.0,
                            1.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Phenology].Stage",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.WeightedTemperatureFunction, Models",
                      "MaximumTemperatureWeighting": 0.5,
                      "Name": "TemperatureFactor",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            -20.0,
                            -5.0,
                            0.0,
                            5.0,
                            17.0,
                            32.2,
                            45.0
                          ],
                          "Y": [
                            1.5,
                            0.25,
                            0.1,
                            0.1,
                            1.0,
                            1.0,
                            1.25
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Functions.LinearInterpolationFunction, Models",
                      "Name": "SoilMoistureFactor",
                      "ResourceName": None,
                      "Children": [
                        {
                          "$child": "Models.Functions.XYPairs, Models",
                          "X": [
                            0.0,
                            0.5,
                            1.0
                          ],
                          "Y": [
                            2.0,
                            1.0,
                            1.0
                          ],
                          "XVariableName": None,
                          "Name": "XYPairs",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        },
                        {
                          "$child": "Models.Functions.VariableReference, Models",
                          "VariableName": "[Leaf].Fw",
                          "Name": "XValue",
                          "ResourceName": None,
                          "Children": [],
                          "Enabled": True,
                          "ReadOnly": False
                        }
                      ],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.CompositeBiomass, Models",
              "OrganNames": [
                "Leaf",
                "Stem"
              ],
              "IncludeLive": True,
              "IncludeDead": True,
              "Name": "AboveGround",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.CompositeBiomass, Models",
              "OrganNames": [
                "Leaf",
                "Stem"
              ],
              "IncludeLive": True,
              "IncludeDead": False,
              "Name": "AboveGroundLive",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.CompositeBiomass, Models",
              "OrganNames": [
                "Leaf",
                "Stem"
              ],
              "IncludeLive": False,
              "IncludeDead": True,
              "Name": "AboveGroundDead",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.CompositeBiomass, Models",
              "OrganNames": [
                "Root",
                "StorageOrgan"
              ],
              "IncludeLive": True,
              "IncludeDead": True,
              "Name": "BelowGround",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.CompositeBiomass, Models",
              "OrganNames": [
                "Leaf",
                "Stem",
                "Root",
                "StorageOrgan"
              ],
              "IncludeLive": True,
              "IncludeDead": True,
              "Name": "Total",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.CompositeBiomass, Models",
              "OrganNames": [
                "Leaf",
                "Stem",
                "Root",
                "StorageOrgan"
              ],
              "IncludeLive": True,
              "IncludeDead": False,
              "Name": "TotalLive",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.CompositeBiomass, Models",
              "OrganNames": [
                "Leaf",
                "Stem",
                "Root",
                "StorageOrgan"
              ],
              "IncludeLive": False,
              "IncludeDead": True,
              "Name": "TotalDead",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.CompositeBiomass, Models",
              "OrganNames": [
                "Leaf",
                "Stem"
              ],
              "IncludeLive": True,
              "IncludeDead": True,
              "Name": "Forage",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.Functions.Constant, Models",
              "FixedValue": 0.0,
              "Units": None,
              "Name": "MortalityRate",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.Functions.Constant, Models",
              "FixedValue": 0.0,
              "Units": None,
              "Name": "SeedMortalityRate",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.Cultivar, Models",
              "Command": [
                "[Leaf].Photosynthesis.RUE.FixedValue =2.45",
                "[Leaf].ExtinctionCoefficient.FixedValue = 0.65",
                "[Root].MaximumRootDepth.FixedValue = 3000",
                "//[Root].SpecificRootLength.FixedValue = 100",
                "",
                "/// frost kill ratio",
                "[Leaf].LeafKillFraction.XYPairs.X = -35,  -2, //0,  17, 28, 42, 45",
                "[Leaf].LeafKillFraction.XYPairs.Y = 0.014, 0.0, //0, 0, 0, 0, 0.001",
                "",
                "///Temperature factors",
                "[Leaf].Photosynthesis.FT.XYPairs.X =     0, 2.5,  15,   32, 50",
                "[Leaf].Photosynthesis.FT.XYPairs.Y =      0,  0.0,  1,  1, 0",
                "[Leaf].Photosynthesis.FT.MaximumTemperatureWeighting = 0.75",
                "[Leaf].MinimumLAI.FixedValue = 0.1",
                "[Leaf].InitialWtFunction.InitialPlantWt.FixedValue = 0.8",
                "[Leaf].MaximumNConc.FixedValue = 0.023",
                "[Leaf].MinimumNConc.FixedValue = 0.0045"
              ],
              "Name": "SwitchGrass8",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.Cultivar, Models",
              "Command": [
                "[Leaf].Photosynthesis.RUE.FixedValue =2.45",
                "[Leaf].ExtinctionCoefficient.FixedValue = 0.5",
                "[Root].MaximumRootDepth.FixedValue = 3000",
                "//[Root].SpecificRootLength.FixedValue = 100",
                "",
                "/// frost kill ratio",
                "[Leaf].LeafKillFraction.XYPairs.X = -35,  -5, ",
                "[Leaf].LeafKillFraction.XYPairs.Y = 0.014, 0.0,",
                "",
                "///Temperature factors",
                "[Leaf].Photosynthesis.FT.XYPairs.X =     0, 2.5,  15,   32, 43",
                "[Leaf].Photosynthesis.FT.XYPairs.Y =      0,  0.0,  1,  1, 0",
                "[Leaf].Photosynthesis.FT.MaximumTemperatureWeighting = 0.75",
                "[Leaf].MinimumLAI.FixedValue = 0.1",
                "[Leaf].InitialWtFunction.InitialPlantWt.FixedValue = 0.8",
                "[Leaf].MaximumNConc.FixedValue = 0.023",
                "[Leaf].MinimumNConc.FixedValue = 0.0045"
              ],
              "Name": "SwitchGrass72",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.Core.Folder, Models",
              "ShowInDocs": True,
              "GraphsPerPage": 6,
              "Name": "Cultivar",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Leaf].Photosynthesis.RUE.FixedValue =4.75",
                    "[Leaf].ExtinctionCoefficient.FixedValue = 0.65",
                    "[Root].MaximumRootDepth.FixedValue = 3000",
                    "//[Root].SpecificRootLength.FixedValue = 100",
                    "",
                    "/// frost kill ratio",
                    "[Leaf].LeafKillFraction.XYPairs.X = -25,  -2, ",
                    "[Leaf].LeafKillFraction.XYPairs.Y = 0.014, 0.0,",
                    "",
                    "///Temperature factors",
                    "[Leaf].Photosynthesis.FT.XYPairs.X =     0, 6.5,  15,   32, 50",
                    "[Leaf].Photosynthesis.FT.XYPairs.Y =      0,  0.0,  1,  1, 0",
                    "",
                    "//[Leaf].Photosynthesis.FT.XYPairs.X =     0,  6, 9, 12, 17, 21, 43",
                    "//[Leaf].Photosynthesis.FT.XYPairs.Y =     0,  0, 0.4, 1,  1, 1, 0",
                    "",
                    "",
                    "[Leaf].Photosynthesis.FT.MaximumTemperatureWeighting = 0.95",
                    "[Leaf].MinimumLAI.FixedValue = 0.1",
                    "[Leaf].InitialWtFunction.InitialPlantWt.FixedValue = 0.8",
                    "[Leaf].MaximumNConc.FixedValue = 0.023",
                    "//[Leaf].MinimumNConc.FixedValue = 0.0045",
                    "[Leaf].MinimumNConc.FixedValue = 0.0045",
                    "",
                    "[Root].NUptakeSWFactor.XYPairs.X = 0, 1",
                    "[Root].NUptakeSWFactor.XYPairs.Y = 0, 1",
                    "[Phenology].Vegetative.Target.FixedValue =30"
                  ],
                  "Name": "SwitchGrass9",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Leaf].Photosynthesis.RUE.FixedValue =3.45",
                    "[Leaf].ExtinctionCoefficient.FixedValue = 0.65",
                    "[Root].MaximumRootDepth.FixedValue = 3000",
                    "//[Root].SpecificRootLength.FixedValue = 100",
                    "",
                    "/// frost kill ratio",
                    "[Leaf].LeafKillFraction.XYPairs.X = -35,  -2, //0,  17, 28, 42, 45",
                    "[Leaf].LeafKillFraction.XYPairs.Y = 0.014, 0.0, //0, 0, 0, 0, 0.001",
                    "",
                    "///Temperature factors",
                    "[Leaf].Photosynthesis.FT.XYPairs.X =     0, 2.5,  15,   32, 50",
                    "[Leaf].Photosynthesis.FT.XYPairs.Y =      0,  0.0,  1,  1, 0",
                    "[Leaf].Photosynthesis.FT.MaximumTemperatureWeighting = 0.75",
                    "[Leaf].MinimumLAI.FixedValue = 0.1",
                    "[Leaf].InitialWtFunction.InitialPlantWt.FixedValue = 0.8",
                    "[Leaf].MaximumNConc.FixedValue = 0.023",
                    "[Leaf].MinimumNConc.FixedValue = 0.0045"
                  ],
                  "Name": "SwitchGrass8",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Leaf].Photosynthesis.RUE.FixedValue =2.45",
                    "[Leaf].ExtinctionCoefficient.FixedValue = 0.5",
                    "[Root].MaximumRootDepth.FixedValue = 3000",
                    "//[Root].SpecificRootLength.FixedValue = 100",
                    "",
                    "/// frost kill ratio",
                    "[Leaf].LeafKillFraction.XYPairs.X = -25,  -2, ",
                    "[Leaf].LeafKillFraction.XYPairs.Y = 0.014, 0.0,",
                    "",
                    "///Temperature factors",
                    "[Leaf].Photosynthesis.FT.XYPairs.X =     0, 6.5,  15,   32, 50",
                    "[Leaf].Photosynthesis.FT.XYPairs.Y =      0,  0.0,  1,  1, 0",
                    "[Leaf].Photosynthesis.FT.MaximumTemperatureWeighting = 0.5",
                    "[Leaf].MinimumLAI.FixedValue = 0.1",
                    "[Leaf].InitialWtFunction.InitialPlantWt.FixedValue = 0.8",
                    "[Leaf].MaximumNConc.FixedValue = 0.023",
                    "[Leaf].MinimumNConc.FixedValue = 0.0045"
                  ],
                  "Name": "SwitchGrass7",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Leaf].Photosynthesis.RUE.FixedValue =1.80",
                    "[Leaf].ExtinctionCoefficient.FixedValue = 0.5",
                    "[Root].MaximumRootDepth.FixedValue = 3000",
                    "//[Root].SpecificRootLength.FixedValue = 100",
                    "",
                    "/// frost kill ratio",
                    "[Leaf].LeafKillFraction.XYPairs.X = -35,  -2, //0,  17, 28, 42, 45",
                    "[Leaf].LeafKillFraction.XYPairs.Y = 0.014, 0.0, //0, 0, 0, 0, 0.001",
                    "",
                    "///Temperature factors",
                    "[Leaf].Photosynthesis.FT.XYPairs.X =      6.5,     32.5,  45",
                    "[Leaf].Photosynthesis.FT.XYPairs.Y =       0.2,    1,  0",
                    "[Leaf].Photosynthesis.FT.MaximumTemperatureWeighting = 0.5",
                    "[Leaf].MinimumLAI.FixedValue = 0.1",
                    "[Leaf].InitialWtFunction.InitialPlantWt.FixedValue = 0.8",
                    "[Leaf].MaximumNConc.FixedValue = 0.023",
                    "[Leaf].MinimumNConc.FixedValue = 0.0045"
                  ],
                  "Name": "SwitchGrass6",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Leaf].Photosynthesis.RUE.FixedValue =1.80",
                    "[Leaf].ExtinctionCoefficient.FixedValue = 0.5",
                    "[Root].MaximumRootDepth.FixedValue = 3000",
                    "//[Root].SpecificRootLength.FixedValue = 100",
                    "",
                    "/// frost kill ratio",
                    "[Leaf].LeafKillFraction.XYPairs.X = -35,  -2, //0,  17, 28, 42, 45",
                    "[Leaf].LeafKillFraction.XYPairs.Y = 0.014, 0.0, //0, 0, 0, 0, 0.001",
                    "",
                    "///Temperature factors",
                    "[Leaf].Photosynthesis.FT.XYPairs.X =     6.5,    32, 35, 45",
                    "[Leaf].Photosynthesis.FT.XYPairs.Y =       0.2,   1,  1, 0",
                    "[Leaf].Photosynthesis.FT.MaximumTemperatureWeighting = 0.5",
                    "[Leaf].MinimumLAI.FixedValue = 0.1",
                    "[Leaf].InitialWtFunction.InitialPlantWt.FixedValue = 0.8",
                    "[Leaf].MaximumNConc.FixedValue = 0.023",
                    "[Leaf].MinimumNConc.FixedValue = 0.0045"
                  ],
                  "Name": "SwitchGrass5",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Leaf].Photosynthesis.RUE.FixedValue =1.7",
                    "[Leaf].ExtinctionCoefficient.FixedValue = 0.5",
                    "[Root].MaximumRootDepth.FixedValue = 3000",
                    "//[Root].SpecificRootLength.FixedValue = 100",
                    "",
                    "/// frost kill ratio",
                    "[Leaf].LeafKillFraction.XYPairs.X = -35,  -5, //0,  17, 28, 42, 45",
                    "[Leaf].LeafKillFraction.XYPairs.Y = 0.014, 0.0, //0, 0, 0, 0, 0.001",
                    "",
                    "///Temperature factors",
                    "[Leaf].Photosynthesis.FT.XYPairs.X =     0, 6.5,    32, 32.5, 45",
                    "[Leaf].Photosynthesis.FT.XYPairs.Y =     0,  0.2,   1,  1, 0",
                    "[Leaf].Photosynthesis.FT.MaximumTemperatureWeighting = 0.5",
                    "[Leaf].MinimumLAI.FixedValue = 0.1",
                    "[Leaf].InitialWtFunction.InitialPlantWt.FixedValue = 0.01",
                    "[TropicalPasture].AboveGroundHarvestable = 0.8"
                  ],
                  "Name": "SwitchGrass4",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Leaf].Photosynthesis.RUE.FixedValue =4.3370",
                    "[Leaf].ExtinctionCoefficient.FixedValue = 0.465",
                    "[Root].MaximumRootDepth.FixedValue = 2000",
                    "//[Root].SpecificRootLength.FixedValue = 100",
                    "",
                    "/// frost kill ratio",
                    "[Leaf].LeafKillFraction.XYPairs.X = -35,  -5, //0,  17, 28, 42, 45",
                    "[Leaf].LeafKillFraction.XYPairs.Y = 0.014, 0.0, //0, 0, 0, 0, 0.001",
                    "",
                    "///Temperature factors",
                    "[Leaf].Photosynthesis.FT.XYPairs.X =    9, 12, 17, 21, 45",
                    "[Leaf].Photosynthesis.FT.XYPairs.Y =   0.3, 0.5, 1, 1, 0",
                    "[Leaf].Photosynthesis.FT.MaximumTemperatureWeighting = 0.5",
                    "[Leaf].MinimumLAI.FixedValue = 0.1",
                    "[Leaf].InitialWtFunction.InitialPlantWt.FixedValue = 0.2",
                    "//Nitrogen _variables",
                    "[Leaf].MaximumNConc.FixedValue = 0.023",
                    "[Leaf].MinimumNConc.FixedValue = 0.0045"
                  ],
                  "Name": "Cave_in_rock111",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    ""
                  ],
                  "Name": "BrachiariaBRSPiata",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    ""
                  ],
                  "Name": "PanicumBambatsi",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Leaf].Photosynthesis.RUE.FixedValue =1.70",
                    "[Leaf].ExtinctionCoefficient.FixedValue = 0.5",
                    "[Root].MaximumRootDepth.FixedValue = 3000",
                    "//[Root].SpecificRootLength.FixedValue = 100",
                    "",
                    "/// frost kill ratio",
                    "[Leaf].LeafKillFraction.XYPairs.X = -35,  -5, //0,  17, 28, 42, 45",
                    "[Leaf].LeafKillFraction.XYPairs.Y = 0.014, 0.0, //0, 0, 0, 0, 0.001",
                    "",
                    "///Temperature factors",
                    "[Leaf].Photosynthesis.FT.XYPairs.X =    9, 12, 17, 21, 45",
                    "[Leaf].Photosynthesis.FT.XYPairs.Y =  0.5, 0.5, 0.8, 1, 0",
                    "[Leaf].Photosynthesis.FT.MaximumTemperatureWeighting = 0.75",
                    "[Leaf].MinimumLAI.FixedValue = 0.1",
                    "[Leaf].InitialWtFunction.InitialPlantWt.FixedValue = 0.5"
                  ],
                  "Name": "SwitchGrass3",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Memo, Models",
                  "Text": "This is a *large-leaf* cultivar developed in New Zealand from crossings between local varieties and north-american ladino genotypes.  It succeeded Pitau and Kopu cultivars, being more persistent than these due to greater stolon density; it has also greater resistance to nematodes and clover root weevil.  As a *large-leaf* cultivar, Kopu2 has leaves that are 25-50% larger than Huia, with corresponding longer petioles; it has a semi-prostrate growth habit, enabling higher harvest index than *medium-leaf* cultivars.  Kopu2 has been shown to produce high yields, especially in early season, being more winter active that the previous cultivars; it is intended for lowland pastures under rotational cattle grazing.  \r\n",
                  "Name": "Description",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Leaf].Photosynthesis.RUE.FixedValue =1.65",
                    "[Leaf].ExtinctionCoefficient.FixedValue = 0.65",
                    "[Root].MaximumRootDepth.FixedValue = 3000",
                    "[Root].SpecificRootLength.FixedValue = 100",
                    "",
                    "/// frost kill ratio",
                    "[Leaf].LeafKillFraction.XYPairs.X = -35,  -5, ",
                    "[Leaf].LeafKillFraction.XYPairs.Y = 0.014, 0.0,",
                    "",
                    "///Temperature factors",
                    "//[Leaf].Photosynthesis.FT.XYPairs.X =     0, 6.5,  15,   32, 43",
                    "//[Leaf].Photosynthesis.FT.XYPairs.Y =      0,  0.0,  1,  1, 0",
                    "[Leaf].Photosynthesis.FT.MaximumTemperatureWeighting =0.75",
                    "[Leaf].MinimumLAI.FixedValue = 0.1",
                    "[Leaf].InitialWtFunction.InitialPlantWt.FixedValue = 0.01",
                    "//[Leaf].MaximumNConc.FixedValue = 0.023",
                    "//[Leaf].MinimumNConc.FixedValue = 0.0045",
                    "[Leaf].MaximumNConc.FixedValue = 0.04",
                    "[Leaf].MinimumNConc.FixedValue = 0.007",
                    "//[Phenology].Vegetative.Target.FixedValue =90",
                    "//[Phenology].Inductive.Target.FixedValue =45"
                  ],
                  "Name": "SwitchGrass71",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.PMF.Cultivar, Models",
                  "Command": [
                    "[Leaf].Photosynthesis.RUE.FixedValue =1.75",
                    "[Leaf].ExtinctionCoefficient.FixedValue = 0.65",
                    "[Root].MaximumRootDepth.FixedValue = 3000",
                    "[Root].SpecificRootLength.FixedValue = 100",
                    "",
                    "/// frost kill ratio",
                    "[Leaf].LeafKillFraction.XYPairs.X = -35,  -5, ",
                    "[Leaf].LeafKillFraction.XYPairs.Y = 0.014, 0.0,",
                    "",
                    "///Temperature factors",
                    "//[Leaf].Photosynthesis.FT.XYPairs.X =     0, 6.5,  15,   32, 43",
                    "//[Leaf].Photosynthesis.FT.XYPairs.Y =      0,  0.0,  1,  1, 0",
                    "[Leaf].Photosynthesis.FT.MaximumTemperatureWeighting =0.75",
                    "[Leaf].MinimumLAI.FixedValue = 0.1",
                    "[Leaf].InitialWtFunction.InitialPlantWt.FixedValue = 0.01",
                    "//[Leaf].MaximumNConc.FixedValue = 0.023",
                    "//[Leaf].MinimumNConc.FixedValue = 0.0045",
                    "[Leaf].MaximumNConc.FixedValue = 0.047",
                    "[Leaf].MinimumNConc.FixedValue = 0.007",
                    "//[Phenology].Vegetative.Target.FixedValue =90",
                    "//[Phenology].Inductive.Target.FixedValue =45"
                  ],
                  "Name": "SwitchGrass711",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            }
          ],
          "Enabled": True,
          "ReadOnly": False
        }
      ],
      "Enabled": True,
      "ReadOnly": False
    },
    {
      "$child": "Models.Graph, Models",
      "Caption": None,
      "Axis": [
        {
          "$child": "APSIM.Shared.Graphing.Axis, APSIM.Shared",
          "Title": None,
          "Position": 3,
          "Inverted": False,
          "CrossesAtZero": False,
          "Minimum": None,
          "Maximum": None,
          "Interval": None
        },
        {
          "$child": "APSIM.Shared.Graphing.Axis, APSIM.Shared",
          "Title": None,
          "Position": 0,
          "Inverted": False,
          "CrossesAtZero": False,
          "Minimum": None,
          "Maximum": None,
          "Interval": None
        }
      ],
      "LegendPosition": 0,
      "LegendOrientation": 0,
      "AnnotationLocation": 0,
      "DisabledSeries": [],
      "LegendOutsideGraph": False,
      "Name": "Graph",
      "ResourceName": None,
      "Children": [
        {
          "$child": "Models.Series, Models",
          "Type": 1,
          "XAxis": 3,
          "YAxis": 0,
          "ColourArgb": -11094807,
          "FactorToVaryColours": None,
          "FactorToVaryMarkers": None,
          "FactorToVaryLines": None,
          "Marker": 11,
          "MarkerSize": 0,
          "Line": 0,
          "LineThickness": 0,
          "TableName": "Annual_",
          "XFieldName": "Date",
          "YFieldName": "runoff",
          "X2FieldName": None,
          "Y2FieldName": None,
          "ShowInLegend": True,
          "IncludeSeriesNameInLegend": False,
          "Cumulative": False,
          "CumulativeX": False,
          "Filter": None,
          "Name": "change",
          "ResourceName": None,
          "Children": [],
          "Enabled": True,
          "ReadOnly": False
        },
        {
          "$child": "Models.Series, Models",
          "Type": 1,
          "XAxis": 3,
          "YAxis": 0,
          "ColourArgb": -2793984,
          "FactorToVaryColours": None,
          "FactorToVaryMarkers": None,
          "FactorToVaryLines": None,
          "Marker": 11,
          "MarkerSize": 0,
          "Line": 0,
          "LineThickness": 0,
          "TableName": "Annual",
          "XFieldName": "Date",
          "YFieldName": "runoff",
          "X2FieldName": None,
          "Y2FieldName": None,
          "ShowInLegend": True,
          "IncludeSeriesNameInLegend": False,
          "Cumulative": False,
          "CumulativeX": False,
          "Filter": None,
          "Name": "baseline",
          "ResourceName": None,
          "Children": [],
          "Enabled": True,
          "ReadOnly": False
        }
      ],
      "Enabled": True,
      "ReadOnly": False
    },
    {
      "$child": "Models.Core.Simulation, Models",
      "Descriptors": None,
      "Name": "P30511",
      "ResourceName": None,
      "Children": [
        {
          "$child": "Models.Clock, Models",
          "Start": "2011-01-01T00:00:00",
          "End": "2021-12-31T00:00:00",
          "Name": "Clock",
          "ResourceName": None,
          "Children": [],
          "Enabled": True,
          "ReadOnly": False
        },
        {
          "$child": "Models.Summary, Models",
          "Verbosity": 100,
          "Name": "SummaryFile",
          "ResourceName": None,
          "Children": [],
          "Enabled": True,
          "ReadOnly": False
        },
        {
          "$child": "Models.Climate.Weather, Models",
          "ConstantsFile": None,
          "FileName": "C:\\Users\\rmagala\\Box\\PhD thesis\\Objective two\\data\\gGIS\\Iem_IA0200.met",
          "ExcelWorkSheetName": "",
          "Name": "Weather",
          "ResourceName": None,
          "Children": [],
          "Enabled": True,
          "ReadOnly": False
        },
        {
          "$child": "Models.Soils.Arbitrator.SoilArbitrator, Models",
          "Name": "Soil Arbitrator",
          "ResourceName": None,
          "Children": [],
          "Enabled": True,
          "ReadOnly": False
        },
        {
          "$child": "Models.Core.Zone, Models",
          "Area": 1.0,
          "Slope": 2.0,
          "AspectAngle": 0.0,
          "Altitude": 50.0,
          "Name": "Field1",
          "ResourceName": None,
          "Children": [
            {
              "$child": "Models.Report, Models",
              "VariableNames": [
                "diff of ([Soil].Nutrient.TotalC[1])/1000 from [Clock].StartOfSimulation to [Clock].EndOfSimulation as changeincarbon",
                "diff of sum([Soil].Nutrient.TotalC[1:2])/1000 from [Clock].StartOfSimulation to [Clock].EndOfSimulation as changeincarbon2",
                "",
                "changeincarbon + changeincarbon2 as to_carb",
                "",
                "(-1 * to_carb *(44/12))*1000 as co_eq"
              ],
              "EventNames": [
                "[Clock].EndOfSimulation"
              ],
              "GroupByVariableName": None,
              "Name": "Carbon",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.Report, Models",
              "VariableNames": [
                "[Clock].Today as date_in_wheat",
                "[Clock].Today.Year as Year_wheat",
                "[Clock].Today.Month as Month_in_wheat",
                "[Wheat].AboveGround.Wt *10 as w_AGB"
              ],
              "EventNames": [
                "[wheat].Harvesting"
              ],
              "GroupByVariableName": None,
              "Name": "WheatR",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.Report, Models",
              "VariableNames": [
                "// Time",
                "[Clock].Today as Date",
                "[Clock].Today.Year as an_Year",
                "",
                "// Carbon",
                "[Soil].Nutrient.TotalC[1]/1000 as SOC1",
                "sum([Soil].Nutrient.TotalC[1:2])/1000 as SOC2",
                "sum of sum([Nutrient].Catm[1:10]) from 01-jan to [Clock].Today as Whole_repsiration",
                "sum of sum([Nutrient].Catm[1:6]) from 01-jan to [Clock].Today  as  Top_respiration",
                "sum([Soil].Nutrient.TotalC[1:2])/1000 as TPC",
                "// Nitrogen ",
                "[SoilWater].FlowNO3[1] as flowNitrate",
                "sum of Sum( [Soil].NO3.ppm[1:2]) from 01-jan to [Clock].Today as NConc",
                "[Soil].NO3.ppm[1] as N03Ncon",
                "[Soil].NO3.ppm[2] as N03Ncon2",
                "[Soil].NO3.ppm[3] *0.8 as N03Ncon3",
                "Sum of [Soil].SoilWater.LeachNO3 from 01-jan to [Clock].Today as CumuAnnualLeach",
                "sum([Soil].NO3.kgha) as NO3Total",
                "",
                "sum of sum([Nutrient].N2Oatm[1:6]) from 01-Mar to [Clock].Today as TopN2O",
                "",
                "sum of sum([Nutrient].N2Oatm[1:8]) from 01-Mar to [Clock].Today as  n20",
                "sum([Soil].Nutrient.NO3.kgha[1:2]) as MineralN",
                "sum of [Soil].SoilWater.LeachNO3 from [Annual].DayAfterLastOutput to [clock].Today as LeachNO3",
                "// Water",
                "sum of ([Soil].SoilWater.Runoff) from 01-jan to [Clock].Today as runoff",
                "sum of [Soil].SoilWater.Drainage from [Annual].DayAfterLastOutput to [clock].Today as drainaged_run_off",
                "SOC1 + SOC2 as total_carbon",
                "// Calculate Nitrate Loss. 0.01 converts to kg per hectare",
                "",
                "runoff * (N03Ncon + N03Ncon2 + N03Ncon3) *0.01  as Nloss",
                "",
                "// calculalate carbon dioxide equivalent",
                "",
                " n20 *(44/28) *165 as no2_eq ",
                " TopN2O *(44/28) *165 as top_no2_eq ",
                " ",
                " // Weather and soil water",
                "",
                "sum of [weather].rain from 01-jan to [clock].Today as mean_rain",
                "sum of [weather].MaxT  from 01-jan to [Clock].Today as mean_maxt",
                "sum  of [weather].MinT from 01-jan to [Clock].Today as mean_mint",
                "sum  of [weather].Radn from 01-jan to [Clock].Today as total_radn",
                "mean  of [Water].PAW[1] from 01-jan to [Clock].Today as PAWC1",
                "mean  of [Water].PAW[2] from 01-jan to [Clock].Today as PAWC2",
                "mean  of [Soil].SoilWater.SW[1] from 01-jan to [Clock].Today as SW1",
                "mean  of [Soil].SoilWater.SW[2] from 01-jan to [Clock].Today as SW2",
                "mean  of [Soil].SoilWater.SW[3] from 01-jan to [Clock].Today as SW3"
              ],
              "EventNames": [
                "[Clock].EndOfYear"
              ],
              "GroupByVariableName": None,
              "Name": "Annual",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.Report, Models",
              "VariableNames": [
                "[Clock].Today as date_in_soybean",
                "[Clock].Today.Year as Year_in_soybean",
                "[Soybean].AboveGround.Wt*10  as soy_AGB",
                "[Soybean].Grain.Total.Wt * 10 as SoyYield"
              ],
              "EventNames": [
                "[Soybean].Harvesting"
              ],
              "GroupByVariableName": None,
              "Name": "SoybeanR",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.Surface.SurfaceOrganicMatter, Models",
              "SurfOM": [],
              "Canopies": [],
              "InitialResidueName": "grass",
              "InitialResidueType": "grass",
              "InitialResidueMass": 800.0,
              "InitialStandingFraction": 0.0,
              "InitialCPR": 0.0,
              "InitialCNR": 80.0,
              "Name": "SurfaceOrganicMatter",
              "ResourceName": "SurfaceOrganicMatter",
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.Report, Models",
              "VariableNames": [
                "[Clock].Today",
                "[Clock].Today.Year as Year",
                "[Maize].AboveGround.Wt*10  as M_AGB",
                "[Maize].AboveGround.N as MAIZE_AGB_NITROGEN",
                "[Maize].Grain.Total.Wt * 10 as Maize_Yield",
                "[Maize].Grain.Total.Wt * 10 as mYield"
              ],
              "EventNames": [
                "[Maize].Harvesting"
              ],
              "GroupByVariableName": None,
              "Name": "MaizeR",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.Fertiliser, Models",
              "Name": "Fertiliser",
              "ResourceName": "Fertiliser",
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.Plant, Models",
              "Name": "Maize",
              "ResourceName": "Maize",
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.MicroClimate, Models",
              "a_interception": 0.0,
              "b_interception": 1.0,
              "c_interception": 0.0,
              "d_interception": 0.0,
              "SoilHeatFluxFraction": 0.4,
              "MinimumHeightDiffForNewLayer": 0.0,
              "NightInterceptionFraction": 0.5,
              "ReferenceHeight": 2.0,
              "Name": "MicroClimate",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.Plant, Models",
              "Name": "Soybean",
              "ResourceName": "Soybean",
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.PMF.Plant, Models",
              "Name": "Wheat",
              "ResourceName": "Wheat",
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.Manager, Models",
              "Code": "using Models.Soils;\nusing System;\nusing Models.Core;\nusing Models.PMF;\nusing Models.Climate;\nusing APSIM.Shared.Utilities;\nnamespace Models\n\n{\n    [Serializable]\n    public class Script : Model\n    {\n        [Link] Clock Clock;\n        \n        [Link] Fertiliser Fertiliser;\n        [Link] ISummary Summary;\n        [Link] private Weather Weather;\n        [Separator(\"Crop in the rotation to apply fertilizers\")]\n        [Description(\"Crop\")]\n        public IPlant Crop { get; set; }\n        \n        [Separator(\"Fertilizer application details\")]\n\n        [Description(\"Type of fertiliser to apply? Select from a list. \")] \n        public Fertiliser.Types FertiliserType { get; set; }\n        \n        [Description(\"Depth to apply the fertilizer. \")] \n        public double Depth2Apply { get; set; }\n        \n        [Description(\"Fraction to apply at sowing (0-1). if it is a single application, it should be 1\")]\n        public double FractionToApplyFirst { get; set; }\n\n        [Description(\"Enter the total amount of fertiliser to be applied (kg /ha)\")] \n        public double Amount { get; set; }\n        \n        [Description(\"Enter other potential dates as dd-mmm with comma separation (any year information entered will be ignored): \")] \n        public string[] FertiliserDates { get; set; } \n        \n        \n        private double Rain;\n        \n        [EventSubscribe(\"Sowing\")]\n        private void OnSowing(object sender, EventArgs e)\n        {\n            Model crop = sender as Model;\n            var nit = FractionToApplyFirst * Amount;\n            if (Crop != None && crop.Name.ToLower() == (Crop as IModel).Name.ToLower())\n                Fertiliser.Apply(Amount: nit, Type: FertiliserType, Depth:Depth2Apply);\n        }\n        \n        [EventSubscribe(\"DoManagement\")]\n        private void OnDoManagement(object sender, EventArgs e)\n        {\n         Depth2Apply  = 0.0;\n         var Depth = 0.0;\n         \n        \n            foreach (string ddMMM in FertiliserDates)\n            {\n             Rain = Weather.Rain;\n             \n             var nit = FractionToApplyFirst * Amount;\n             var otherNapp = Amount - nit;\n                if (Crop.IsAlive & DateUtilities.DatesEqual(ddMMM, Clock.Today)& Rain ==0)\n                {\n                        Fertiliser.Apply(Amount: otherNapp / FertiliserDates.Length, Type: FertiliserType, Depth: Depth2Apply);\n                        Summary.WriteMessage(this, $\"At other date(s), {Crop.Name} received {otherNapp} Kg of Nitrogen applied as {FertiliserType}.\", MessageType.Information);\n                        Summary.WriteMessage(this, $\"Rain was:, {Rain}\",  MessageType.Information);\n                    break;   \n                }\n            }\n        }\n    }\n}\n",
              "Parameters": [
                {
                  "Key": "Crop",
                  "Value": "[Maize]"
                },
                {
                  "Key": "FertiliserType",
                  "Value": "UreaNO3"
                },
                {
                  "Key": "Depth2Apply",
                  "Value": "0"
                },
                {
                  "Key": "FractionToApplyFirst",
                  "Value": "1"
                },
                {
                  "Key": "Amount",
                  "Value": "202"
                },
                {
                  "Key": "FertiliserDates",
                  "Value": "29-may, 30-may, 31-may, 01-jul, 02-jul"
                }
              ],
              "Name": "MaizeNitrogenManager",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.Manager, Models",
              "Code": "using Models.Interfaces;\r\nusing APSIM.Shared.Utilities;\r\nusing Models.Utilities;\r\nusing Models.Soils;\r\nusing Models.PMF;\r\nusing Models.Core;\r\nusing System;\r\nusing Models.Climate;\r\n\r\nnamespace Models\r\n{\r\n    [Serializable]\r\n    public class Script : Model\r\n    {\r\n        [Link] private Clock Clock;\r\n        [Link] private Summary Summary;\r\n        private Accumulator accumulatedRain;\r\n        [Link]\r\n        private ISoilWater waterBalance;\n        \n        [Description(\"Simple Rotation Manager\")]\n        public Manager Rotations { get; set; }\n        \n        [Separator(\"Sowing Info\")]\r\n        \r\n        [Description(\"Crop\")]\r\n        public IPlant Crop { get; set; }\r\n\r\n        [Description(\"Start of sowing window (d-mmm)\")]\r\n        public string StartDate { get; set; }\r\n\r\n        [Description(\"End of sowing window (d-mmm)\")]\r\n        public string EndDate { get; set; }\r\n\r\n        [Description(\"Minimum extractable soil water for sowing (mm)\")]\r\n        public double MinESW { get; set; }\r\n\r\n        [Description(\"Accumulated rainfall required for sowing (mm)\")]\r\n        public double MinRain { get; set; }\r\n\r\n        [Description(\"Duration of rainfall accumulation (d)\")]\r\n        public int RainDays { get; set; }\r\n\r\n        [Display(Type = DisplayType.CultivarName)]\r\n        [Description(\"Cultivar to be sown\")]\r\n        public string CultivarName { get; set; }\r\n\r\n        [Description(\"Sowing depth (mm)\")]\r\n        public double SowingDepth { get; set; }\r\n\r\n        [Description(\"Row spacing (mm)\")]\r\n        public double RowSpacing { get; set; }\r\n\r\n        [Description(\"Plant population (/m2)\")]\r\n        public double Population { get; set; }\n        \n        [Description(\"Must sow by end date?\")]\n        public bool MustSow { get; set; }\r\n        \r\n        \r\n        [EventSubscribe(\"StartOfSimulation\")]\r\n        private void OnSimulationCommencing(object sender, EventArgs e)\r\n        {\r\n            accumulatedRain = new Accumulator(this, \"[Weather].Rain\", RainDays);\r\n        }\r\n\r\n        [EventSubscribe(\"DoManagement\")]\r\n        private void OnDoManagement(object sender, EventArgs e)\r\n        {\r\n            accumulatedRain.Update();\r\n            if (!Crop.IsAlive &&\n           \t    DateUtilities.WithinDates(StartDate, Clock.Today, EndDate))\n           \t{\n           \t\tvar rainSum = accumulatedRain.Sum;\n           \t\tvar eswSum  = MathUtilities.Sum(waterBalance.ESW);\n           \t\tif (rainSum > MinRain && eswSum > MinESW ||\n           \t\t    DateUtilities.DatesEqual(EndDate, Clock.Today) && MustSow)\n           \t\t{\n           \t\t\tvar thisCrop = Crop.Name.ToLower();\n           \t\t\tvar nextCrop = (string)Rotations?.FindByPath(\"Script.NextCrop\")?.Value;\n           \t\t\tif (thisCrop == (nextCrop?.ToLower() ?? thisCrop))\n           \t\t\t\tCrop.Sow(population: Population,\n                \t\t\t\t cultivar: CultivarName,\n                \t\t\t\t depth: SowingDepth,\n                \t\t\t\t rowSpacing: RowSpacing);\n           \t\t}\n           \t}\r\n        }\r\n    }\r\n}\r\n",
              "Parameters": [
                {
                  "Key": "Rotations",
                  "Value": "[Simple Rotation]"
                },
                {
                  "Key": "Crop",
                  "Value": "[Maize]"
                },
                {
                  "Key": "StartDate",
                  "Value": "1-May"
                },
                {
                  "Key": "EndDate",
                  "Value": "15-May"
                },
                {
                  "Key": "MinESW",
                  "Value": "200"
                },
                {
                  "Key": "MinRain",
                  "Value": "30"
                },
                {
                  "Key": "RainDays",
                  "Value": "3"
                },
                {
                  "Key": "CultivarName",
                  "Value": "B_110"
                },
                {
                  "Key": "SowingDepth",
                  "Value": "50"
                },
                {
                  "Key": "RowSpacing",
                  "Value": "750"
                },
                {
                  "Key": "Population",
                  "Value": "8.65"
                },
                {
                  "Key": "MustSow",
                  "Value": "True"
                }
              ],
              "Name": "SowMaize",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.Manager, Models",
              "Code": "using Models.Interfaces;\r\nusing APSIM.Shared.Utilities;\r\nusing Models.Utilities;\r\nusing Models.Soils;\r\nusing Models.PMF;\r\nusing Models.Core;\r\nusing System;\r\nusing Models.Climate;\r\n\r\nnamespace Models\r\n{\r\n    [Serializable]\r\n    public class Script : Model\r\n    {\r\n        [Link] private Clock Clock;\r\n        [Link] private Summary Summary;\r\n        private Accumulator accumulatedRain;\r\n        [Link]\r\n        private ISoilWater waterBalance;\n        \n        [Description(\"Simple Rotation Manager\")]\n        public Manager Rotations { get; set; }\n        \n        [Separator(\"Sowing Info\")]\r\n        \r\n        [Description(\"Crop\")]\r\n        public IPlant Crop { get; set; }\r\n\r\n        [Description(\"Start of sowing window (d-mmm)\")]\r\n        public string StartDate { get; set; }\r\n\r\n        [Description(\"End of sowing window (d-mmm)\")]\r\n        public string EndDate { get; set; }\r\n\r\n        [Description(\"Minimum extractable soil water for sowing (mm)\")]\r\n        public double MinESW { get; set; }\r\n\r\n        [Description(\"Accumulated rainfall required for sowing (mm)\")]\r\n        public double MinRain { get; set; }\r\n\r\n        [Description(\"Duration of rainfall accumulation (d)\")]\r\n        public int RainDays { get; set; }\r\n\r\n        [Display(Type = DisplayType.CultivarName)]\r\n        [Description(\"Cultivar to be sown\")]\r\n        public string CultivarName { get; set; }\r\n\r\n        [Description(\"Sowing depth (mm)\")]\r\n        public double SowingDepth { get; set; }\r\n\r\n        [Description(\"Row spacing (mm)\")]\r\n        public double RowSpacing { get; set; }\r\n\r\n        [Description(\"Plant population (/m2)\")]\r\n        public double Population { get; set; }\n        \n        [Description(\"Must sow by end date?\")]\n        public bool MustSow { get; set; }\r\n        \r\n        \r\n        [EventSubscribe(\"StartOfSimulation\")]\r\n        private void OnSimulationCommencing(object sender, EventArgs e)\r\n        {\r\n            accumulatedRain = new Accumulator(this, \"[Weather].Rain\", RainDays);\r\n        }\r\n\r\n        [EventSubscribe(\"DoManagement\")]\r\n        private void OnDoManagement(object sender, EventArgs e)\r\n        {\r\n            accumulatedRain.Update();\r\n            if (!Crop.IsAlive &&\n           \t    DateUtilities.WithinDates(StartDate, Clock.Today, EndDate))\n           \t{\n           \t\tvar rainSum = accumulatedRain.Sum;\n           \t\tvar eswSum  = MathUtilities.Sum(waterBalance.ESW);\n           \t\tif (rainSum > MinRain && eswSum > MinESW ||\n           \t\t    DateUtilities.DatesEqual(EndDate, Clock.Today) && MustSow)\n           \t\t{\n           \t\t\tvar thisCrop = Crop.Name.ToLower();\n           \t\t\tvar nextCrop = (string)Rotations?.FindByPath(\"Script.NextCrop\")?.Value;\n           \t\t\tif (thisCrop == (nextCrop?.ToLower() ?? thisCrop))\n           \t\t\t\tCrop.Sow(population: Population,\n                \t\t\t\t cultivar: CultivarName,\n                \t\t\t\t depth: SowingDepth,\n                \t\t\t\t rowSpacing: RowSpacing);\n           \t\t}\n           \t}\r\n        }\r\n    }\r\n}\r\n",
              "Parameters": [
                {
                  "Key": "Rotations",
                  "Value": "[Simple Rotation]"
                },
                {
                  "Key": "Crop",
                  "Value": "[Wheat]"
                },
                {
                  "Key": "StartDate",
                  "Value": "27-Sep"
                },
                {
                  "Key": "EndDate",
                  "Value": "5-Oct"
                },
                {
                  "Key": "MinESW",
                  "Value": "100"
                },
                {
                  "Key": "MinRain",
                  "Value": "15"
                },
                {
                  "Key": "RainDays",
                  "Value": "7"
                },
                {
                  "Key": "CultivarName",
                  "Value": "YecoraCovercrop"
                },
                {
                  "Key": "SowingDepth",
                  "Value": "10"
                },
                {
                  "Key": "RowSpacing",
                  "Value": "250"
                },
                {
                  "Key": "Population",
                  "Value": "50"
                },
                {
                  "Key": "MustSow",
                  "Value": "True"
                }
              ],
              "Name": "SowWheat",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.Manager, Models",
              "Code": "using APSIM.Shared.Utilities;\r\nusing Models.Utilities;\r\nusing Models.Soils;\r\nusing Models.PMF;\r\nusing Models.Core;\r\nusing System;\r\n\r\nnamespace Models\r\n{\r\n    [Serializable] \r\n    public class Script : Model\r\n    {\n    \t[Link] IClock Clock;\n    \t\n    \t[Separator(\"Harvesting script that forces harvesting\")]\r\n        [Description(\"Crop\")]\r\n        public IPlant Crop { get; set; }\n        \n        [Description(\"Date to force harvesting by if crop is not ready (d-mmm)\")]\n        public string Date { get; set; }\r\n        \r\n        [EventSubscribe(\"DoManagement\")]\r\n        private void OnDoManagement(object sender, EventArgs e)\r\n        {\r\n            if (Crop.IsReadyForHarvesting || \n                (DateUtilities.DatesAreEqual(Date, Clock.Today) && Crop.IsAlive))\r\n            {\r\n                Crop.Harvest();\r\n                Crop.EndCrop();\r\n            }\r\n        }\r\n    }\r\n}\r\n",
              "Parameters": [
                {
                  "Key": "Crop",
                  "Value": "[Maize]"
                },
                {
                  "Key": "Date",
                  "Value": "25-Sep"
                }
              ],
              "Name": "HarvestMaize",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.Manager, Models",
              "Code": "using Models.Interfaces;\r\nusing APSIM.Shared.Utilities;\r\nusing Models.Utilities;\r\nusing Models.Soils;\r\nusing Models.PMF;\r\nusing Models.Core;\r\nusing System;\r\nusing Models.Climate;\r\n\r\nnamespace Models\r\n{\r\n    [Serializable]\r\n    public class Script : Model\r\n    {\r\n        [Link] private Clock Clock;\r\n        [Link] private Summary Summary;\r\n        private Accumulator accumulatedRain;\r\n        [Link]\r\n        private ISoilWater waterBalance;\n        \n        [Description(\"Simple Rotation Manager\")]\n        public Manager Rotations { get; set; }\n        \n        [Separator(\"Sowing Info\")]\r\n        \r\n        [Description(\"Crop\")]\r\n        public IPlant Crop { get; set; }\r\n\r\n        [Description(\"Start of sowing window (d-mmm)\")]\r\n        public string StartDate { get; set; }\r\n\r\n        [Description(\"End of sowing window (d-mmm)\")]\r\n        public string EndDate { get; set; }\r\n\r\n        [Description(\"Minimum extractable soil water for sowing (mm)\")]\r\n        public double MinESW { get; set; }\r\n\r\n        [Description(\"Accumulated rainfall required for sowing (mm)\")]\r\n        public double MinRain { get; set; }\r\n\r\n        [Description(\"Duration of rainfall accumulation (d)\")]\r\n        public int RainDays { get; set; }\r\n\r\n        [Display(Type = DisplayType.CultivarName)]\r\n        [Description(\"Cultivar to be sown\")]\r\n        public string CultivarName { get; set; }\r\n\r\n        [Description(\"Sowing depth (mm)\")]\r\n        public double SowingDepth { get; set; }\r\n\r\n        [Description(\"Row spacing (mm)\")]\r\n        public double RowSpacing { get; set; }\r\n\r\n        [Description(\"Plant population (/m2)\")]\r\n        public double Population { get; set; }\n        \n        [Description(\"Must sow by end date?\")]\n        public bool MustSow { get; set; }\r\n        \r\n        \r\n        [EventSubscribe(\"StartOfSimulation\")]\r\n        private void OnSimulationCommencing(object sender, EventArgs e)\r\n        {\r\n            accumulatedRain = new Accumulator(this, \"[Weather].Rain\", RainDays);\r\n        }\r\n\r\n        [EventSubscribe(\"DoManagement\")]\r\n        private void OnDoManagement(object sender, EventArgs e)\r\n        {\r\n            accumulatedRain.Update();\r\n            if (!Crop.IsAlive &&\n           \t    DateUtilities.WithinDates(StartDate, Clock.Today, EndDate))\n           \t{\n           \t\tvar rainSum = accumulatedRain.Sum;\n           \t\tvar eswSum  = MathUtilities.Sum(waterBalance.ESW);\n           \t\tif (rainSum > MinRain && eswSum > MinESW ||\n           \t\t    DateUtilities.DatesEqual(EndDate, Clock.Today) && MustSow)\n           \t\t{\n           \t\t\tvar thisCrop = Crop.Name.ToLower();\n           \t\t\tvar nextCrop = (string)Rotations?.FindByPath(\"Script.NextCrop\")?.Value;\n           \t\t\tif (thisCrop == (nextCrop?.ToLower() ?? thisCrop))\n           \t\t\t\tCrop.Sow(population: Population,\n                \t\t\t\t cultivar: CultivarName,\n                \t\t\t\t depth: SowingDepth,\n                \t\t\t\t rowSpacing: RowSpacing);\n           \t\t}\n           \t}\r\n        }\r\n    }\r\n}\r\n",
              "Parameters": [
                {
                  "Key": "Rotations",
                  "Value": "[Simple Rotation]"
                },
                {
                  "Key": "Crop",
                  "Value": "[Soybean]"
                },
                {
                  "Key": "StartDate",
                  "Value": "15-may"
                },
                {
                  "Key": "EndDate",
                  "Value": "25-May"
                },
                {
                  "Key": "MinESW",
                  "Value": "200"
                },
                {
                  "Key": "MinRain",
                  "Value": "30"
                },
                {
                  "Key": "RainDays",
                  "Value": "3"
                },
                {
                  "Key": "CultivarName",
                  "Value": "Generic_MG2"
                },
                {
                  "Key": "SowingDepth",
                  "Value": "50"
                },
                {
                  "Key": "RowSpacing",
                  "Value": "750"
                },
                {
                  "Key": "Population",
                  "Value": "30"
                },
                {
                  "Key": "MustSow",
                  "Value": "True"
                }
              ],
              "Name": "SowSoy",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.Manager, Models",
              "Code": "using APSIM.Shared.Utilities;\r\nusing Models.Utilities;\r\nusing Models.Soils;\r\nusing Models.PMF;\r\nusing Models.Core;\r\nusing System;\r\n\r\nnamespace Models\r\n{\r\n    [Serializable] \r\n    public class Script : Model\r\n    {\n    \t[Link] IClock Clock;\n    \t\n    \t[Separator(\"Harvesting script that forces harvesting\")]\r\n        [Description(\"Crop\")]\r\n        public IPlant Crop { get; set; }\n        \n        [Description(\"Date to force harvesting by if crop is not ready (d-mmm)\")]\n        public string Date { get; set; }\r\n        \r\n        [EventSubscribe(\"DoManagement\")]\r\n        private void OnDoManagement(object sender, EventArgs e)\r\n        {\r\n            if (Crop.IsReadyForHarvesting || \n                (DateUtilities.DatesAreEqual(Date, Clock.Today) && Crop.IsAlive))\r\n            {\r\n                Crop.Harvest();\r\n                Crop.EndCrop();\r\n            }\r\n        }\r\n    }\r\n}\r\n",
              "Parameters": [
                {
                  "Key": "Crop",
                  "Value": "[Soybean]"
                },
                {
                  "Key": "Date",
                  "Value": "19-sep"
                }
              ],
              "Name": "HarvestSoy",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.Manager, Models",
              "Code": "using Models.Soils;\r\nusing System.Collections.Generic;\r\nusing APSIM.Shared.Utilities;\r\nusing Models.PMF;\r\nusing Models.Interfaces;\r\nusing Models.Core;\r\nusing System;\r\n\r\nnamespace Models\r\n{\r\n    [Serializable]\r\n    public class Script : Model\r\n    {\r\n        [Link] IClock Clock;\r\n        [Link] ISummary Summary;\r\n        \r\n        [Description(\"Rotation Sequence (comma separated crop names)\")]\r\n        public string[] Crops { get; set; }\r\n        \r\n        // Repeating cycle of input crops. Don't loop on this, loop on Crops\r\n        private IEnumerable<string> cropsEnum()\r\n        {\r\n            var counter = 0;\r\n            while (True)\r\n            {\r\n                yield return Crops[counter];\r\n                ++counter;\r\n                if (counter >= Crops.Length)\r\n                    counter = 0;\r\n            }\r\n        }\r\n        \r\n        private IEnumerator<string> cropCycle;\r\n        \r\n        [Description(\"Next crop to sow in the simple rotation.\")]\r\n        public string NextCrop\r\n        { get => cropCycle.Current; }\r\n        \r\n        [EventSubscribe(\"StartOfSimulation\")]\r\n        private void Init(object sender, EventArgs e)\r\n        {\r\n            // TODO: Error if Crops is empty.\r\n            cropCycle = cropsEnum().GetEnumerator();\r\n            // Set to first value.\r\n            cropCycle.MoveNext();\r\n        }\r\n\r\n        [EventSubscribe(\"PlantEnding\")]\r\n        private void OnEndCrop(object sender, EventArgs e)\r\n        {\r\n            // Currently assumes only one crop in the ground, so does not check if the  \r\n            // harvested crop is this script's current crop.\r\n            var old = NextCrop;\r\n            cropCycle.MoveNext();\r\n            var _new = NextCrop;\r\n            Summary.WriteMessage(this, $\"Rotation finished with {old}, next up: {_new}.\", MessageType.Information);\r\n        }\r\n    }\r\n}\r\n",
              "Parameters": [
                {
                  "Key": "Crops",
                  "Value": "Maize,Soybean,Wheat,Maize,Soybean"
                }
              ],
              "Name": "Simple Rotation",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.Manager, Models",
              "Code": "using Models.Soils;\r\nusing System;\r\nusing Models.Core;\r\nusing Models.PMF;\r\nusing APSIM.Shared.Utilities;\r\nnamespace Models\r\n{\r\n    [Serializable]\r\n    public class Script : Model\r\n    {\r\n        [Link] Clock Clock;\r\n        \r\n        [Link] Fertiliser Fertiliser;\r\n        [Link] ISummary Summary;\r\n        [Separator(\"Crop in the rotation to apply fertilizers\")]\r\n        [Description(\"Crop\")]\r\n        public IPlant Crop { get; set; }\r\n        \r\n        [Separator(\"Fertilizer application details\")]\r\n\r\n        [Description(\"Type of fertiliser to apply? Select from a list. \")] \r\n        public Fertiliser.Types FertiliserType { get; set; }\r\n        \r\n        [Description(\"Fraction to apply at sowing (0-1). if it is a single application, it should be 1\")]\r\n        public double Fraction { get; set; }\r\n\r\n        [Description(\"Enter the total amount of fertiliser to be applied (kg /ha)\")] \r\n        public double Amount { get; set; }\r\n        \r\n        [Description(\"Enter other dates as dd-mmm with comma separation (any year information entered will be ignored): \")] \r\n        public string[] FertiliserDates { get; set; } \r\n        \r\n        [Description(\"Is the remaining amount to be applied each time? (True/ticked)? Or is it the total annual amount across all dates (no(False)/unticked)\")] \r\n        public bool AmountType { get; set; }\r\n        \r\n        [EventSubscribe(\"Sowing\")]\r\n        private void OnSowing(object sender, EventArgs e)\r\n        {\r\n            Model crop = sender as Model;\r\n            var nit = Fraction * Amount;\r\n            if (Crop != None && crop.Name.ToLower() == (Crop as IModel).Name.ToLower())\r\n                Fertiliser.Apply(Amount: nit, Type: FertiliserType);\r\n        }\r\n        \r\n        [EventSubscribe(\"DoManagement\")]\r\n        private void OnDoManagement(object sender, EventArgs e)\r\n        {\r\n            foreach (string ddMMM in FertiliserDates)\r\n            {\r\n             var nit = Fraction * Amount;\r\n             var otherNapp = Amount - nit;\r\n                if (Crop.IsAlive & DateUtilities.DatesEqual(ddMMM, Clock.Today))\r\n                {\r\n                    if (AmountType)\r\n                        Fertiliser.Apply(Amount: otherNapp, Type: FertiliserType);\r\n                        \r\n                    else\r\n                        Fertiliser.Apply(Amount: otherNapp / FertiliserDates.Length, Type: FertiliserType);\r\n                        Summary.WriteMessage(this, $\"At other date(s), {Crop.Name} received {otherNapp} Kg of Nitrogen applied as {FertiliserType}. {AmountType} was selected: True implies that the user specified amount was applied on each date and False implies that the amount was applied across dates\", MessageType.Information);\r\n                }\r\n            }\r\n        }\r\n    }\r\n}\r\n",
              "Parameters": [
                {
                  "Key": "Crop",
                  "Value": "[Wheat]"
                },
                {
                  "Key": "FertiliserType",
                  "Value": "UreaN"
                },
                {
                  "Key": "Fraction",
                  "Value": "0.3777778"
                },
                {
                  "Key": "Amount",
                  "Value": "NaN"
                },
                {
                  "Key": "FertiliserDates",
                  "Value": "20-march"
                },
                {
                  "Key": "AmountType",
                  "Value": "False"
                }
              ],
              "Name": "AddfertlizerRotationWheat",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.Manager, Models",
              "Code": "using APSIM.Shared.Utilities;\r\nusing Models.Utilities;\r\nusing Models.Soils;\r\nusing Models.PMF;\r\nusing Models.Core;\r\nusing System;\r\n\r\nnamespace Models\r\n{\r\n    [Serializable] \r\n    public class Script : Model\r\n    {\n    \t[Link] IClock Clock;\n    \t\n    \t[Separator(\"Harvesting script that forces harvesting\")]\r\n        [Description(\"Crop\")]\r\n        public IPlant Crop { get; set; }\n        \n        [Description(\"Date to force harvesting by if crop is not ready (d-mmm)\")]\n        public string Date { get; set; }\r\n        \r\n        [EventSubscribe(\"DoManagement\")]\r\n        private void OnDoManagement(object sender, EventArgs e)\r\n        {\r\n            if (Crop.IsReadyForHarvesting || \n                (DateUtilities.DatesAreEqual(Date, Clock.Today) && Crop.IsAlive))\r\n            {\r\n                Crop.Harvest();\r\n                Crop.EndCrop();\r\n            }\r\n        }\r\n    }\r\n}\r\n",
              "Parameters": [
                {
                  "Key": "Crop",
                  "Value": "[Wheat]"
                },
                {
                  "Key": "Date",
                  "Value": "28-apr"
                }
              ],
              "Name": "HarvestWheat",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.Soils.Soil, Models",
              "RecordNumber": 1,
              "ASCOrder": None,
              "ASCSubOrder": None,
              "SoilType": "Shaffton:67263817",
              "LocalName": None,
              "Site": "11",
              "NearestTown": None,
              "Region": None,
              "State": "IL",
              "Country": "USA",
              "NaturalVegetation": None,
              "ApsoilNumber": "0",
              "Latitude": 41.76297555833147,
              "Longitude": -90.6807514145342,
              "LocationAccuracy": None,
              "YearOfSampling": None,
              "DataSource": "SSURGO https://SDMDataAccess.nrcs.usda.gov/Tabular/SDMTabularService.asmx through PyAPSIMX function Replace_Soilprofile. Date downloaded: 2022-12-27 02:49:44.558240",
              "Comments": "cockey = 22743816 Shaffton loam, 0 to 2 percent slopes Component percentage: 95%",
              "Name": "Muscatune:244911",
              "ResourceName": None,
              "Children": [
                {
                  "$child": "Models.Soils.Physical, Models",
                  "Thickness": [
                    200.0,
                    100.0,
                    200.0,
                    200.0,
                    200.0,
                    250.0,
                    300.0,
                    300.0,
                    350.0,
                    250.0
                  ],
                  "ParticleSizeClay": [
                    32.0,
                    31.759033203125,
                    31.51806640625,
                    31.277099609375,
                    31.0361328125,
                    29.397460937499993,
                    27.514477539062494,
                    26.057324218749997,
                    25.006823730468746,
                    23.956323242187494
                  ],
                  "ParticleSizeSand": [
                    1.7001953125,
                    2.0134521484375,
                    2.326708984375,
                    2.6399658203125007,
                    2.953222656250001,
                    2.8182373046874996,
                    2.604666137695312,
                    2.3253906249999985,
                    1.9834472656249984,
                    1.641503906249998
                  ],
                  "ParticleSizeSilt": [
                    66.3125,
                    66.23719787597656,
                    66.16189575195312,
                    66.08659362792969,
                    66.01129150390625,
                    67.7724609375,
                    69.85510253906251,
                    71.593359375,
                    73.00014648437501,
                    74.40693359375001
                  ],
                  "Rocks": None,
                  "Texture": None,
                  "BD": [
                    1.35,
                    1.3625147928994086,
                    1.3750295857988168,
                    1.387544378698225,
                    1.4000591715976334,
                    1.4203321678321679,
                    1.420804195804196,
                    1.4425174825174827,
                    1.521346153846154,
                    1.5810576923076924
                  ],
                  "AirDry": [
                    0.057499999999999996,
                    0.08873004844617748,
                    0.08546009689235495,
                    0.10273768167316552,
                    0.09865024223088738,
                    0.09303100125680769,
                    0.08714369501466272,
                    0.08187306242144951,
                    0.07719340874179581,
                    0.07251375506214212
                  ],
                  "LL15": [
                    0.11499999999999999,
                    0.11091256055772183,
                    0.10682512111544368,
                    0.10273768167316552,
                    0.09865024223088738,
                    0.09303100125680769,
                    0.08714369501466272,
                    0.08187306242144951,
                    0.07719340874179581,
                    0.07251375506214212
                  ],
                  "DUL": [
                    0.45499999999999996,
                    0.4329251506625542,
                    0.4108503013251085,
                    0.3887754519876628,
                    0.3686449949074143,
                    0.36670060265021703,
                    0.3594339116377837,
                    0.3568958338196487,
                    0.3502228283681532,
                    0.3492558697205646
                  ],
                  "SAT": [
                    0.47056603773584904,
                    0.46584347437758167,
                    0.4611209110193144,
                    0.45639834766104714,
                    0.45167578430277977,
                    0.44402559704446487,
                    0.4438474732814355,
                    0.435653780182082,
                    0.40590711175616834,
                    0.38337445573294626
                  ],
                  "KS": [
                    792.4499999999999,
                    792.4499999999999,
                    792.4499999999999,
                    792.4499999999999,
                    792.4499999999999,
                    792.4499999999999,
                    792.4499999999999,
                    792.4499999999999,
                    792.4499999999999,
                    792.4499999999999
                  ],
                  "BDMetadata": None,
                  "AirDryMetadata": None,
                  "LL15Metadata": None,
                  "DULMetadata": None,
                  "SATMetadata": None,
                  "KSMetadata": None,
                  "RocksMetadata": None,
                  "TextureMetadata": None,
                  "ParticleSizeSandMetadata": None,
                  "ParticleSizeSiltMetadata": None,
                  "ParticleSizeClayMetadata": None,
                  "Name": "Physical",
                  "ResourceName": None,
                  "Children": [
                    {
                      "$child": "Models.Soils.SoilCrop, Models",
                      "LL": [
                        0.11499999999999999,
                        0.11091256055772183,
                        0.10682512111544368,
                        0.10273768167316552,
                        0.09865024223088738,
                        0.09303100125680769,
                        0.08714369501466272,
                        0.08187306242144951,
                        0.07719340874179581,
                        0.07251375506214212
                      ],
                      "KL": [
                        0.08,
                        0.06549846024623855,
                        0.053625603682851145,
                        0.04390493088752211,
                        0.03594631712937773,
                        0.029430355293715384,
                        0.024095536952976167,
                        0.019727757115328516,
                        0.016151721439572435,
                        0.013223911057726925
                      ],
                      "XF": [
                        1.0,
                        1.0,
                        1.0,
                        1.0,
                        1.0,
                        1.0,
                        1.0,
                        1.0,
                        1.0,
                        1.0
                      ],
                      "LLMetadata": None,
                      "KLMetadata": None,
                      "XFMetadata": None,
                      "Name": "MaizeSoil",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Soils.SoilCrop, Models",
                      "LL": [
                        0.11499999999999999,
                        0.11091256055772183,
                        0.10682512111544368,
                        0.10273768167316552,
                        0.09865024223088738,
                        0.09303100125680769,
                        0.08714369501466272,
                        0.08187306242144951,
                        0.07719340874179581,
                        0.07251375506214212
                      ],
                      "KL": [
                        0.08,
                        0.06549846024623855,
                        0.053625603682851145,
                        0.04390493088752211,
                        0.03594631712937773,
                        0.029430355293715384,
                        0.024095536952976167,
                        0.019727757115328516,
                        0.016151721439572435,
                        0.013223911057726925
                      ],
                      "XF": [
                        1.0,
                        1.0,
                        1.0,
                        1.0,
                        1.0,
                        1.0,
                        1.0,
                        1.0,
                        1.0,
                        1.0
                      ],
                      "LLMetadata": None,
                      "KLMetadata": None,
                      "XFMetadata": None,
                      "Name": "SoybeanSoil",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Soils.SoilCrop, Models",
                      "LL": [
                        0.11499999999999999,
                        0.11091256055772183,
                        0.10682512111544368,
                        0.10273768167316552,
                        0.09865024223088738,
                        0.09303100125680769,
                        0.08714369501466272,
                        0.08187306242144951,
                        0.07719340874179581,
                        0.07251375506214212
                      ],
                      "KL": [
                        0.08,
                        0.06549846024623855,
                        0.053625603682851145,
                        0.04390493088752211,
                        0.03594631712937773,
                        0.029430355293715384,
                        0.024095536952976167,
                        0.019727757115328516,
                        0.016151721439572435,
                        0.013223911057726925
                      ],
                      "XF": [
                        1.0,
                        1.0,
                        1.0,
                        1.0,
                        1.0,
                        1.0,
                        1.0,
                        1.0,
                        1.0,
                        1.0
                      ],
                      "LLMetadata": None,
                      "KLMetadata": None,
                      "XFMetadata": None,
                      "Name": "WheatSoil",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    },
                    {
                      "$child": "Models.Soils.SoilCrop, Models",
                      "LL": [
                        0.11499999999999999,
                        0.11091256055772183,
                        0.10682512111544368,
                        0.10273768167316552,
                        0.09865024223088738,
                        0.09303100125680769,
                        0.08714369501466272,
                        0.08187306242144951,
                        0.07719340874179581,
                        0.07251375506214212
                      ],
                      "KL": [
                        0.08,
                        0.06549846024623855,
                        0.053625603682851145,
                        0.04390493088752211,
                        0.03594631712937773,
                        0.029430355293715384,
                        0.024095536952976167,
                        0.019727757115328516,
                        0.016151721439572435,
                        0.013223911057726925
                      ],
                      "XF": [
                        1.0,
                        1.0,
                        1.0,
                        1.0,
                        1.0,
                        1.0,
                        1.0,
                        1.0,
                        1.0,
                        1.0
                      ],
                      "LLMetadata": None,
                      "KLMetadata": None,
                      "XFMetadata": None,
                      "Name": "TropicalPastureSoil",
                      "ResourceName": None,
                      "Children": [],
                      "Enabled": True,
                      "ReadOnly": False
                    }
                  ],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Soils.Organic, Models",
                  "FOMCNRatio": 40.0,
                  "Thickness": [
                    200.0,
                    100.0,
                    200.0,
                    200.0,
                    200.0,
                    250.0,
                    300.0,
                    300.0,
                    350.0,
                    250.0
                  ],
                  "Carbon": [
                    2.4699999999999998,
                    1.71787533364344,
                    1.1947755716358548,
                    0.8309617343129455,
                    0.5779306342420175,
                    0.4019484943810814,
                    0.2795536048147282,
                    0.19442843811430266,
                    0.13522421781189506,
                    0.09404791428756372
                  ],
                  "CarbonUnits": 0,
                  "SoilCNRatio": [
                    12.0,
                    12.0,
                    12.0,
                    12.0,
                    12.0,
                    12.0,
                    12.0,
                    12.0,
                    12.0,
                    12.0
                  ],
                  "FBiom": [
                    0.0395,
                    0.035,
                    0.030164402071603768,
                    0.024696523624231186,
                    0.020219803385274973,
                    0.0165545748527149,
                    0.013553739536049093,
                    0.01109686337737229,
                    0.009085343309759493,
                    0.007438449969971395
                  ],
                  "FInert": [
                    0.65,
                    0.668,
                    0.8467671122222075,
                    0.8552772631814191,
                    0.8638729425796824,
                    0.8725550099921,
                    0.8813243336326486,
                    0.8901817904409999,
                    0.8991282661702156,
                    0.9081646554753248
                  ],
                  "FOM": [
                    150.0,
                    122.80961296169728,
                    100.5480069053459,
                    82.32174541410396,
                    67.39934461758324,
                    55.18191617571634,
                    45.17913178683031,
                    36.98954459124097,
                    30.284477699198312,
                    24.794833233237984
                  ],
                  "CarbonMetadata": None,
                  "FOMMetadata": None,
                  "Name": "Organic",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Soils.Chemical, Models",
                  "Thickness": [
                    200.0,
                    100.0,
                    200.0,
                    200.0,
                    200.0,
                    250.0,
                    300.0,
                    300.0,
                    350.0,
                    250.0
                  ],
                  "PH": [
                    6.2,
                    5.866,
                    5.531,
                    5.3,
                    5.3,
                    5.3,
                    5.3,
                    5.3,
                    5.535,
                    6.118
                  ],
                  "PHUnits": 0,
                  "EC": None,
                  "ESP": None,
                  "ECMetadata": None,
                  "CLMetadata": None,
                  "ESPMetadata": None,
                  "PHMetadata": None,
                  "Name": "Chemical",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Soils.Water, Models",
                  "Thickness": [
                    200.0,
                    100.0,
                    200.0,
                    200.0,
                    200.0,
                    250.0,
                    300.0,
                    300.0,
                    350.0,
                    250.0
                  ],
                  "InitialValues": [
                    0.45499999999999996,
                    0.4329251506625542,
                    0.4108503013251085,
                    0.3887754519876628,
                    0.3686449949074143,
                    0.36670060265021703,
                    0.3594339116377837,
                    0.3568958338196487,
                    0.3502228283681532,
                    0.3492558697205646
                  ],
                  "RelativeTo": "LL15",
                  "FilledFromTop": True,
                  "Name": "Water",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Soils.Nutrients.Nutrient, Models",
                  "Name": "Nutrient",
                  "ResourceName": "Nutrient",
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.WaterModel.WaterBalance, Models",
                  "SummerDate": "1-jun",
                  "SummerU": 8.55,
                  "SummerCona": 3.775,
                  "WinterDate": "1-nov",
                  "WinterU": 8.55,
                  "WinterCona": 3.775,
                  "DiffusConst": 40.0,
                  "DiffusSlope": 16.0,
                  "Salb": 0.13,
                  "CN2Bare": 76.0,
                  "CNRed": 20.0,
                  "CNCov": 0.8,
                  "DischargeWidth": 160.0,
                  "CatchmentArea": 10000.0,
                  "PSIDul": -100.0,
                  "Thickness": [
                    200.0,
                    100.0,
                    200.0,
                    200.0,
                    200.0,
                    250.0,
                    300.0,
                    300.0,
                    350.0,
                    250.0
                  ],
                  "SWCON": [
                    0.15220540364583338,
                    0.18515385150806513,
                    0.21877764915203454,
                    0.2530977757183513,
                    0.28436150314218656,
                    0.27586244631565077,
                    0.2899272006646377,
                    0.2816709153195647,
                    0.24340303510649974,
                    0.20110916070288432
                  ],
                  "KLAT": None,
                  "Name": "SoilWater",
                  "ResourceName": "WaterBalance",
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Soils.Solute, Models",
                  "Thickness": [
                    200.0,
                    200.0,
                    200.0,
                    200.0,
                    200.0,
                    200.0,
                    200.0,
                    200.0,
                    200.0,
                    200.0
                  ],
                  "InitialValues": [
                    1.45,
                    1.342,
                    1.235,
                    1.096,
                    0.889,
                    0.683,
                    0.476,
                    0.269,
                    0.145,
                    0.145
                  ],
                  "InitialValuesUnits": 0,
                  "WaterTableConcentration": 0.0,
                  "D0": 0.0,
                  "Exco": None,
                  "FIP": None,
                  "DepthConstant": 0.0,
                  "MaxDepthSoluteAccessible": 0.0,
                  "RunoffEffectivenessAtMovingSolute": 0.0,
                  "MaxEffectiveRunoff": 0.0,
                  "Name": "NO3",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Soils.Solute, Models",
                  "Thickness": [
                    200.0,
                    200.0,
                    200.0,
                    200.0,
                    200.0,
                    200.0,
                    200.0,
                    200.0,
                    200.0,
                    200.0
                  ],
                  "InitialValues": [
                    1.45,
                    1.342,
                    1.235,
                    1.096,
                    0.889,
                    0.683,
                    0.476,
                    0.269,
                    0.145,
                    0.145
                  ],
                  "InitialValuesUnits": 0,
                  "WaterTableConcentration": 0.0,
                  "D0": 0.0,
                  "Exco": None,
                  "FIP": None,
                  "DepthConstant": 0.0,
                  "MaxDepthSoluteAccessible": 0.0,
                  "RunoffEffectivenessAtMovingSolute": 0.0,
                  "MaxEffectiveRunoff": 0.0,
                  "Name": "NH4",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Soils.Solute, Models",
                  "Thickness": [
                    200.0,
                    200.0,
                    200.0,
                    200.0,
                    200.0,
                    200.0,
                    200.0,
                    200.0,
                    200.0,
                    200.0
                  ],
                  "InitialValues": [
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0
                  ],
                  "InitialValuesUnits": 1,
                  "WaterTableConcentration": 0.0,
                  "D0": 0.0,
                  "Exco": None,
                  "FIP": None,
                  "DepthConstant": 0.0,
                  "MaxDepthSoluteAccessible": 0.0,
                  "RunoffEffectivenessAtMovingSolute": 0.0,
                  "MaxEffectiveRunoff": 0.0,
                  "Name": "Urea",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                },
                {
                  "$child": "Models.Soils.SoilTemp.SoilTemperature, Models",
                  "Name": "SoilTemperature",
                  "ResourceName": None,
                  "Children": [],
                  "Enabled": True,
                  "ReadOnly": False
                }
              ],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.Manager, Models",
              "Code": "using Models.Soils;\nusing System;\nusing Models.Core;\nusing Models.PMF;\nusing APSIM.Shared.Utilities;\nusing Models.Surface;\n\nnamespace Models\n{\n    [Serializable]\n    public class Script : Model\n    {\n        [Link] Clock Clock;\n        [Link] SurfaceOrganicMatter SOM;\n        [Link(IsOptional = True)] private IPlant existingCrop;\n        [Link] ISummary Summary;\n        [Separator(\"Applies tillage to an existing crop when it is ending\")]\n        //[Description(\"Crop\")]\n        //public IPlant Crop { get; set; }\n        \n        [Description(\"Fraction of Residues To Incorporate (0-1)\")]\n        public double Fraction { get; set; }\n        \n        [Description(\"Depth of Tillage (mm)\")]\n        public double Depth { get; set; }\n        \n        \n        public enum ApplyTillage { yes, no }\n        [Description(\"Do you want to till the land?\")] \n        public ApplyTillage AppplyTillageOption { get; set; }\n\n        [EventSubscribe(\"PlantEnding\")]\n        private void OnEndCrop(object sender, EventArgs e)\n        {\n            Model existingCrop = sender as Model;\n            if (existingCrop != None && existingCrop.Name.ToLower() == (existingCrop as IModel).Name.ToLower()&& ApplyTillage.yes ==AppplyTillageOption)\n            \n            {\n           \n            SOM.Incorporate(Fraction, Depth);\n            \n           }\n           \n           \n             Summary.WriteMessage(this, $\"Tillage is applied to: {existingCrop.Name}\", MessageType.Information);\n        }\n    }\n}",
              "Parameters": [
                {
                  "Key": "Fraction",
                  "Value": "0.65"
                },
                {
                  "Key": "Depth",
                  "Value": "350"
                },
                {
                  "Key": "AppplyTillageOption",
                  "Value": "no"
                }
              ],
              "Name": "Tillage",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            },
            {
              "$child": "Models.Report, Models",
              "VariableNames": [
                "diff of ([Soil].Nutrient.TotalC[1])/1000 from [Clock].StartOfSimulation to [Clock].EndOfSimulation as changeincarbon",
                "diff of sum([Soil].Nutrient.TotalC[1:2])/1000 from [Clock].StartOfSimulation to [Clock].EndOfSimulation as changeincarbon2",
                "",
                "changeincarbon + changeincarbon2 as to_carb",
                "",
                "(-1 * to_carb *(44/12))*1000 as co_eq"
              ],
              "EventNames": [
                "[Clock].EndOfSimulation"
              ],
              "GroupByVariableName": None,
              "Name": "Carbon1",
              "ResourceName": None,
              "Children": [],
              "Enabled": True,
              "ReadOnly": False
            }
          ],
          "Enabled": True,
          "ReadOnly": False
        },
        {
          "$child": "Models.Memo, Models",
          "Text": "When working with perennial crops that are harvested or grazed frequently, often an output that plots the accumulation of growth during a year is wanted. That output should be zeroed between years. If simulating a site in the Southern Hemisphere, usually the end of the year will be in June or July (winter). Here this is achieved as \"CumulativeAnnualNetGrowth\" using reporting at the end of the month (see in MonthlyReporting) and:\n\n\t\tSum of ([Ryegrass].NetGrowthWt + [WhiteClover].NetGrowthWt) \n\t\t\tfrom 1-Jul to [Clock].Today as CumulativeAnnualNetGrowth\n\nwith the output shown as the orange line below. Note that here it is necessary to calculate the output summing ryegrass and white clover. For comparison, the pattern of accumulation is plotted against the annual total (green line) and that was specified in AnnualReporting using:\n\n\t\tSum of ([Ryegrass].NetGrowthWt + [WhiteClover].NetGrowthWt) \n\t\t\tfrom [AnnualReporting].DayAfterLastOutput to [Clock].Today \n\t\t\tas SumYearlyNetGrowth\n",
          "Name": "Memo",
          "ResourceName": None,
          "Children": [],
          "Enabled": True,
          "ReadOnly": False
        }
      ],
      "Enabled": True,
      "ReadOnly": False
    }
  ],
  "Enabled": True,
  "ReadOnly": False
}

